#!/usr/bin/env perl
use warnings;
use strict;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=pod

=head1 NAME

mapsnps.pl - Map genomic SNPs to known human genes / proteins

=head1 SYNOPSIS

mapsnsp.pl [options] infile [>outfile]

where options are:

  -c          annotate coding SNPs only

  -m          annotate missense coding SNPs only

  -n          skip alleles which are missing a matching reference nt

  -y <batch>  save PolyPhen-2 queries for missense SNPs to 'batch' file

  -g hgXX     genome assembly annotations used; default is hg19 (GRCh37),
              also supported is hg18 (NCBI36)

  -i hgYY     input SNPs chromosome coordinate assembly version; default
              is to assume the same assembly as used for the annotations,
              otherwise coordinates are transformed to match annotations
              assembly version using liftOver tool

  -x N        Apply filtering of increased stringency when mapping SNP
              genomic positions to genes, N=0..3:
                0 - include all known gene transcripts, no filtering
                1 - include only known canonical transcripts (default)
                2 - include only known canonical transcripts which
                    are also annotated as part of CCDS
                3 - include only known canonical transcripts with
                    exact match to CCDS

  -u          attempt to map translated CDS to a sequence in the local
              UniProtKB database, default is to utilize UCSC mappings

  -U          as above but utilize (a much slower) BLAST search instead
              of the exact matches; successful match requires 98% identity
              level and no gaps in alignement

  -s prot.fa  save protein sequences for all CDS's to which at least
              a single missense SNP was mapped to a FASTA-formatted
              file 'prot.fa'; if '-u/-U' option was also in effect then
              only save those sequences that failed to match UniProtKB

  -d [user:password@]host[:port]
              MySQL database server specification; default is to use
              anonymous localhost connection with no password

  -r N/M      partition input into M equally sized stripes and process
              stripe N

  -v N        verbosity level; default N=1, set N=0 for quiet operation

=head1 DESCRIPTION

Input file should be in a tab-delimited text format, with single
genomic SNPs specification per each line:

chrom:pos a1/a2[/a3[/a4]] [a1/a2 ...]

   * chrom  - chromosome identifier, e.g. chr10
   * pos    - SNP chromosome position, one-based coordinates
   * a1/a2  - reference/observed allelic variants (nucleotides)
              in plus strand orientation

Example:

   chr1:1267483   G/A
   chr1:1158631   A/C/G/T
   chr2:167262274 C/T
   chr4:264904    G/A
   chr7:122261636 G/T
   chr16:53698869 T/C

=head1 AUTHORS

IAA

=head1 SUBVERSION

 $LastChangedDate: 2011-05-20 02:38:48 -0400 (Fri, 20 May 2011) $
 $LastChangedRevision: 364 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Getopt::Std;
use Pod::Usage;
use DBI;
use Storable;
use File::Temp;

my $PROGRESSBAR;

BEGIN {
  $Getopt::Std::STANDARD_HELP_VERSION = 1;
  use FindBin qw($RealBin);
  BEGIN {
    unless ($ENV{'PPH'}) {
      my $binpath = $RealBin;
      $binpath =~ s|/[^/]+$||;
      $ENV{'PPH'} = $binpath;
    }
  }
  use lib "$ENV{PPH}/perl";
  eval { require Term::ProgressBar; };
  $PROGRESSBAR = $@ ? 0 : 1;
}

use PPH::Config;
use PPH::Seq;

# Standard genetic code
my %TRANS = (
  AAA => 'K',   ATA => 'I',   AGA => 'R',   ACA => 'T',
  AAG => 'K',   ATG => 'M',   AGG => 'R',   ACG => 'T',
  AAT => 'N',   ATT => 'I',   AGT => 'S',   ACT => 'T',
  AAC => 'N',   ATC => 'I',   AGC => 'S',   ACC => 'T',

  TAA => '*',   TTA => 'L',   TGA => '*',   TCA => 'S',
  TAG => '*',   TTG => 'L',   TGG => 'W',   TCG => 'S',
  TAT => 'Y',   TTT => 'F',   TGT => 'C',   TCT => 'S',
  TAC => 'Y',   TTC => 'F',   TGC => 'C',   TCC => 'S',

  GAA => 'E',   GTA => 'V',   GGA => 'G',   GCA => 'A',
  GAG => 'E',   GTG => 'V',   GGG => 'G',   GCG => 'A',
  GAT => 'D',   GTT => 'V',   GGT => 'G',   GCT => 'A',
  GAC => 'D',   GTC => 'V',   GGC => 'G',   GCC => 'A',

  CAA => 'Q',   CTA => 'L',   CGA => 'R',   CCA => 'P',
  CAG => 'Q',   CTG => 'L',   CGG => 'R',   CCG => 'P',
  CAT => 'H',   CTT => 'L',   CGT => 'R',   CCT => 'P',
  CAC => 'H',   CTC => 'L',   CGC => 'R',   CCC => 'P',
);

# Degenerate positions in codons (for standard genetic code):
#
# Nei M & Kumar S (2000) Molecular Evolution and Phylogenetics. Oxford University Press, New York.
# http://www.megasoftware.net/WebHelp/part_iv___evolutionary_analysis/computing_evolutionary_distances/distance_models/synonymouse_and_nonsynonymous_substitution_models/hc_kumar_comeron_method.htm
#
#   L (Leu):
#   TTA, TTG, CTT, CTC, CTA, CTG
#   202  202  004  004  204  204
#   s s  s s            s    s
#
#   R (Arg):
#   AGA, AGG, CGT, CGC, CGA, CGG
#   202  202  004  004  204  204
#   c s  c s            c    c
#
#   S (Ser):
#   AGT, AGC, TCT, TCC, TCA, TCG
#   202  202  004  004  204  204
#   c s  c s            c    c
#
#   c - complex twofold site; s - simple twofold site
#
# Note: in the following table, '3' denotes complex twofold sites
#
my %DGEN = (
  AAA => [0,0,2],   ATA => [0,0,3],   AGA => [3,0,2],   ACA => [0,0,4],
  AAG => [0,0,2],   ATG => [0,0,0],   AGG => [3,0,2],   ACG => [0,0,4],
  AAT => [0,0,2],   ATT => [0,0,3],   AGT => [3,0,2],   ACT => [0,0,4],
  AAC => [0,0,2],   ATC => [0,0,3],   AGC => [3,0,2],   ACC => [0,0,4],

  TAA => [0,0,0],   TTA => [2,0,2],   TGA => [0,0,0],   TCA => [3,0,4],
  TAG => [0,0,0],   TTG => [2,0,2],   TGG => [0,0,0],   TCG => [3,0,4],
  TAT => [0,0,2],   TTT => [0,0,2],   TGT => [0,0,2],   TCT => [0,0,4],
  TAC => [0,0,2],   TTC => [0,0,2],   TGC => [0,0,2],   TCC => [0,0,4],

  GAA => [0,0,2],   GTA => [0,0,4],   GGA => [0,0,4],   GCA => [0,0,4],
  GAG => [0,0,2],   GTG => [0,0,4],   GGG => [0,0,4],   GCG => [0,0,4],
  GAT => [0,0,2],   GTT => [0,0,4],   GGT => [0,0,4],   GCT => [0,0,4],
  GAC => [0,0,2],   GTC => [0,0,4],   GGC => [0,0,4],   GCC => [0,0,4],

  CAA => [0,0,2],   CTA => [2,0,4],   CGA => [3,0,4],   CCA => [0,0,4],
  CAG => [0,0,2],   CTG => [2,0,4],   CGG => [3,0,4],   CCG => [0,0,4],
  CAT => [0,0,2],   CTT => [0,0,4],   CGT => [0,0,4],   CCT => [0,0,4],
  CAC => [0,0,2],   CTC => [0,0,4],   CGC => [0,0,4],   CCC => [0,0,4],
);

# 0 - transition, 1 - transversion
my %TRNV = (
  'A' => {
    'A' => undef,
    'C' => 1,
    'G' => 0,
    'T' => 1
  },
  'C' => {
    'A' => 1,
    'C' => undef,
    'G' => 1,
    'T' => 0
  },
  'G' => {
    'A' => 0,
    'C' => 1,
    'G' => undef,
    'T' => 1
  },
  'T' => {
    'A' => 1,
    'C' => 0,
    'G' => 1,
    'T' => undef
  }
);

my @header = qw(
  snp_pos
  str
  gene
  transcript
  ccid
  ccds
  cciden
  refa
  type
  nt1
  nt2
  flanks
  ntpos
  trv
  cpg
  jxdon
  jxacc
  exon
  cexon
  jxc
  dgn
  frame
  cdn1
  cdn2
  aa1
  aa2
  aapos
  spmap
  spacc
  spname
  refs_acc
  dbrsid
  dbobsrvd
  dbavHet
  dbavHetSE
  dbRmPaPt
);

my @format = qw(
  %-16s
  %3s
  %12s
  %12s
  %5s
  %12s
  %6s
  %4s
  %12s
  %3s
  %3s
  %6s
  %6s
  %3s
  %3s
  %7s
  %7s
  %7s
  %7s
  %3s
  %3s
  %5s
  %4s
  %4s
  %3s
  %3s
  %8s
  %5s
  %10s
  %12s
  %12s
  %10s
  %8s
  %10s
  %10s
  %8s
);

my $formatstr = join("\t", @format);
my $headerstr = sprintf('#'.$formatstr, @header) . "\tcomments\n";

# binRange() constants
my $binFirstShift          =   17;
my $binNextShift           =    3;
my $binOffsetOldToExtended = 4681;
my @binOffsets             = ( 512+64+8+1, 64+8+1, 8+1, 1, 0 );
my @binOffsetsExtended     = ( 4096+512+64+8+1, 512+64+8+1, 64+8+1, 8+1, 1, 0 );

# Aux files
my $fasta_uniprot  = "$CONFIG{UNIPROT}.seq";
my $fasta_varsplic = "$CONFIG{UNIPROT}.varsplic";

$|++;

my %opts;
pod2usage('ERROR: No arguments specified') unless @ARGV;
getopts('chmnuUd:g:i:r:s:v:x:y:', \%opts) or pod2usage('ERROR: Wrong argument(s)');
pod2usage(-verbose=>2) if $opts{'h'};
pod2usage('ERROR: No input file specified') unless @ARGV;
my $infile = shift;

# Process options
my $VERBOSE   = (exists $opts{'v'} && length $opts{'v'}) ? $opts{'v'} : 1;
my $target_assembly  = $opts{'g'} || 'hg19';
die "ERROR: Unsupported assembly version specified: $target_assembly\n"
  unless $target_assembly eq 'hg18' || $target_assembly eq 'hg19';
my $query_assembly = $opts{'i'} || $target_assembly;
die "ERROR: Unsupported assembly version specified: $query_assembly\n"
  unless $query_assembly eq 'hg18' || $query_assembly eq 'hg19';
my $CHRPATH   = "$CONFIG{GOLDENPATH}/$target_assembly/chromosomes";
my $dbopt     = $opts{'d'} || $CONFIG{'GOLDENDSN'} || '';
my $coding_only   = $opts{'c'} || 0;
my $missense_only = $opts{'m'} || 0;
my $skip_noref    = $opts{'n'} || 0;
my $map2uniprot   = $opts{'u'} || 0;
$map2uniprot = 2 if $opts{'U'}; # super-meticious UniProt mapping using BLAST
my $stringency    = exists $opts{'x'} ? $opts{'x'} : 1;
die "ERROR: Illegal option argument: -x $stringency\n" unless $stringency =~ /^[0-3]$/;
my $pph2output    = $opts{'y'} || 0;
my $seqfile = $opts{'s'} || '';
my ($pph2formatstr, $pph2headerstr);
if ($pph2output) {
  open(PPH2OUT, ">$pph2output") or die "ERROR: Can't create file: $pph2output\n";
  $pph2formatstr = join("\t", qw{ %-30s %12s %6s %3s %3s });
  $pph2headerstr = sprintf('#'.$pph2formatstr, qw{ snp_id acc pos aa1 aa2 }) . "\tcomments\n";
}
$coding_only = $missense_only if $missense_only;

# Check if we are running under SGE or LSF and if this is an array job
my ($grid_mode, $jobid, $taskid, $maxtask, $grid_msg);
if (exists $ENV{'SGE_ROOT'} && exists $ENV{'JOB_ID'}) {       # SGE
  $grid_mode = 'SGE';
  $jobid  = $ENV{'JOB_ID'} ne 'undefined' ? $ENV{'JOB_ID'} : 0;
  $taskid = exists $ENV{'SGE_TASK_ID'} && $ENV{'SGE_TASK_ID'} ne 'undefined' ? $ENV{'SGE_TASK_ID'} : 0;
} elsif ($ENV{'LSB_QUEUE'} && exists $ENV{'LSB_JOBID'}) {     # LSF
  $grid_mode = 'LSF';
  $jobid  = $ENV{'LSB_JOBID'};
  $taskid = exists $ENV{'LSB_JOBINDEX'} ? $ENV{'LSB_JOBINDEX'} : 0;
}
my $array_mode = $jobid && $taskid ? 1 : 0;

if ($grid_mode) {
  # Adjust PATH for a grid job
  if ($CONFIG{'GRID_PATH'}) {
    my $ARCH = `uname -m`; chomp $ARCH;
    $ENV{'PATH'} = eval "qq{$CONFIG{GRID_PATH}}";
  }
  $grid_msg = "I am $grid_mode job $jobid \@ $ENV{HOSTNAME}\n";
}

if ($array_mode) {
  # GGI_STRIPES overwrites any array size specified initially
  $maxtask = $ENV{'GGI_STRIPES'} || $ENV{'SGE_TASK_LAST'} || $ENV{'LSB_JOBINDEX_END'};
  $grid_msg = "I am $grid_mode job $jobid task $taskid of $maxtask \@ $ENV{HOSTNAME}\n";
}

my ($stripe_no, $stripes);
if ($opts{'r'}) {
  ($stripe_no, $stripes) = $opts{'r'} =~ m|^(\d+)/(\d+)$|;
  die "ERROR: Illegal option's argument: -r $opts{'r'}\n" unless $stripe_no && $stripes;
}

if ($seqfile) {
  warn "WARNING: File exists, overwriting: $seqfile\n" if -e $seqfile && -s $seqfile;
  open(SEQFILE, ">$seqfile") or die "Can't create file: $seqfile\n";
}

# Debugging
warn $grid_msg if defined $grid_msg;

if ($array_mode) {
  warn "WARNING: Explicit input stripe specification via '-r' option ignored in $grid_mode array mode\n" if $opts{'r'};
  # Additional sanity checks specific to PPHWeb2 service:
  # In the web service pipeline mode (under apache2 user), a special environment
  # variable has to be set since SGE_TASK_LAST cannot be altered for a pending job
  if ($grid_mode eq 'SGE' && $ENV{'USER'} eq 'apache2') {
    die "ERROR: Missing GGI_STRIPES environment variable while in SGE array mode\n"
      unless exists $ENV{'GGI_STRIPES'} && length $ENV{'GGI_STRIPES'};
    die "ERROR: Illegal GGI_STRIPES value while in SGE array mode: $ENV{GGI_STRIPES}\n"
      unless $ENV{'GGI_STRIPES'} > 0;
  }
  ($stripe_no, $stripes) = ($taskid, $maxtask);
}

warn "Initializing ...\n" if $VERBOSE;

# Count how many input lines we have and how many of these are valid query lines,
# e.g., sans comments and empty lines. Used by progress bar and with the stripes option.
my $input_line_count  = 0;
my $input_query_count = 0;
if ($VERBOSE || $stripes) {
  open(FIN, $infile) or die "ERROR: Can't open input file: $infile\n";
  while (<FIN>) {
    $input_line_count++;
    # Skip comments and empty lines
    next if /^#/ || /^\s*$/;
    $input_query_count++;
  }
  close(FIN);
}

# Calculate stripe range
my ($range_start, $range_end) = stripe_range($stripe_no, $stripes, $input_query_count) if $stripes;

my $username  = '';
my $password  = '';
my $hostname  = 'localhost';
my $portno    = '';
if ($dbopt) {
  my ($user, $host) = split /@/, $dbopt;
  my ($port, $pass);
  # embedded @ sign, this is a user[:pass]@host[:port] specification
  if ($host) {
    ($user, $pass) = split /:/, $user;
    ($host, $port) = split /:/, $host;
  # no embedded @ sign, this is a host[:port] specification
  } else {
    $host = $user; $user = '';
    ($host, $port) = split /:/, $host;
  }
  $username = $user if $user;
  $password = $pass if $pass;
  $hostname = $host || 'localhost';
  $portno   = $port if $port;
}
my $dsn = "DBI:mysql:database=$target_assembly;host=$hostname";
$dsn .= ";port=$portno" if $portno;

my $dbsnp = $target_assembly eq 'hg18' ? 'snp130' : 'snp131';

my $dbh = DBI->connect($dsn, $username, $password, {'RaiseError'=>1, 'PrintError'=>1})
  or die $DBI::errstr;

# Default is to map to all known protein-coding genes
#
# Note: The following odd-looking clause:
#    LEFT JOIN ccdsKgMap AS ccds2...
#    WHERE ... ccds2.geneId IS NULL
# is here to ensure we only select a single CCDS match
# with the highest similarity to this known transcript.
# For details see:
#   http://dev.mysql.com/doc/refman/5.1/en/example-maximum-column-group-row.html
#
my $selectKnownGene = q{
  SELECT kg.*, kc.clusterId, ccds.ccdsId, ccds.cdsSimilarity
    FROM
      knownGene AS kg
      LEFT JOIN knownCanonical AS kc    ON kg.name = kc.transcript
      LEFT JOIN ccdsKgMap      AS ccds  ON kg.name = ccds.geneId
      LEFT JOIN ccdsKgMap      AS ccds2 ON ccds.geneId = ccds2.geneId AND ccds.cdsSimilarity < ccds2.cdsSimilarity
    WHERE
      kg.chrom    = ? AND
      kg.txStart <= ? AND
      kg.txEnd   >= ? AND
      kg.cdsEnd - kg.cdsStart > 0 AND
      kg.exonCount > 0 AND
      ccds2.geneId IS NULL
};
if ($stringency >= 1) { $selectKnownGene .= ' AND kc.clusterId IS NOT NULL'; }
if ($stringency >= 2) { $selectKnownGene .= ' AND ccds.ccdsId  IS NOT NULL'; }
if ($stringency == 3) { $selectKnownGene .= ' AND ccds.cdsSimilarity = 1'; }
# In rare cases, several different CCDS entries can map to a single knownGene transcript
# with exactly the same similarity score. GROUP BY clause ensures that only first of
# such matches is selected thus avoiding duplicate knownGene transcripts mappings
# in output (which only differ in their CCDS mappings). Note that this would result
# in selecting only the first CCDS entry for output and dropping the rest.
$selectKnownGene .= ' GROUP BY kg.name, ccds.cdsSimilarity';

my $selectXref = q{
  SELECT * FROM kgXref WHERE kgID = ?
};

my $selectdbSNP = qq{
  SELECT
    s.name, s.chrom, s.chromStart, s.chromEnd, s.strand,
    s.observed, s.class, s.valid, s.avHet, s.avHetSE, s.func,
    a.humanAllele, a.humanStrand, a.chimpAllele, a.chimpStrand,
    a.orangAllele, a.orangStrand, a.macaqueAllele, a.macaqueStrand
  FROM $dbsnp AS s LEFT JOIN ${dbsnp}OrthoPt2Pa2Rm2 AS a
  USING(name)
  WHERE
    s.chrom      = ? AND
    s.bin        = ? AND
    s.chromStart = ? AND
    s.chromEnd   = ? AND
    s.class      = 'single'
};

my $sthKnownGene = $dbh->prepare($selectKnownGene) or die $dbh->errstr;
my $sthXref      = $dbh->prepare($selectXref)      or die $dbh->errstr;
my $sthdbSNP     = $dbh->prepare($selectdbSNP)     or die $dbh->errstr;

warn "Processing ...\n" if $VERBOSE;

my %totals;
my $limit = 0;
my $input_line_no  = 0;
my $input_query_no = 0;
my $range_line_no  = 0;
my %genes;
my $input_range_count;
if      ($stripes) {
  if ($range_start == 0 && $range_end == 0) {
    $input_range_count = 0;
  } else {
    $input_range_count = $range_end - $range_start + 1
  }
  die "ERROR: Input out of range (stripe $stripe_no / $stripes with $input_query_count query lines in input)\n"
    unless $input_range_count;
} elsif ($VERBOSE) {    # $input_query_count is only set in either -r/array or verbose modes
  $input_range_count = $input_query_count;
  die "ERROR: Nothing to do ($input_query_count query lines in input)\n"
    unless $input_range_count;
}
my $progress = Term::ProgressBar->new({count=>$input_range_count}) if $PROGRESSBAR && $VERBOSE;  # set up a progress bar
my $next_update = 0;
my $header_printed = 0;
my $pph2header_printed = 0;
open(FIN, $infile) or die "ERROR: Can't open input file: $infile\n";
LINE: while (<FIN>) {
  $input_line_no++;   # unfiltered input line number, used in error/warning messages
  # Skip comments and empty lines
  next if /^#/ || /^\s*$/;
  $input_query_no++;  # input query line number
  if ($stripes) {
    next LINE if $input_query_no < $range_start;
    last LINE if $input_query_no > $range_end;
  }
  $totals{'input'}++; # total query input lines to process (i.e., within range but without comments)
  $range_line_no++;   # count, same as above
  $next_update = $progress->update($range_line_no) if $PROGRESSBAR && $VERBOSE && $input_line_no >= $next_update;
  chomp;
  # Process optional user comments as separated by a '#' character
  my $comments = $1 if s/\s *#\s*(.*)$//;
  my ($chrpos, @a) = split;
  my ($chrom, $snppos);
  if ($chrpos =~ /^chr(\d{1,2}|[MXY]):(\d+)/i) {
    $chrom  = 'chr' . $1;
    $snppos = $2;
  } else {
    warn "WARNING: Illegal genomic coordinate specification skipped at input line ($input_line_no): $chrpos\n";
    next;
  }
  my @o_alleles;   # these are original alleles specified in input,
  my %o_alleles;   # they are always on plus strand
  for (my $i=0; $i<=$#a; $i++) {
    $a[$i] =~ s/\s+//g;
    warn("WARNING: Illegal allele specification ($a[$i]) skipped at input line ($input_line_no)\n"), next
      unless $a[$i] =~ m|^[ACGTacgt]/[ACGTacgt](?:[/,][ACGTacgt])*$|;
    foreach my $nuc (split m|[/,]|, $a[$i]) {
      $nuc = uc $nuc;
      push @o_alleles, $nuc unless $o_alleles{$nuc};
      $o_alleles{$nuc} = 1;
    }
  }
  warn("WARNING: No valid alleles found at input line ($input_line_no)\n"), next LINE unless @o_alleles;
  last LINE if $limit and $input_line_no > $limit;

  my $snpstart = $snppos   - 1; # convert to zero-based UCSC start coord
  my $snpend   = $snpstart + 1; # dummy UCSC-style end coord

  # liftOver input coordinates to target assembly if query uses different assembly version
  if ($query_assembly ne $target_assembly) {
    my $fhq = File::Temp->new();
    my $fnq = $fhq->filename;     # query in BED format
    my $fht = File::Temp->new();
    my $fnt = $fht->filename;     # target in BED format
    my $fhu = File::Temp->new();
    my $fnu = $fhu->filename;     # unMapped
    print($fhq "$chrom\t$snpstart\t$snpend\n");
    my $lofile =
      "$CONFIG{GOLDENPATH}/$query_assembly/liftOver/" .
      $query_assembly . 'To' . ucfirst($target_assembly) .'.over.chain';
    die "ERROR: liftOver file not found: $lofile\n" unless -e $lofile;
    my $rc = system('liftOver', $fnq, $lofile, $fnt, $fnu) == 0 or
      warn("WARNING: liftOver failed at input line ($input_line_no): $!\n"), next LINE;
    my ($mapped, $unmapped);
    { local $/; $mapped   = <$fht>; }
    { local $/; $unmapped = <$fhu>; }
    chomp($mapped, $unmapped);
    if (length $unmapped) {
      $unmapped =~ s/^#//s;
      $unmapped =~ s/\n/: /mg;
      $unmapped =~ s/\t+/ /mg;
      warn "WARNING: liftOver error at input line ($input_line_no): $unmapped\n";
      next LINE;
    } else {
      warn("WARNING: liftOver output empty at input line ($input_line_no)\n"), next LINE
        unless length $mapped;
      ($chrom, $snpstart, $snpend) = split ' ', $mapped;
    }
  }

  $totals{'processed'}++; # total query input lines processed (after validating & without comments)

  # Execute query vs knownGene table of gene transcripts
  $sthKnownGene->execute($chrom, $snpstart, $snpend) or die $sthKnownGene->errstr;

  GENE: while (my $knownGene = $sthKnownGene->fetchrow_hashref()) {

    my $plusflag = $knownGene->{strand} eq '+' ? 1 : 0;
    my $nbstrand = $plusflag ? '+' : 'm';   # nibFrag strand option argument is '+' or 'm'

    # Fetch gene's full nucleotide sequence unless already cached
    my $full_nucseq;
    if (exists $genes{$knownGene->{name}}{full_nucseq}) {
      $full_nucseq = $genes{$knownGene->{name}}{full_nucseq};
    } else {
      my $nibcommand =
          "nibFrag -upper -name= $CHRPATH/$chrom.nib $knownGene->{txStart} $knownGene->{txEnd} $nbstrand stdout";
      $full_nucseq = `$nibcommand`;
      unless (defined $full_nucseq) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Failed to extract full sequence ($!), CDS skipped\n";
        next GENE;
      }
      $full_nucseq =~ s/^>.*?\n//s;
      $full_nucseq =~ s/\n//sg;
      unless (length $full_nucseq) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Empty full sequence, CDS skipped\n";
        next GENE;
      }
      if ($full_nucseq =~ /[^ACTG]/) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal or ambiguous character(s) found in gene sequence\n";
      }
      $genes{$knownGene->{name}}{full_nucseq} = $full_nucseq;
    }

    # Extract chromosome nt at the variation position and at its 5' and 3' flanks (for CpG annotation)
    # and complement working copies of input alleles (array and hash) for minus strand transcripts.
    # Note1: all three nt's are on the coding strand, with flanks enumerated in the direction of transcription (5'3')
    # Note2: corresponding flanking nt will be undef for a SNP the in first / last positions of the transcript
    my @alleles = @o_alleles;   # these are working copies of the original input alleles, they will be
    my %alleles = %o_alleles;   # complemented for positions mapped to transcripts on minus strand
    my $nucseq_len    = length $full_nucseq;
    my $nucseq_offset = $plusflag ? $snpstart - $knownGene->{txStart} : $knownGene->{txEnd} - $snpend;
    # Transcript nt at the SNP position
    my $trn_nt = substr($full_nucseq, $nucseq_offset, 1);
    # Reference assembly nt at the SNP position, will be complemented later for transcripts on minus strand
    my $ref_nt = $trn_nt;
    unless (defined $trn_nt) {
      warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Failed to extract transcript nt at SNP position ($snpend), transcript skipped\n";
      next GENE;
    }
    unless ($trn_nt =~ /^[ACTG]$/) {
      warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal nucleotide ($trn_nt) in transcript sequence at SNP position ($snpend), transcript skipped\n";
      next GENE;
    }
    my $trn_nt5 = $nucseq_offset > 0               ? substr($full_nucseq, $nucseq_offset - 1, 1) : 'N';
    my $trn_nt3 = $nucseq_offset < $nucseq_len - 1 ? substr($full_nucseq, $nucseq_offset + 1, 1) : 'N';
    unless ($plusflag) {
      # Complement back to plus (reference) strand of assembly
      $ref_nt  =~ tr/ACGT/TGCA/;
      # Complement input alleles to minus (transcript coding) strand
      %alleles = ();
      for (my $i=0; $i<@alleles; $i++) {
        $alleles[$i] =~ tr/ACGT/TGCA/;
        $alleles{$alleles[$i]} = 1;
      }
    }
    ###my $o_ref_nt = $chr_nt; # alias, for historical reasons (always on plus strand)
    # Check for the situation when the reference sequence nt is different from any of the variant alleles submitted for analysis.
    unless ($o_alleles{$ref_nt}) {
      warn("WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) None of the input alleles (".join('/', @o_alleles).") matches reference nt ($ref_nt)\n");
      $totals{'processed'}--, next LINE if $skip_noref;
    }

    $knownGene->{cdsSimilarity} = sprintf "%.3g", $knownGene->{cdsSimilarity} if defined $knownGene->{cdsSimilarity};

    my @exstarts = split /,/, $knownGene->{exonStarts};
    my @exends   = split /,/, $knownGene->{exonEnds};

    my $excount     = 0;
    my $cds_excount = 0;
    my $cds_length  = 0;
    my $cds_nucseq  = '';

    # Scan exons in reverse direction if gene is on the minus strand
    my ($exon_first, $exon_last, $exon_step); # first / last exon indices, zero-based
    if ($plusflag) {
      ($exon_first, $exon_last, $exon_step) = (0, $knownGene->{exonCount} - 1,  1);
    } else {
      ($exon_first, $exon_last, $exon_step) = ($knownGene->{exonCount} - 1, 0, -1);
    }
    my $exon_limit = $exon_last + $exon_step;

    my %chr2cds;
    my $snptype;
    my ($exno, $cds_exno);
    my ($jdist_donor, $jdist_acceptor);
    my ($cds_firstcodon, $cds_lastcodon);
    EXON: for (my $i=$exon_first; $i!=$exon_limit; $i+=$exon_step) {

      $excount++;

      my $start = $exstarts[$i];
      my $end   = $exends[$i];

      my $cds_exstart = $start <= $knownGene->{cdsStart} ? $knownGene->{cdsStart} : $start;
      my $cds_exend   = $end   <= $knownGene->{cdsEnd}   ? $end : $knownGene->{cdsEnd};

      # Annotate SNP position
      unless ($snptype) {
        # SNP position is inside this exon
        if ($snpstart >= $start && $snpend <= $end) {
          $exno = $excount;
          # SNP position is in 5'-UTR
          if      ($snpstart < $cds_exstart) {
            $snptype = $plusflag ? 'utr-5' : 'utr-3';
          # SNP position is in 3'-UTR
          } elsif ($snpend  > $cds_exend)    {
            $snptype = $plusflag ? 'utr-3' : 'utr-5';
          # SNP position is in the coding exon sequence
          } else {
            $snptype  = 'exon';
            $cds_exno = $cds_excount + 1;
          }
        }
      }

      # Calculate distances to the exon/intron junctions nearest to SNP position.
      # This only makes sense if the transcript has at least one intron.
      if ($knownGene->{exonCount} > 1) {
        if ($plusflag) {
          unless ($start == $knownGene->{txStart}) {  # skip start of first exon if at the transcript start -- no junction
            my $dstart = $start - $snpend;
            if (defined $jdist_acceptor) {
              $jdist_acceptor = $dstart if abs($dstart) < abs($jdist_acceptor);
            } else {
              $jdist_acceptor = $dstart;
            }
          }
          unless ($end == $knownGene->{txEnd}) {      # skip end of last exon if at the transcript end -- no junction
            my $dend   = $end - $snpstart;
            if (defined $jdist_donor) {
              $jdist_donor = $dend if abs($dend) < abs($jdist_donor);
            } else {
              $jdist_donor = $dend;
            }
          }
        # Exon start and exon end are reversed for a transcript on minus strand
        } else {
          unless ($start == $knownGene->{txStart}) {
            my $dend = $snpend - $start;
            if (defined $jdist_donor) {
              $jdist_donor = $dend if abs($dend) < abs($jdist_donor);
            } else {
              $jdist_donor = $dend;
            }
          }
          unless ($end == $knownGene->{txEnd}) {
            my $dstart   = $snpstart - $end;
            if (defined $jdist_acceptor) {
              $jdist_acceptor = $dstart if abs($dstart) < abs($jdist_acceptor);
            } else {
              $jdist_acceptor = $dstart;
            }
          }
        }
      }

      # Check that at least part of the exon falls within the CDS
      if ($end - 1 < $knownGene->{cdsStart} || $start > $knownGene->{cdsEnd} - 1) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Exon ($excount) is outside CDS, skipped\n" if $VERBOSE > 1;
        next EXON;
      }

      $cds_firstcodon = $i unless defined $cds_firstcodon;
      $cds_lastcodon  = $i;
      $cds_excount++;

      if ($plusflag) {
        @chr2cds{$cds_exstart .. ($cds_exend - 1)} = $cds_length .. ($cds_length + $cds_exend - $cds_exstart - 1);
      } else {
        @chr2cds{reverse($cds_exstart .. ($cds_exend - 1))} = $cds_length .. ($cds_length + $cds_exend - $cds_exstart - 1);
      }
      $cds_length += $cds_exend - $cds_exstart;

      # Extract the exon sequence unless coding sequence for this gene was already cached
      unless (exists $genes{$knownGene->{name}}{cds_nucseq}) {
        my $nibcommand =
            "nibFrag -upper -name= $CHRPATH/$chrom.nib $cds_exstart $cds_exend $nbstrand stdout";
        my $nibfa = `$nibcommand`;
        unless (defined $nibfa) {
          warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Failed to extract exon ($i), CDS skipped\n";
          next GENE;
        }
        $nibfa =~ s/^>.*?\n//s;
        $nibfa =~ s/\n//sg;
        unless (length $nibfa) {
          warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Exon ($i) sequence is empty, CDS skipped\n";
          next GENE;
        }
        if ($nibfa =~ /[^ACTG]/) {
          warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal sequence character(s) in exon ($i), CDS skipped\n";
          next GENE;
        }
        $cds_nucseq .= $nibfa;
      }

    } # EXON

    $jdist_donor = sprintf('%+d', $jdist_donor) if defined $jdist_donor;
    $jdist_acceptor = sprintf('%+d', $jdist_acceptor) if defined $jdist_acceptor;

    $exno     = '?' unless defined $exno;
    $cds_exno = '?' unless defined $cds_exno;

    $snptype = 'intron' unless $snptype;

    if (exists $genes{$knownGene->{name}}{cds_nucseq}) {
      $cds_nucseq = $genes{$knownGene->{name}}{cds_nucseq};
    } else {
      $genes{$knownGene->{name}}{cds_nucseq} = $cds_nucseq;
    }

    if (length($cds_nucseq) % 3) {
      warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Incomplete CDS, skipped\n";
      next GENE;
    }

    # Translate CDS (unless cached)
    my $cds_proseq = '';
    if (exists $genes{$knownGene->{name}}{cds_proseq}) {
      $cds_proseq = $genes{$knownGene->{name}}{cds_proseq};
    } else {
      for (my $i=0; $i<int(length($cds_nucseq) / 3); $i++) {
        my $cdn = substr($cds_nucseq, $i * 3, 3);
        my $res = $TRANS{$cdn} if exists $TRANS{$cdn};
        unless (defined $res) {
          warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal codon $cdn, CDS skipped\n";
          next GENE;
        }
        $cds_proseq .= $res;
      }
      $genes{$knownGene->{name}}{cds_proseq} = $cds_proseq;
    }

    # Obtain cross-reference mappings for transcript (unless already cached)
    unless (exists $genes{$knownGene->{name}}{xref}) {
      # Try to get mappings from kgXref annotations
      $sthXref->execute($knownGene->{name}) or die $sthXref->errstr;
      my $n = 0;
      while (my $xref = $sthXref->fetchrow_hashref()) {
        unless (exists $genes{$knownGene->{name}}{xref}) {
          # Occasionally, UniProt mappings are missing from kgXref but refSeq ID
          # may still be present; attempt to use UniProt idmapping in such case
          unless ($xref->{spID}) {
            my $xdbid = PPH::Seq::xdb2acc($xref->{protAcc}) if $xref->{protAcc};
            $xref->{spID} = $xdbid if $xdbid;
          }
          $genes{$knownGene->{name}}{xref} = $xref;
        }
        $n++;
      }
      warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Multiple xref matches ($n) skipped for transcript ($knownGene->{transcript})\n" if $n > 1;
    }

    # Annotate all coding SNP alleles
    if ($snptype eq 'exon') {
      unless (exists $chr2cds{$snpstart} && $chr2cds{$snpstart} >= 0) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal exonic SNP position ($snpend), CDS skipped\n";
        next GENE;
      }
      my $pro_codon_pos   = int($chr2cds{$snpstart} / 3);
      my $nuc_codon_pos   = $pro_codon_pos * 3;
      my $nuc_codon_frame = $chr2cds{$snpstart} % 3;
      my $trn_codon       = substr($cds_nucseq, $nuc_codon_pos, 3);           # transcript (reference) codon
      my $trn_aa          = $TRANS{$trn_codon} if exists $TRANS{$trn_codon};  # transcript (reference) AA residue
      unless (defined $trn_aa) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Illegal codon ($trn_codon) at CDS position ($nuc_codon_pos), CDS skipped\n";
        next GENE;
      }
      my $aa1 = substr($cds_proseq, $pro_codon_pos, 1); # aa1 should be the same as trn_aa, the following test is redundant
      unless ($trn_aa eq $aa1) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Wrong aa residue ($trn_aa != $aa1) at CDS position ($nuc_codon_pos), CDS skipped\n";
        next GENE;
      }
      unless ($trn_nt eq substr($trn_codon, $nuc_codon_frame, 1)) {
        warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Mismatched nt ($trn_nt) at codon ($trn_codon), frame ($nuc_codon_frame), CDS skipped\n";
        next GENE;
      }
      # Check if the mutated codon is split by an exon/intron junction boundary
      #        d   a        (d - donor site; a - acceptor site)
      #   =====|---|ACG=     d << 0 / f(0),a(-1) f(1),a(-2) f(2),a(-3)
      #   ====A|---|CG==     f(0),d(+1) / f(1),a(-1) f(2),a(-2)
      #   ===AC|---|G===     f(0),d(+2) f(1),d(+1) / f(2),a(-1)
      #   ==ACG|---|====     f(0),d(+3) f(1)d(+2) f(2),d(+3) / a >> 0
      my $codon_split;
      if (defined $jdist_donor && defined $jdist_acceptor) {
        if      ($nuc_codon_frame == 0) {
          $codon_split = 1 if $jdist_donor == 1 || $jdist_donor == 2;
        } elsif ($nuc_codon_frame == 1) {
          $codon_split = 1 if $jdist_donor == 1 || $jdist_acceptor == -1;
        } elsif ($nuc_codon_frame == 2) {
          $codon_split = 1 if $jdist_acceptor == -1 || $jdist_acceptor == -2;
        }
      }
      # Obtain transcript->protein mapping if one of -u or -U options was
      # specified and this mapping was not already cached
      if ($map2uniprot && !exists $genes{$knownGene->{name}}{mapped}) {
        # Attempt to map protein sequence of this CDS to a local version of UniProtKB
        my $seq  = $cds_proseq; $seq =~ s/\*$//;
        my $qlen = length $seq;   # query sequence length
        my ($macc, $mid, $mindex);
        my $mident = 0;
        # Try fgrep (-u) mapping first
        my ($mdesc, $mseq);
        ($macc, $mid, $mdesc, $mseq) = PPH::Seq::grep_in_FASTA($seq, $fasta_uniprot);
        ($macc, $mid, $mdesc, $mseq) = PPH::Seq::grep_in_FASTA($seq, $fasta_varsplic) unless defined $macc;
        if (defined $macc) {
          $mindex = $qlen == length($mseq) ? 0 : index($mseq, $seq);
          $mident = 1;
        }
        # Next try BLAST (-U) mapping if fgrep failed
        if ($map2uniprot > 1 && !$mident) {
          # sacc sname pident length mismatch gaps qstart qend sstart send
          my ($sacc, $sname, $pident, $alen, $mismatch, $gaps, $qstart, $qend, $sstart, $send) = map_by_BLAST($seq, $CONFIG{'UNIPROT'});
          # Found full length HSP match for query sequence with at least 98% identity and no gaps
          # Only full query length matches are accepted, partial overlaps with target are rejected
          if (defined $sacc && $alen == $qlen && $pident >= 98.0 && $gaps == 0) {
            $macc   = $sacc;
            $mid    = $sname;
            $mident = sprintf("%.3f", $pident / 100); # from percent to fraction
            $mindex = $sstart - 1;                    # zero-based index, 0 if query matches at the target start
            if ($mident < 1.0) {
              warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Mapped UCSC protein ($genes{$knownGene->{name}}{xref}{spID}) to UniProtKB ($macc) with ident=$mident, index=$mindex\n";
            }
          }
        }
        # Successfully mapped
        if ($mident) {
          # Note that it is likely that the RefSeq ID would (still) be incorrectly mapped
          $genes{$knownGene->{name}}{xref}{spID}         = $macc;
          $genes{$knownGene->{name}}{xref}{spDisplayID}  = $mid;
          $genes{$knownGene->{name}}{xref}{posIndex}     = $mindex;
        }
        # Ident > 0 if successfully mapped, zero otherwise
        $genes{$knownGene->{name}}{mapped} = $mident;
      }
      # Correct protein substitution position in case protein sequence has been re-mapped
      $pro_codon_pos += $genes{$knownGene->{name}}{xref}{posIndex} if $genes{$knownGene->{name}}{mapped};

      ALLELE: for (my $i=0; $i<@alleles; $i++) {
        my $allele = $alleles[$i];
        next ALLELE if $allele eq $trn_nt;
        my $o_allele = $o_alleles[$i];
        my $codon2 = $trn_codon; substr($codon2, $nuc_codon_frame, 1, $allele);
        my $aa2 = $TRANS{$codon2} if exists $TRANS{$codon2};
        if      ($aa2 eq $aa1) {
          $snptype = 'coding-synon';
        } elsif ($aa2 eq '*' || $aa1 eq '*') {
          $snptype = 'nonsense';
        } else {
          $snptype = 'missense';
        }
        $totals{$snptype}++; $totals{'annotated'}++;
        next ALLELE if $missense_only && $snptype ne 'missense';
        my $flanks = ($trn_nt5 || '?') . ($trn_nt3 || '?');
        my $cpg    = CpG($trn_nt5, $trn_nt, $allele, $trn_nt3);
        my $dbsnp  = map2dbSNP($chrom, $snpstart, $snpend, $trn_nt, $allele, $knownGene);
        # Save CDS protein sequence if -s option is specified
        if ($seqfile && !$genes{$knownGene->{name}}{saved} && $snptype eq 'missense') {
          # If mapping to UniProt was enabled then save the protein
          # sequence only if mapping attempt did not succeed
          unless ($map2uniprot && $genes{$knownGene->{name}}{mapped}) {
            my $seq = $cds_proseq; $seq =~ s/\*$//;
            my $defline = '>lcl|' . $knownGene->{name} . ' Len=' . length($seq);
            $defline .= ' ccdsId=' . $knownGene->{ccdsId}
              if exists $knownGene->{ccdsId} && defined $knownGene->{ccdsId} && length $knownGene->{ccdsId};
            $defline .= ' spID=' . $genes{$knownGene->{name}}{xref}{spID}
              if exists $genes{$knownGene->{name}}{xref}{spID} && defined $genes{$knownGene->{name}}{xref}{spID} &&
                length $genes{$knownGene->{name}}{xref}{spID};
            $defline .= ' spDisplayID=' . $genes{$knownGene->{name}}{xref}{spDisplayID}
              if exists $genes{$knownGene->{name}}{xref}{spDisplayID} && defined $genes{$knownGene->{name}}{xref}{spDisplayID} &&
                length $genes{$knownGene->{name}}{xref}{spDisplayID};
            $defline .= ' protAcc=' . $genes{$knownGene->{name}}{xref}{protAcc}
              if exists $genes{$knownGene->{name}}{xref}{protAcc} && defined $genes{$knownGene->{name}}{xref}{protAcc} &&
                length $genes{$knownGene->{name}}{xref}{protAcc};
            $defline .= ' refSeq=' . $genes{$knownGene->{name}}{xref}{refseq}
              if exists $genes{$knownGene->{name}}{xref}{refseq} && defined $genes{$knownGene->{name}}{xref}{refseq} &&
                length $genes{$knownGene->{name}}{xref}{refseq};
            $defline .= ' geneSymbol=' . $genes{$knownGene->{name}}{xref}{geneSymbol}
              if exists $genes{$knownGene->{name}}{xref}{geneSymbol} && defined $genes{$knownGene->{name}}{xref}{geneSymbol} &&
                length $genes{$knownGene->{name}}{xref}{geneSymbol};
            print SEQFILE $defline, "\n", $seq, "\n";
          }
          $genes{$knownGene->{name}}{saved} = 1;
        }
        # Print out full allele annotation
        my @values = map { (defined) ? $_ : '' } (
          "$chrom:$snpend",
          $knownGene->{strand},
          $genes{$knownGene->{name}}{xref}{geneSymbol},
          $knownGene->{name},
          $knownGene->{clusterId},
          $knownGene->{ccdsId},
          $knownGene->{cdsSimilarity},
          "$ref_nt/$o_allele",
          $snptype,
          $trn_nt,
          $allele,
          $flanks,
          defined $nucseq_offset ? $nucseq_offset + 1 : '',
          $TRNV{$trn_nt}{$allele},
          $cpg,
          $jdist_donor,
          $jdist_acceptor,
          "$exno/$excount",
          "$cds_exno/$cds_excount",
          $codon_split,
          $DGEN{$trn_codon}[$nuc_codon_frame],
          $nuc_codon_frame,
          $trn_codon,
          $codon2,
          $aa1,
          $aa2,
          $pro_codon_pos + 1,
          $map2uniprot ? $genes{$knownGene->{name}}{mapped} : undef,
          $genes{$knownGene->{name}}{xref}{spID},
          $genes{$knownGene->{name}}{xref}{spDisplayID},
          $genes{$knownGene->{name}}{xref}{protAcc},
          $dbsnp->{name},
          $dbsnp->{observed},
          $dbsnp->{avHet},
          $dbsnp->{avHetSE},
          $dbsnp->{RmPaPt}
        );
        unless ($header_printed || $stripes && $stripe_no > 1) {
          print $headerstr; $header_printed = 1;
        }
        printf $formatstr, @values;
        print "\t$comments" if defined $comments && length $comments;
        print "\n";
        # Now write query string for missense SNPs in PolyPhen-2
        # input format to a filename specified via '-y' option.
        if ($pph2output && $snptype eq 'missense') {
          my $spID = $genes{$knownGene->{name}}{xref}{spID} if
            exists $genes{$knownGene->{name}}{xref}{spID} &&
            defined $genes{$knownGene->{name}}{xref}{spID} &&
            length $genes{$knownGene->{name}}{xref}{spID};
          # Mapping to UniProtKB was attempted...
          if ($map2uniprot) {
            # ...and failed, skip unless '-s' option was also specified,
            # in which case substitute UCSC transcript name for spID
            unless ($genes{$knownGene->{name}}{mapped}) {
              if ($seqfile) { $spID = $knownGene->{name}; }
              else { $spID = ''; }
            }
          # No UniProtKB mapping, use original kgXref accession if available...
          } else {
            # ...otherwise skip unless '-s' option was also specified,
            # in which case substitute UCSC transcript name for spID
            unless ($spID) {
              if ($seqfile) { $spID = $knownGene->{name}; }
              else { $spID = ''; }
            }
          }
          if ($spID) {
            # Remove '/' character from original (input) allele spec since snp_id is
            # used in XML file names and slash is not allowed in filenames
            $o_allele =~ tr|/||d;
            my @values = map { (defined) ? $_ : '' } (
              "$chrom:$snpend.$ref_nt$o_allele.$knownGene->{name}",
              $spID,
              $pro_codon_pos + 1,
              $aa1,
              $aa2
            );
            unless ($pph2header_printed || $stripes && $stripe_no > 1) {
              print PPH2OUT $pph2headerstr; $pph2header_printed = 1;
            }
            printf PPH2OUT $pph2formatstr, @values;
            if (defined $comments && length $comments) {
              # PolyPhen-2 input format requires comments to be additionally delimited with the '#' character
              $comments = "# $comments";
              print PPH2OUT "\t$comments";
            }
            print PPH2OUT "\n";
          }
        }
      } # ALLELE
    # Short annotation for non-coding SNP alleles
    } else {
      ALLELE: for (my $i=0; $i<@alleles; $i++) {
        my $allele = $alleles[$i];
        next ALLELE if $allele eq $trn_nt;
        $totals{$snptype}++; $totals{'annotated'}++;
        # Skip reporting all non-coding alleles if '-c' option was specified
        # but still count totals for them properly in the above line.
        next ALLELE if $coding_only;
        my $o_allele = $o_alleles[$i];
        my $dbsnp = map2dbSNP($chrom, $snpstart, $snpend, $trn_nt, $allele, $knownGene);
        my $cpg   = CpG($trn_nt5, $trn_nt, $allele, $trn_nt3);
        my $flanks = ($trn_nt5 || '?') . ($trn_nt3 || '?');
        my @values = map { (defined) ? $_ : '' } (
          "$chrom:$snpend",
          $knownGene->{strand},
          $genes{$knownGene->{name}}{xref}{geneSymbol},
          $knownGene->{name},
          $knownGene->{clusterId},
          $knownGene->{ccdsId},
          $knownGene->{cdsSimilarity},
          "$ref_nt/$o_allele",
          $snptype,
          $trn_nt,
          $allele,
          $flanks,
          defined $nucseq_offset ? $nucseq_offset + 1 : '',
          $TRNV{$trn_nt}{$allele},
          $cpg,
          $jdist_donor,
          $jdist_acceptor,
          (undef) x 14,
          $dbsnp->{name},
          $dbsnp->{observed},
          $dbsnp->{avHet},
          $dbsnp->{avHetSE},
          $dbsnp->{RmPaPt}
        );
        unless ($header_printed || $stripes && $stripe_no > 1) {
          print $headerstr; $header_printed = 1;
        }
        printf $formatstr, @values;
        print "\t$comments" if defined $comments && length $comments;
        print "\n";
      } # ALLELE
    }

  } # GENE

} # LINE
close(FIN);

$sthdbSNP->finish();
$sthXref->finish();
$sthKnownGene->finish();
$dbh->disconnect();

close(SEQFILE) if $seqfile;
close(PPH2OUT) if $pph2output;

# Print out totals
my $skipped = ($totals{'input'} || 0) - ($totals{'processed'} || 0);
print "## Totals:\n";
printf "##   %-18s %8s\n", 'lines input', $totals{'input'} || 0;
printf "##   %-18s %8s\n", 'lines skipped', $skipped;
printf "##   %-18s %8s\n", 'alleles annotated', $totals{'annotated'} || 0;
delete $totals{'input'}; delete $totals{'processed'}; delete $totals{'annotated'};
for my $key (qw{ missense nonsense coding-synon intron utr-3 utr-5 }) {
  printf "##     %-16s %8s\n", $key, $totals{$key} || 0;
  delete $totals{$key};
}
# In case something non-standard was still left
for my $key (sort keys %totals) {
  printf "##     %-16s %8s\n", $key, $totals{$key} || 0;
}

warn "Finished.\n" if $VERBOSE;

# Note: snp130OrthoPt2Pa2Rm2.humanAllele contains a *reference* (plus strand) allele,
#       unlike the corresponding fields for all other species, for which alleles are
#       always relative to the human SNP strand. I assume this is a bug in UCSC data.
#       On the contrast, snp130OrthoPt2Pa2Rm2.humanObserved as well as snp130.observed
#       are listed relative to the current SNP strand, just like all other alleles.
sub map2dbSNP {
  my ($chrom, $snpstart, $snpend, $trn_nt, $allele, $knownGene) = @_;
  my $bin = binFromRangeStandard($snpstart, $snpend);
  $sthdbSNP->execute($chrom, $bin, $snpstart, $snpend) or die $sthdbSNP->errstr;
  my $firstdbsnp;
  my @extrasnps;
  while (my $dbsnp = $sthdbSNP->fetchrow_hashref()) {
    # Complement observed alleles if SNP orientation does not match gene orientation
    $dbsnp->{observed} =~ tr/ACGTacgt/TGCAtgca/ if $knownGene->{strand} ne $dbsnp->{strand};
    # Both reference and variant allele should match (in any order)
    unless ($dbsnp->{observed} =~ /$trn_nt/ && $dbsnp->{observed} =~ /$allele/) {
      warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Allele ($trn_nt/$allele) does not match dbSNP ($dbsnp->{name}) observed ($dbsnp->{observed})\n" if $VERBOSE > 1;
      next;
    }
    # Skip all but first match
    if (defined $firstdbsnp) {
      push @extrasnps, $dbsnp->{name};
    } else {
      $firstdbsnp = $dbsnp;
    }
  }
  if (@extrasnps) {
    warn "WARNING: ($chrom:$snpend $knownGene->{strand} $knownGene->{name}) Allele ($trn_nt/$allele) matches multiple dbSNPs: $firstdbsnp->{name}, " .
      join(', ', @extrasnps) . "\n";
  }
  if (defined $firstdbsnp) {
    $firstdbsnp->{avHet}   = (exists $firstdbsnp->{avHet} && defined $firstdbsnp->{avHet}) ?
                              sprintf('%g', $firstdbsnp->{avHet}) : '';
    $firstdbsnp->{avHetSE} = (exists $firstdbsnp->{avHetSE} && defined $firstdbsnp->{avHetSE}) ?
                              sprintf('%g', $firstdbsnp->{avHetSE}) : '';
    # If JOIN with the snp130OrthoPt2Pa2Rm2 table was successful then all alleles
    # are defined, even though any number of them could be ?, or unknown
    if (defined $firstdbsnp->{macaqueAllele}) {
      $firstdbsnp->{RmPaPt} =
        "$firstdbsnp->{macaqueAllele}>$firstdbsnp->{orangAllele}>$firstdbsnp->{chimpAllele}";
      # Complement all orthologs alleles if SNP orientation does not match gene orientation
      $firstdbsnp->{RmPaPt} =~ tr/ACGTacgt/TGCAtgca/ if $firstdbsnp->{strand} ne $knownGene->{strand};
    }
  }
  return $firstdbsnp;
}

# From (Jim Kent's):
#   blatSrc/lib/binRange.c
sub binFromRangeStandard {
  my ($start, $end) = @_;
  my $startBin = $start;
  my $endBin   = $end - 1;
  $startBin  >>= $binFirstShift;
  $endBin    >>= $binFirstShift;
  for (my $i=0; $i<@binOffsets; ++$i) {
    if ($startBin == $endBin) { return $binOffsets[$i] + $startBin; }
    $startBin >>= $binNextShift;
    $endBin   >>= $binNextShift;
  }
  warn sprintf("WARNING (binRange): start %d, end %d out of range (max is 512M)\n", $start, $end);
  return 0;
}

# Calculate [start,end] range of a stripe in dataset
# Note: all indices passed and returned are one-based
#
# Arguments:  stripe index, number of stripes, dataset size
# Returns:    data start index, data end index; both indices
#             returned are zero if no more data fall within
#             a stripe specified (e.g. number of stripes is
#             larger than the number of data items)
#
sub stripe_range {
  my ($stripe_no, $stripes, $size) = @_;

  die "ERROR: Stripe index too large: $stripe_no > $stripes\n" if $stripe_no > $stripes;

  my $p = int($size / $stripes);
  my $r = $size % $stripes;

  my $shift_start = ($stripe_no <= $r ? $stripe_no : $r + 1) - 1;
  my $shift_end   = ($stripe_no <= $r ? $stripe_no : $r);

  my $start = ($stripe_no - 1)*$p + 1 + $shift_start;
  my $end   = $stripe_no*$p + $shift_end;

  if ($start > $end) {
    return (0, 0);
  } else {
    return ($start, $end);
  }
}

sub CpG {
  my ($nt5, $allele1, $allele2, $nt3) = @_;
  my $t1  = $nt5 . $allele1 . $nt3;
  my $t2  = $nt5 . $allele2 . $nt3;
  return undef unless length($t1) == 3 && length($t2) == 3;
  my $cpg = 0;
  $cpg    = 1 if index($t1, 'CG') >= 0;
  $cpg   += 2 if index($t2, 'CG') >= 0;
  return $cpg;
}

sub map_by_BLAST {
  my ($query, $db) = @_;
  my $blast = File::Temp->new();
  my $cmd = qq($CONFIG{BLAST} -db $db -max_target_seqs 1 -ungapped -comp_based_stats F -seg no -evalue 0.01 -outfmt '10 sseqid pident length mismatch gaps qstart qend sstart send' -out $blast >/dev/null 2>&1);
  open(FCMD, '|-', $cmd) || die "ERROR: $CONFIG{BLAST}: pipe failed ($!)\n";
  print FCMD $query;
  close(FCMD) or die "ERROR: $CONFIG{BLAST}: broken pipe ($?)\n";
  my $rc = <$blast>;
  if ($rc) {
    chomp($rc);
    # sseqid pident length mismatch gaps qstart qend sstart send
    my @a = split /,/, $rc, 9;
    my @b = (split /\|/, $a[0], 3)[1..2];
    # sacc sname pident length mismatch gaps qstart qend sstart send
    return (@b, @a[1..$#a]);
  } else {
    return undef;
  }
}
