#!/usr/bin/env perl

use warnings;
use strict;

=head1 NAME

merge_xml.pl - merge results of PolyPhen-2 classifier run into corresponding
               SNP features file(s) in XML format

=head1 SYNOPSIS

merge_xml.pl snp.xml modelfile prediction pph2_class pph2_prob pph2_FPR pph2_TPR pph2_FDR

--or--

merge_xml.pl -f pph2summary [-m model] snp1.xml [snp2.xml]...

where mandatory argument is:

  -f pph2summary   PolyPhen-2 SNP features tabulated summary file

and options are:

  -m modelfile     WEKA model file used by classifier; default
                   is $PPH/models/HumDiv.UniRef100.NBd.f11.model

=head1 SUBVERSION

 $LastChangedDate: 2009-11-28 00:15:09 -0500 (Sat, 28 Nov 2009) $
 $LastChangedRevision: 254 $
 $LastChangedBy: ivan $

=cut

use Getopt::Std;
use Pod::Usage;
use File::Path;

BEGIN {
  use FindBin qw($RealBin);
  BEGIN {
    unless ($ENV{'PPH'}) {
      my $binpath = $RealBin;
      $binpath =~ s|/[^/]+$||;
      $ENV{'PPH'} = $binpath;
    }
  }
  use lib "$ENV{PPH}/perl";
}

use PPH::Config;
use PPH;
use SNP;

my %opts;
pod2usage('No arguments specified!') unless @ARGV;
getopts('f:m:', \%opts) or pod2usage('Wrong arguments!');
unless ($opts{'f'}) {
  pod2usage('Wrong number of arguments!') unless @ARGV == 8;
  pod2usage('Options not allowed in single SNP mode!') if %opts;
}

# Read classifier data from a tabulated summary file
if ($opts{'f'}) {
  my $modf = $opts{'m'} || $CONFIG{'WEKAMODEL'}; # WEKA model filename (not checked!)
  my $sumf = $opts{'f'};
  die "File does not exist: $sumf\n" unless -e $sumf;
  die "File empty: $sumf\n" unless -s $sumf;
  my %summary;
  open(SUMF, $sumf) or die "Can't open file: $sumf\n";
  while (<SUMF>) {
    # Skip comments
    next if /^#/;
    # Skip lines with errors & warnings and empty lines
    next if /\tError:\s/;
    next if /^(ERROR|WARNING):\s/;
    next if /^\s*$/;
    chomp;
    my @a = split /\t/;
    map { if (defined) { s/^\s+//; s/\s+$//; } } @a;
    # Verify data
    warn("Invalid input: $_\n"), next
      # column 13: prediction
      # columns 16-20: pph2_class pph2_prob pph2_FPR pph2_TPR pph2_FDR
      unless
        # valid prediction
        (defined $a[15] && ($a[15] eq 'deleterious' || $a[15] eq 'neutral') &&
         defined $a[16] &&  $a[16] =~ /^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $a[16] >= 0 && $a[16] <= 1 &&
         defined $a[17] &&  $a[17] =~ /^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $a[17] >= 0 && $a[17] <= 1 &&
         defined $a[18] &&  $a[18] =~ /^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $a[18] >= 0 && $a[18] <= 1 &&
         defined $a[19] &&  $a[19] =~ /^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $a[19] >= 0 && $a[19] <= 1) ||
        # dummy but properly formatted prediction (aka "unknown")
        (defined $a[12] && $a[12] eq 'unknown' && defined $a[15] && $a[15] eq 'none')
      ;
    push @{ $summary{$a[0]} }, @a[12, 15..19];
  }
  close(SUMF);
  my @xmlfiles = @ARGV ? @ARGV : glob('*.xml');
  die "WARNING: No XML files found\n" unless @xmlfiles;
  foreach my $xmlf (@xmlfiles) {
    warn("File does not exist: $xmlf\n"), next unless -e $xmlf;
    warn("File empty: $xmlf\n"), next unless -s $xmlf;
    my $SNP = SNP->new_empty();
    $SNP->xml2snp($xmlf);
    my $snpid = exists $SNP->{Orig_SNP_Id} ? $SNP->{Orig_SNP_Id} : $SNP->{SNP_Id};
    if (exists $summary{$snpid}) {
      my @a = @{ $summary{$snpid} };
      push @{ $SNP->{WEKA} }, {
        Prediction  => $a[0],
        Modelfile   => $modf,
        Class       => $a[1],
        Score       => $a[2],
        FPR         => $a[3],
        TPR         => $a[4],
        FDR         => $a[5]
      };
      $SNP->snp2xml($xmlf);
    } else {
      warn "Unmatched SNP_Id ($snpid): $xmlf\n";
    }
  }
# Single SNP command-line arguments mode
} else {
  my $xmlf = shift; # PPH2 original XML dump file
  my $modf = shift; # WEKA model filename (not checked!)
  my ($prediction, $class, $prob, $fpr, $tpr, $fdr) = @ARGV;

  die "File does not exist: $xmlf\n" unless -e $xmlf;
  die "File empty: $xmlf\n" unless -s $xmlf;

  my $SNP = SNP->new_empty();
  $SNP->xml2snp($xmlf);
  # No data verification!
  push @{ $SNP->{WEKA} }, { Modelfile=>$modf, Prediction=>$prediction, Class=>$class, Score=>$prob, FPR=>$fpr, TPR=>$tpr, FDR=>$fdr };
  $SNP->snp2xml($xmlf);
}
