#!/usr/bin/env perl

use Getopt::Std;
use File::Basename;

# PPHv2 features (50 total) vs. WEKA features
# (Note: PPHv2 input Pos is zero-based, WEKA output Pos are 1-based)
#
#       PPHv2                   WEKA
#  --------------- -----------------------------------
#  Pos  Name        Name              Pos33 Pos11 Pos8
#  --------------- -----------------------------------
#    0  o_snp_id
#    1  o_acc
#    2  o_pos
#    3  o_aa1
#    4  o_aa2
#    5  snp_id
#    6  acc
#    7  pos
#    8  aa1
#    9  aa2
#   10  nt1         nt1                   1
#   11  nt2         nt2                   2
#   12  prediction
#   13  based on
#   14  effect
#   15  site        site                  3
#   16  region      region                4
#   17  PHAT        phat                  5
#   18  dScore      score_delta           6     1    1
#   19  Score1      score1                7     2    2
#   20  Score2      score2                8
#   21  Nobs        num_obs               9     3    3
#   22  Nstruct
#   23  Nfilt
#   24  PDB_id
#   25  PDB_pos
#   26  PDB_ch
#   27  ident       ali_ide              21
#   28  length      ali_len              22
#   29  NormAcc     acc_normed           23     9
#   30  SecStr      sec_str              24
#   31  MapReg      map_region           25
#   32  dVol        delta_volume         10     4
#   33  dProp       delta_prop           26
#   34  B-fact      b_fact               27    10
#   35  H-bonds
#   36  AveNHet     het_cont_ave_num     28
#   37  MinDHet     het_cont_min_dist    29
#   38  AveNInt     inter_cont_ave_num   30
#   39  MinDInt     inter_cont_min_dist  31
#   40  AveNSit
#   41  MinDSit
#   42  Transv      transversion         11
#   43  CodPos
#   44  CpG         cpg                  12
#   45  MinDJnc
#   46  PfamHit     pfam_hit             13     5    4
#   47  IdPmax      id_p_max             14     6    5
#   48  IdPSNP      id_p_snp             15
#   49  IdQmin      id_q_min             16     7    6
#   50  Comment
#
# The following 6 extra features have no direct equivalents in PPH2 output:
#
#   51  cpg_var1_var2                    17
#   52  cpg_transition                   18     8    7
#   53  charge_change                    19
#   54  hydroph_change                   20
#   55  delta_volume_new                 32
#   56  delta_prop_new                   33    11    8

# Same mappings as above encoded as array of arrays
my @PPH2WEKA = (
  [ 'o_snp_id',   '',                     '',           0,  0,  0 ],
  [ 'o_acc',      '',                     '',           0,  0,  0 ],
  [ 'o_pos',      '',                     '',           0,  0,  0 ],
  [ 'o_aa1',      '',                     '',           0,  0,  0 ],
  [ 'o_aa2',      '',                     '',           0,  0,  0 ],
  [ 'snp_id',     '',                     '',           0,  0,  0 ],
  [ 'acc',        '',                     '',           0,  0,  0 ],
  [ 'pos',        '',                     '',           0,  0,  0 ],
  [ 'aa1',        '',                     '',           0,  0,  0 ],
  [ 'aa2',        '',                     '',           0,  0,  0 ],
  [ 'nt1',        'nt1',                  '',           1,  0,  0 ],
  [ 'nt2',        'nt2',                  '',           2,  0,  0 ],
  [ 'prediction', '',                     '',           0,  0,  0 ],
  [ 'based on',   '',                     '',           0,  0,  0 ],
  [ 'effect',     '',                     '',           0,  0,  0 ],
  [ 'site',       'site',                 'yesno',      3,  0,  0 ],
  [ 'region',     'region',               'missno',     4,  0,  0 ],
  [ 'PHAT',       'phat',                 'numeric',    5,  0,  0 ],
  [ 'dScore',     'score_delta',          'numeric',    6,  1,  1 ],
  [ 'Score1',     'score1',               'numeric',    7,  2,  2 ],
  [ 'Score2',     'score2',               'numeric',    8,  0,  0 ],
  [ 'Nobs',       'num_obs',              'numeric',    9,  3,  3 ],
  [ 'Nstruct',    '',                     '',           0,  0,  0 ],
  [ 'Nfilt',      '',                     '',           0,  0,  0 ],
  [ 'PDB_id',     '',                     '',           0,  0,  0 ],
  [ 'PDB_pos',    '',                     '',           0,  0,  0 ],
  [ 'PDB_ch',     '',                     '',           0,  0,  0 ],
  [ 'ident',      'ali_ide',              'misszero',  21,  0,  0 ],
  [ 'length',     'ali_len',              'misszero',  22,  0,  0 ],
  [ 'NormAcc',    'acc_normed',           'numeric',   23,  9,  0 ],
  [ 'SecStr',     'sec_str',              '',          24,  0,  0 ],
  [ 'MapReg',     'map_region',           '',          25,  0,  0 ],
  [ 'dVol',       'delta_volume',         'numeric',   10,  4,  0 ],
  [ 'dProp',      'delta_prop',           'numeric',   26,  0,  0 ],
  [ 'B-fact',     'b_fact',               'numeric',   27, 10,  0 ],
  [ 'H-bonds',    '',                     '',           0,  0,  0 ],
  [ 'AveNHet',    'het_cont_ave_num',     'numeric',   28,  0,  0 ],
  [ 'MinDHet',    'het_cont_min_dist',    'numeric',   29,  0,  0 ],
  [ 'AveNInt',    'inter_cont_ave_num',   'numeric',   30,  0,  0 ],
  [ 'MinDInt',    'inter_cont_min_dist',  'numeric',   31,  0,  0 ],
  [ 'AveNSit',    '',                     '',           0,  0,  0 ],
  [ 'MinDSit',    '',                     '',           0,  0,  0 ],
  [ 'Transv',     'transversion',         'yesno',     11,  0,  0 ],
  [ 'CodPos',     '',                     '',           0,  0,  0 ],
  [ 'CpG',        'cpg',                  'yesno',     12,  0,  0 ],
  [ 'MinDJnc',    '',                     '',           0,  0,  0 ],
  [ 'PfamHit',    'pfam_hit',             'yesno',     13,  5,  4 ],
  [ 'IdPmax',     'id_p_max',             'numeric',   14,  6,  5 ],
  [ 'IdPSNP',     'id_p_snp',             'yesno',     15,  0,  0 ],
  [ 'IdQmin',     'id_q_min',             'numeric',   16,  7,  6 ]
);

# Mappings for the additionally generated features
my @GENERATED =(
  # note: cpg_var1_var2() takes care of both cpg_var1_var2 and cpg_transition
  [ \&cpg_var1_var2,    'cpg_var1_var2',    'missno',    17,  0,  0 ],
  [ undef,              'cpg_transition',   'missno',    18,  8,  7 ],
  [ \&charge_change,    'charge_change',    'numeric',   19,  0,  0 ],
  [ \&hydroph_change,   'hydroph_change',   'numeric',   20,  0,  0 ],
  # note: delta_volume_new takes care of both delta_volume_new and delta_prop_new
  [ \&delta_volume_new, 'delta_volume_new', 'numeric',   32,  0,  0 ],
  [ undef,              'delta_prop_new',   'numeric',   33, 11,  8 ]
);

my %VOLUME = (
  'A' => 88, 'C' => 108, 'D' => 111, 'E' => 138, 'F' => 190,
  'G' => 60, 'H' => 153, 'I' => 167, 'K' => 168, 'L' => 167,
  'M' => 163,'N' => 114, 'P' => 112, 'Q' => 144, 'R' => 173,
  'S' => 89, 'T' => 116, 'V' => 140, 'W' => 227, 'Y' => 193
);

my %PROPENSITY = (
  A => [  0.59,   0.06,  -0.03,  -0.20,  -0.23,  -0.18,   0.18 ],
  R => [ -2.99,  -1.14,   0.02,   0.33,   0.47,   0.25,  -0.31 ],
  N => [ -1.48,  -0.55,  -0.27,   0.06,   0.26,   0.45,   0.39 ],
  D => [ -1.35,  -0.69,  -0.33,   0.10,   0.28,   0.42,   0.50 ],
  C => [  0.61,   0.52,   0.51,  -0.02,  -0.75,  -1.80,  -2.00 ],
  Q => [ -1.55,  -0.58,  -0.25,   0.06,   0.40,   0.44,  -0.17 ],
  E => [ -2.13,  -0.98,  -0.35,  -0.01,   0.40,   0.58,   0.28 ],
  G => [  0.31,  -0.23,  -0.07,  -0.21,  -0.12,   0.05,   0.68 ],
  H => [ -1.09,  -0.11,   0.25,   0.38,   0.10,  -0.18,  -0.54 ],
  I => [  0.87,   0.54,   0.09,  -0.04,  -0.79,  -1.29,  -2.00 ],
  L => [  0.59,   0.61,   0.28,  -0.11,  -0.76,  -1.21,  -1.42 ],
  K => [ -4.06,  -2.28,  -1.22,  -0.08,   0.60,   0.75,   0.39 ],
  M => [  0.53,   0.61,  -0.01,  -0.11,  -0.40,  -1.17,  -0.62 ],
  F => [  0.47,   0.64,   0.36,  -0.01,  -0.86,  -1.31,  -1.75 ],
  P => [ -0.98,  -0.46,  -0.28,  -0.03,   0.11,   0.41,   0.74 ],
  S => [ -0.29,  -0.21,   0.04,  -0.11,  -0.01,   0.21,   0.49 ],
  T => [ -0.49,  -0.16,  -0.11,   0.10,   0.23,   0.09,  -0.15 ],
  W => [ -0.29,   0.53,   0.63,   0.19,  -0.52,  -1.42,  -2.54 ],
  Y => [ -0.76,   0.27,   0.53,   0.43,  -0.22,  -0.95,  -1.87 ],
  V => [  0.69,   0.50,   0.19,  -0.13,  -0.46,  -1.13,  -1.76 ]
);

my %HYDROPHOBICITY = (
  A, 0.02, R, -0.42, N, -0.77, D, -1.04,  C, 0.77,  Q, -1.10,  E, -1.14, G, -0.80,  H, 0.26,  I, 1.81,
  L, 1.14, K, -0.41, M,  1.00, F,  1.35,  P, -0.09, S, -0.97,  T, -0.77, W,  1.71,  Y, 1.11,  V, 1.13,
);

my $ACC_THRESH = 0.05;

my %opt;
die "Illegal option\n" unless getopts('psuvc:f:', \%opt);
my $sumfile = shift;
die "PPH summary file not specified\n" unless defined $sumfile;
die "PPH summary file does not exist: $sumfile\n" unless -e $sumfile;
my $setidx = 3; # pointer to default feature set, full 33 features
if (defined $opt{'f'}) {
  if    ($opt{'f'} == 33) { $setidx = 3; }
  elsif ($opt{'f'} == 11) { $setidx = 4; }
  elsif ($opt{'f'} ==  8) { $setidx = 5; }
  else  { die("Unsupported feature set: ".$opt{'f'}."\n"); }
} else {
  $opt{'f'} = '33';
}
my $signed = 1; # defaults to signed deltas
$signed = 0 if $opt{'u'};
my $parse_comments = $opt{'p'} || 0;  # parse comments to extract sids and class
my $sids = $opt{'s'} || 0; # input format with set IDs at array pos -2
my $forced_class = $opt{'c'} || ''; # user-supplied override for the class value
my $verbose = $opt{'v'} || 0;

# Compile indexing array for mapping 54 input fields
# to a (variable) set of output features
my @subset;
foreach $rec (@PPH2WEKA) {
  my $idx = $rec->[$setidx];
  push @subset, ($idx ? $idx-1 : undef);
}
foreach $rec (@GENERATED) {
  my $idx = $rec->[$setidx];
  push @subset, ($idx ? $idx-1 : undef);
}

open(FH, $sumfile) or die "Can't open summary file: $sumfile\n";
# WEKA features, AoA
my @weka;
# Unique values for each WEKA feature, HoH
my $input_line_no = 0;
my %values;
while (<FH>) {
  $input_line_no++;
  # Skip comments
  next if /^#/;
  # Skip lines with errors & warnings and empty lines
  next if /\tError:\s/;
  next if /^(ERROR|WARNING):\s/;
  next if /^\s*$/;
  chomp;
  # Trim comments if present
  my $comments = $1 if s/\s *#\s*(.+)$//;
  my @pf = split /\t/;
  # 50 mandatory features
  die('Incorrect number of fields (' . scalar(@pf) . ") at line $input_line_no in summary file: $sumfile\n") unless @pf == 50;
  map { s/^\s+//; s/\s+$//; } @pf;
  # Skip lines with missing alignment data, i.e., the ones with 'unknown' prediction outcome
  next if $pf[12] eq 'unknown' && $pf[18] eq '';
  my ($set_id, $class);
  # Parse comments to extract sids and/or class columns
  if ($parse_comments && defined $comments) {
    my @c = split /\t/, $comments;
    map { s/^\s+//; s/\s+$//; } @c;
    $set_id = shift @c if @c > 1;
    $class  = shift @c;
  }
  my ($aa1, $aa2) = ($pf[8], $pf[9]);
  # Sanity check
  die "AA variant(s) missing or incorrect at line $input_line_no: $_\n" unless length($aa1)==1 && length($aa2)==1;

  # NormAcc indicates presence of 3D structure
  my $acc_normed = $pf[29];
  my $with3D = length $acc_normed ? 1 : 0;
  my @wf; # array of 56 fileds (some dummy)
  # Loop over all 50 PPHv2 output fields as defined in mappings array
  for ($i=0; $i<@PPH2WEKA; $i++) {

    my ($pname, $wname, $type) =
       ($PPH2WEKA[$i]->[0], $PPH2WEKA[$i]->[1], $PPH2WEKA[$i]->[2]);
    push(@wf, undef), next
      unless length $wname; # dummy values for fields not included into output
    # Sanity check
    die "Summary file parsing error at line $input_line_no: $_\n" unless length($pname) && length($wname);
    my $pf = $pf[$i];
    if (defined $pf && length $pf) {

      # Specific features:
      # note: will still be set to '?' if undefined/missing
      if      ($pname eq 'SecStr') {
        if    ($pf eq 'H' || $pf eq 'G')  { $pf = 'HELIX'; }
        elsif ($pf eq 'E')                { $pf = 'SHEET'; }
        else                              { $pf = 'OTHER'; }
      } elsif ($pname eq 'MapReg') {
        if    ($pf eq 'A' || $pf eq 'a')  { $pf = 'ALPHA'; }
        elsif ($pf eq 'B' || $pf eq 'b')  { $pf = 'BETA';  }
        else                              { $pf = 'OTHER'; }
      } elsif ($pname eq 'dVol') {
        # Need to recalculate since PPH dVol is abs() value
        die('Side chain volume undefined for residue: '.$aa1."\n") unless exists $VOLUME{$aa1};
        die('Side chain volume undefined for residue: '.$aa2."\n") unless exists $VOLUME{$aa2};
        $pf = $VOLUME{$aa1} - $VOLUME{$aa2};
        $pf = abs($pf) unless $signed;
      } elsif ($pname eq 'dProp') {
        if ($with3D && length $acc_normed) {
          my $j = &propens_interval($acc_normed);
          my $dp = $PROPENSITY{$aa1}->[$j] - $PROPENSITY{$aa2}->[$j];
          $pf = $signed ? $dp : abs($dp);
        } else {
          $pf = '?';
        }
      } elsif ($pname eq 'dScore') {
        # Recalculate if we are in signed mode
        $pf = sprintf('%.3f', $pf[19] - $pf[20]) if $signed;
      } elsif ($pname eq 'CpG') {
        if ($signed) {
          my $cpgmod = $pf % 2;
          $pf = $cpgmod ? 'YES' : 'NO';
        }

      # Generic feature types:
      # numeric: remove extra + sign
      } elsif   ($type eq 'numeric') {
        $pf =~ s/^\+//;
      # binary: translate boolean {0,1} to {YES,NO}
      } elsif ($type eq 'yesno') {
        $pf = $pf ? 'YES' : 'NO';
      }

    } else {

      # Specific features:
      if      ($pname eq 'AveNHet' || $pname eq 'AveNInt') {
        $pf = $with3D ? 0 : '?';
      } elsif ($pname eq 'MinDHet' || $pname eq 'MinDInt') {
        # 100 Angstrom denotes arbitrary large distance
        $pf = $with3D ? 100 : '?';

      # Generic feature types:
      # Force yesno and missno types to NO if undefined/missing
      } elsif ($type eq 'yesno' || $type eq 'missno') {
        $pf = 'NO';
      # Force misszero types to 0 if undefined/missing
      } elsif   ($type eq 'misszero') {
        $pf = 0;
      # All other undefined/missing features denoted by ?
      } else { $pf = '?'; }

    }

    &values($wname, $pf, $type) if $verbose;
    push @wf, $pf;

  }

  # Run all subroutines to calculate additional features
  for ($i=0; $i<@GENERATED; $i++) {
    my $coderef = $GENERATED[$i]->[0];
    next unless defined $coderef;
    &$coderef($i, \@wf, $aa1, $aa2);
  }

  my @wfset; $#wfset = scalar(grep(defined, @subset)) - 1;
  for ($i=0; $i<@wf; $i++) {
    $wfset[ $subset[$i] ] = $wf[$i] if defined $subset[$i];
  }

  push @wfset, length $set_id ? $set_id : '?' if $sids;
  push @wfset, $forced_class || $class || 'NEUTRAL';

  push @weka, \@wfset;

}
close(FH);

$opt{'f'} += 1 if $sids; # set_id is an extra feature
my $basename = fileparse($sumfile, '.output');
my $testfile = $basename . '.f' . $opt{'f'} . '.arff';
open(FH, ">$testfile") or die "Can't create temporary file: $testfile\n";
# Create an empty .arff if no valid lines were found in the PPH file:
# this can happen especially in a cluster (-q) mode when each PPH file
# holds output of a single query (which may easily fail). Empty arrf
# then serves as a flag for further pipeline.
close(FH), exit unless @weka;

my $relation = '@relation PPHv2.f' . $opt{'f'};

print FH $relation, "\n\n";

if      ($setidx == 3) {
  print FH<<'EOT';
@attribute nt1 {A,C,G,T}
@attribute nt2 {A,C,G,T}
@attribute site {NO,YES}
@attribute region {NO,PROPEP,SIGNAL,TRANSMEM}
@attribute phat numeric
@attribute score_delta numeric
@attribute score1 numeric
@attribute score2 numeric
@attribute num_observ numeric
@attribute delta_volume numeric
@attribute transversion {NO,YES}
@attribute cpg {NO,YES}
@attribute pfam_hit {NO,YES}
@attribute id_p_max numeric
@attribute id_p_snp {NO,YES}
@attribute id_q_min numeric
@attribute cpg_var1_var2 {NO,YES_A_T,YES_A_V,YES_C_R,YES_D_N,YES_E_K,YES_G_R,YES_G_S,YES_H_R,YES_I_V,YES_K_E,YES_L_P,YES_L_S,YES_M_T,YES_M_V,YES_N_D,YES_P_L,YES_Q_R,YES_R_C,YES_R_G,YES_R_H,YES_R_Q,YES_R_W,YES_S_G,YES_S_L,YES_T_A,YES_T_M,YES_V_A,YES_V_I,YES_V_M,YES_W_R}
@attribute cpg_transition {NO,YES_TRANSITION,YES_TRANSVERSION}
@attribute charge_change numeric
@attribute hydroph_change numeric
@attribute ali_ide numeric
@attribute ali_len numeric
@attribute acc_normed numeric
@attribute sec_str {HELIX,OTHER,SHEET}
@attribute map_region {ALPHA,BETA,OTHER}
@attribute delta_prop numeric
@attribute b_fact numeric
@attribute het_cont_ave_num numeric
@attribute het_cont_min_dist numeric
@attribute inter_cont_ave_num numeric
@attribute inter_cont_min_dist numeric
@attribute delta_volume_new numeric
@attribute delta_prop_new numeric
EOT
print FH '@attribute set_id numeric', "\n" if $sids;
  print FH<<'EOT';
@attribute class {DELETERIOUS,NEUTRAL}

@data
EOT
} elsif ($setidx == 4) {
  print FH<<'EOT';
@attribute score_delta numeric
@attribute score1 numeric
@attribute num_observ numeric
@attribute delta_volume numeric
@attribute pfam_hit {NO,YES}
@attribute id_p_max numeric
@attribute id_q_min numeric
@attribute cpg_transition {NO,YES_TRANSITION,YES_TRANSVERSION}
@attribute acc_normed numeric
@attribute b_fact numeric
@attribute delta_prop_new numeric
EOT
print FH '@attribute set_id numeric', "\n" if $sids;
  print FH<<'EOT';
@attribute class {DELETERIOUS,NEUTRAL}

@data
EOT
} elsif ($setidx == 5) {
  print FH<<'EOT';
@attribute score_delta numeric
@attribute score1 numeric
@attribute num_observ numeric
@attribute pfam_hit {NO,YES}
@attribute id_p_max numeric
@attribute id_q_min numeric
@attribute cpg_transition {NO,YES_TRANSITION,YES_TRANSVERSION}
@attribute delta_prop_new numeric
EOT
print FH '@attribute set_id numeric', "\n" if $sids;
  print FH<<'EOT';
@attribute class {DELETERIOUS,NEUTRAL}

@data
EOT
}

foreach $rec (@weka) {
  print FH join(',', @$rec), "\n";
}
close(FH);

if ($verbose) {
  my @wnames; $#wnames = scalar(grep(defined, @subset)) - 1;
  for ($i=0; $i<@PPH2WEKA; $i++) {
    $wnames[ $subset[$i] ] = $PPH2WEKA[$i]->[1] if defined $subset[$i];
  }
  for ($i=0; $i<@GENERATED; $i++) {
    $j = $i + @PPH2WEKA;
    $wnames[ $subset[$j] ] = $GENERATED[$i]->[1] if defined $subset[$j];
  }
  foreach $wname (@wnames) {
    printf '%22s', "$wname: ";
    my $type = ref $values{$wname};
    if      ($type eq 'HASH') {
      my $total = 0;
      my $missing = $values{$wname}->{'?'} || 0;
      $total += $missing;
      foreach $name (sort keys %{ $values{$wname} }) {
        next if $name eq '?';
        my $count = $values{$wname}->{$name};
        $total += $count;
        print "$name($count) + ";
      }
      print "?($missing) = $total\n";
    } elsif ($type eq 'ARRAY') {
      my ($min, $max, $sum, $count, $missing) = @{ $values{$wname} };
      print
        sprintf('%.4g:%.4g, AVE=', $min, $max),
        sprintf('%.3g', $sum / $count),
        ', COUNT=', $count,
        ', ?=', $missing,
        ', TOTAL=', ($missing + $count),
        "\n";
    } else {
      die "Illegal type encountered in values: $type\n";
    }
  }
}


# Subroutines
#------------------------------------------------------------------------------
sub values {
  my ($wname, $value, $type) = @_;
  # Count unique values for nominal features and check ranges for type numeric
  if ($type eq 'numeric' || $type eq 'misszero') {
    if ($value eq '?') {
      $values{$wname}->[4]++;
    } else {
      my ($min, $max, $sum, $count, $missing) = @{ $values{$wname} };
      if (defined $min) {
        if    ($value < $min) { $min = $value; }
        elsif ($value > $max) { $max = $value; }
        $count++;
        $sum += $value;
      } else {
        $min = $max = $sum = $value;
        $count = 1;
        $missing = 0 unless defined $missing;
      }
      @{ $values{$wname} } = ( $min, $max, $sum, $count, $missing );
    }
  } else {
    $values{$wname}->{$value}++;
  }
}
#------------------------------------------------------------------------------
sub charge {
  my($aa) = shift;

  if ($aa eq 'K' || $aa eq 'R') {
    return 1;
  } elsif ($aa eq 'E' || $aa eq 'D') {
    return -1;
  }

  return 0;
}
#------------------------------------------------------------------------------
sub propens_interval {
  my($x) = shift;
  ## 0 (0,5) [5,15) [15,30) [30,50) [50,75) [75,...)
  if($x==0) { return 0; }
  elsif (  0<$x && $x<5 ) { return 1; }
  elsif ( 5<=$x && $x<15) { return 2; }
  elsif (15<=$x && $x<30) { return 3; }
  elsif (30<=$x && $x<50) { return 4; }
  elsif (50<=$x && $x<75) { return 5; }
  else { return 6; }
}
#------------------------------------------------------------------------------
sub cpg_var1_var2 {
  my ($i, $wf, $aa1, $aa2) = @_;
  my $cpg          = $wf->[44];
  my $transversion = $wf->[42];
  die "cpg_var1_var2: invalid input values\n" unless defined $cpg && defined $transversion;
  my ($cpg_var1_var2, $cpg_transition);
  if ($cpg ne '?') {
    $cpg_var1_var2 = $cpg . '_' . $aa1 . '_' . $aa2;
    $cpg_var1_var2 = 'NO' unless $cpg eq 'YES' && $transversion eq 'NO';
    $cpg_var1_var2 =~ s/1_/YES_/;
    $cpg_transition =
      $cpg eq 'YES' ?
        ('YES_' . ($transversion eq 'YES' ? 'TRANSVERSION' : 'TRANSITION')) : 'NO';
  } else {
    $cpg_var1_var2 = $cpg_transition = '?';
  }
  # This relies upon the two features being consecutive
  push @{ $wf }, $cpg_var1_var2;
  push @{ $wf }, $cpg_transition;
  if ($verbose) {
    my ($wname, $type) = ($GENERATED[$i]->[1], $GENERATED[$i]->[2]);
    &values($wname, $cpg_var1_var2, $type);
    ($wname, $type) = ($GENERATED[$i+1]->[1], $GENERATED[$i+1]->[2]);
    &values($wname, $cpg_transition, $type);
  }
}
#------------------------------------------------------------------------------
sub charge_change {
  my ($i, $wf, $aa1, $aa2) = @_;
  my $cc;

  if ($aa1 eq 'H' && charge($aa2) == 1 || $aa2 eq 'H' && charge($aa1) == 1) {
    $cc = 0;
  } else {
    $cc = abs(charge($aa1) - charge($aa2));
  }
  push @{ $wf }, $cc;
  if ($verbose) {
    my ($wname, $type) = ($GENERATED[$i]->[1], $GENERATED[$i]->[2]);
    &values($wname, $cc, $type);
  }
}
#------------------------------------------------------------------------------
sub hydroph_change {
  my ($i, $wf, $aa1, $aa2) = @_;
  my $hyd = abs($HYDROPHOBICITY{$aa1} - $HYDROPHOBICITY{$aa2});
  push @{ $wf }, $hyd;
  if ($verbose) {
    my ($wname, $type) = ($GENERATED[$i]->[1], $GENERATED[$i]->[2]);
    &values($wname, $hyd, $type);
  }
}
#------------------------------------------------------------------------------
sub delta_volume_new {
  my ($i, $wf, $aa1, $aa2) = @_;
  my $acc_normed   = $wf->[29];
  my $delta_volume = $wf->[32];
  my $delta_prop   = $wf->[33];
  my ($delta_volume_new, $delta_prop_new);
  if($acc_normed ne '?') {
    $delta_volume_new = $acc_normed > $ACC_THRESH ? 0 : $delta_volume;
    $delta_prop_new   = $acc_normed > $ACC_THRESH ? 0 : $delta_prop;
  } else {
    $delta_volume_new = $delta_prop_new = '?';
  }
  # This relies upon the two features being consecutive
  push @{ $wf }, $delta_volume_new;
  push @{ $wf }, $delta_prop_new;
  if ($verbose) {
    my ($wname, $type) = ($GENERATED[$i]->[1], $GENERATED[$i]->[2]);
    &values($wname, $delta_volume_new, $type);
    ($wname, $type) = ($GENERATED[$i+1]->[1], $GENERATED[$i+1]->[2]);
    &values($wname, $delta_prop_new, $type);
  }
}
#------------------------------------------------------------------------------
