#!/usr/bin/env perl

use warnings;
use strict;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=pod

=head1 NAME

run_pph.pl - run a set of PolyPhen-2 queries

=head1 SYNOPSIS

run_pph.pl [options] infile [>outfile]

where options are:

  -s seqfile  read query protein sequences from FASTA-formatted seqfile

  -b dbase    BLAST sequence database; default is 'nrdb/uniref100'

  -d dumpdir  directory for auxiliary output files; default is 'scratch'

  -r N/M      partition input into M equally sized stripes and process
              stripe N

  -v level    debugging output, verbosity level=[1..3]

=head1 DESCRIPTION

Input file should contain tab-delimited lines describing SNPs with the
following 5 mandatory fields:

snp_id acc pos aa1 aa2

   * snp_id    - unique SNP identifier, arbitrary alphanumeric
                 string containing no spaces
   * acc       - protein accession or entry name; sequences will be
                 retrieved from the user-supplied file (optional)
                 or directly from UniProtKB/Swiss-Prot database
   * pos       - substitution position in the protein sequence
   * aa1       - reference aa residue as found at a query position
                 in the protein sequence
   * aa2       - substitution aa residue (both aa1 & aa2 are standard
                 single-letter residue codes)

See sets/test.input file included for an example of query format.

Advanced options:

  -e evalue  change E-value cutoff for BLAST searches; default is: 1e-3

  -p         use precomputed multiple alignments only

  -f         use full length protein sequences for multiple alignments;
             default is to use BLAST HSPs

  -t         create alignments with BLAST rather than MAFFT/LEON/Cluspack
             pipeline (use for debugging purposes only)

  -m         attempt to map proteins to genomic sequences to extract
             codons (default is to reverse-translate protein sequence)

  -x         save detailed annotation for each SNP in an XML file in
             the current working directory

  -c cnfdir  read configuration files from 'cnfdir'

=head1 AUTHORS

StS, IAA

=head1 SUBVERSION

 $LastChangedDate: 2011-05-28 14:37:25 -0400 (Sat, 28 May 2011) $
 $LastChangedRevision: 367 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

our $VERSION;

use Getopt::Std;
use Pod::Usage;
use File::Basename;
use File::Temp qw(tempdir);
use File::Path;
use Cwd qw(abs_path cwd);
use CGI;

BEGIN {
  $VERSION = '2.1.0r367';
  $Getopt::Std::STANDARD_HELP_VERSION = 1;
  use FindBin qw($RealBin);
  BEGIN {
    unless ($ENV{'PPH'}) {
      my $binpath = $RealBin;
      $binpath =~ s|/[^/]+$||;
      $ENV{'PPH'} = $binpath;
    }
  }
  use lib "$ENV{PPH}/perl";
}

# Need this before BEGIN block to import $DEBUG tested in it
use PPH::Config;

# Use BEGIN block here to ensure that command-line options are
# processed before Smart::Comments module is loaded since DEBUG
# flag can be altered from the command line.
my %opts;
BEGIN {
  # Process command-line options
  pod2usage('No arguments specified!') unless @ARGV;
  getopts('fhmptqxb:c:d:e:r:s:v:', \%opts) or pod2usage('Wrong arguments!');
  pod2usage(-verbose=>2) if $opts{'h'};
  pod2usage('No input file(s) specified!') unless @ARGV;
  # Adjust default configuration if options were specified on command line
  PPH::Config->new(\%opts) if %opts;
  if ($DEBUG) {
    # Test if we have Smart::Comments installed
    eval { require Smart::Comments; Smart::Comments->import(-ENV) };
    die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
  }
}

# All other PPH modules should follow PPH::Config for %CONFIG hash to be
# properly set before the modules that may attempt to utilize its contents.
use PPH;
use SNP;

$| = 1;

#SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
# Starting...
#SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

# Check if we are running under SGE or LSF and if this is an array job
my ($grid_mode, $jobid, $taskid, $maxtask, $grid_msg);
if (exists $ENV{'SGE_ROOT'} && exists $ENV{'JOB_ID'}) {       # SGE
  $grid_mode = 'SGE';
  $jobid  = $ENV{'JOB_ID'} ne 'undefined' ? $ENV{'JOB_ID'} : 0;
  $taskid = exists $ENV{'SGE_TASK_ID'} && $ENV{'SGE_TASK_ID'} ne 'undefined' ? $ENV{'SGE_TASK_ID'} : 0;
} elsif ($ENV{'LSB_QUEUE'} && exists $ENV{'LSB_JOBID'}) {     # LSF
  $grid_mode = 'LSF';
  $jobid  = $ENV{'LSB_JOBID'};
  $taskid = exists $ENV{'LSB_JOBINDEX'} ? $ENV{'LSB_JOBINDEX'} : 0;
}
my $array_mode = $jobid && $taskid ? 1 : 0;

if ($grid_mode) {
  # Adjust PATH for a grid job
  if ($CONFIG{'GRID_PATH'}) {
    my $ARCH = `uname -m`; chomp $ARCH;
    $ENV{'PATH'} = eval "qq{$CONFIG{GRID_PATH}}";
  }
  $grid_msg = "I am $grid_mode job $jobid \@ $ENV{HOSTNAME}\n";
}

if ($array_mode) {
  # GGI_STRIPES overwrites any array size specified initially
  $maxtask  = $ENV{'GGI_STRIPES'} || $ENV{'SGE_TASK_LAST'} || $ENV{'LSB_JOBINDEX_END'};
  $grid_msg = "I am $grid_mode job $jobid task $taskid of $maxtask \@ $ENV{HOSTNAME}\n";
}

# Debugging
warn $grid_msg if defined $grid_msg && !$opts{'q'};

# If -q option was specified then query file has already been processed in PPH::Config
unless ($opts{'q'}) {
  for (my $i=0; $i<@ARGV; $i++) {
    my $abspath = abs_path($ARGV[$i]);
    die "ERROR: Input file does not exist: $ARGV[$i]\n"
      unless defined $abspath && length $abspath && -e $abspath;
    $ARGV[$i] = $abspath;
  }
}

# Set current stripe index and total number of stripes
my ($stripe_no, $stripes);
if  ($opts{'r'}) {
  if      ($opts{'q'}) {
    warn "WARNING: Input stripe specification via '-r' option ignored in query mode\n";
  } elsif ($array_mode) {
    warn "WARNING: Explicit input stripe specification via '-r' option ignored in $grid_mode array mode\n";
  } else {
    ($stripe_no, $stripes) = $opts{'r'} =~ m|^(\d+)/(\d+)$|;
    die "ERROR: Illegal option's argument: -r $opts{'r'}\n" unless $stripe_no && $stripes;
  }
}
if ($array_mode) {
  ($stripe_no, $stripes) = ($taskid, $maxtask);
}

# Calculate range of input lines if striping option was specified
my $input_line_count  = 0;
my $input_query_count = 0;
my ($range_start, $range_end);
if ($stripes) {
  # Count how many input lines we have
  for (my $i=0; $i<@ARGV; $i++) {
    open(FIN, $ARGV[$i]) or die "ERROR: Can't open input file: $ARGV[$i]\n";
    while (<FIN>) {
      $input_line_count++;
      # Skip comments and empty lines
      next if /^#/ || /^\s*$/;
      $input_query_count++;
    }
    close(FIN);
  }
  # Calculate stripe range
  ($range_start, $range_end) = stripe_range($stripe_no, $stripes, $input_query_count);
  # Validate stripe range
  if ($range_start == 0 && $range_end == 0) {
    # In array mode, if $maxtask was set to a value higher than the total number of queries
    # in input then we kill the rest of array as soon as the current task gets beyond the
    # input range, to avoid wasting cluster slots on keeping validating out-of-range tasks.
    # Note: range_start/range_end are both zero if and only if stripe_no > input_query_count
    if ($array_mode) {
      if ($maxtask) {
        warn "WARNING: Task $jobid.$taskid exceeds input range, rest of array omitted\n";
        my $qdel = "sgesh_anko qdel $jobid -t " . ($stripe_no + 1) . "-$maxtask";
        my $qrc = `$qdel`;
        warn "WARNING: qdel error: $qrc\n" unless
          $qrc =~ /^\w+ has deleted job-array tasks \d+-\d+:\d+ of job \d+/ ||
          $qrc =~ /^\w+ has registered the job-array task \d+\.\d+ for deletion/;
        exit 0;
      }
    }
    die "ERROR: Input out of range (stripe $stripe_no / $stripes with $input_query_count queries in input)\n";
  }
}

my $pph = PPH->new($CONFIG{'SEQFILE'});

# Create a temporary directory inside TMP_OUT and chdir into it.
# Also set TMP_OUT to point to this directory. Save old CWD in
# case we later need to know where it pointed originally.
$CONFIG{'TMP_OUT'} = tempdir("$CONFIG{TMP_OUT}/PPH_XXXXXXXX");
$CONFIG{'CWD'} = cwd();
chdir $CONFIG{'TMP_OUT'} or die "WARNING: Failed to chdir into temporary directory: $CONFIG{TMP_OUT}\n";
### |   Scratch directory is set to    : $CONFIG{SCRATCH}
### |   Temporary directory created as : $CONFIG{TMP_OUT}

# Print header (formatted) unless running in array mode
print('#', join("\t", $pph->return_header(-formatted=>1)), "\n") unless $array_mode;

#----------------------------------------
# Read input
#----------------------------------------
my $input_line_no  = 0;
my $input_query_no = 0;
my $range_line_no  = 0;
my $query_no = 1;
my $input_no = 0;
while ($_ = $CONFIG{'QUERYSTRING'} || <>) {

  $input_line_no++;   # unfiltered input line number, used in error/warning messages
  $input_no++;

  # Skip comments and empty lines
  next if /^#/ || /^\s*$/;

  $input_query_no++;  # input query line number

  if ($stripes) {
    next if $input_query_no < $range_start;
    last if $input_query_no > $range_end;
  }

  $range_line_no++;   # total query input lines to process (i.e., within range but without comments)

  chomp;

  # Helps debugging of parallel tasks executed on different cluster nodes
  warn "INPUT: $_\n" if $array_mode;

  # Trim optional user comments
  my $comments = $1 if s/\s *#\s*(.*)$//;

  # We can have either a single dbSNP rsID in input or 5 columns of protein
  # substitution data (snp id, protein id, position, aa1, aa1).
  my $paracount = 0;
  my @params = split /\s+/;
  map { s/^\s+//; s/\s+$//; $paracount++ if defined and length } @params;
  unless ($paracount == 5 || $paracount == 1 && $params[0] =~ /^rs/) {
    warn "ERROR: Incorrect number of parameters ($paracount) in input at line $input_no\n";
    last if $CONFIG{'QUERYSTRING'};
    next;
  }

  my ($SNP_id, $acc, $pos, $aa1, $aa2) = @params;

  #------------------------------------------
  # Creating and verifying SNP object
  #------------------------------------------
  my ($SNP, $msg);
  eval {
    $SNP = SNP->new(
        {
          SNP_Id => $SNP_id,
          Acc    => $acc,
          Pos    => $pos,
          aa1    => $aa1,
          aa2    => $aa2
        }
      )
  };
  if ($@) {
    warn "ERROR: Verification failed for input line $input_no: $@";
    last if $CONFIG{'QUERYSTRING'};
    next;
  }

  # Add optional user comments if present
  $SNP->{'Comments'} = $comments if defined $comments and length $comments;

  $msg = $pph->prepareInput($SNP);

  # Failure of prepararation if $msg set
  if ($msg) {
      warn "ERROR: $msg\n";
      last if $CONFIG{'QUERYSTRING'};
      next;
  }
  #------------------------------------------

  #------------------------------------------
  # Running PolyPhen
  #------------------------------------------
  $pph->analyze($SNP);
  #------------------------------------------

  #------------------------------------------
  # Evaluating data
  #------------------------------------------
  $pph->interpret_results($SNP);
  #------------------------------------------

  #-----------------------------------------------------
  # Saving SNP annotation object to a file in XML format
  # (optional, currently only used by the web interface)
  #-----------------------------------------------------
  if ($CONFIG{'XMLOUTPUT'}) {
    # Add sources versioning information if available
    $SNP->{'Source'}{'Predictor'} = "PolyPhen-2 v$VERSION" if defined $VERSION;
    if (exists $CONFIG{'UNIPROT'}) {
      my $version_file = dirname($CONFIG{'UNIPROT'}) . '/VERSION';
      if (-s $version_file) {
        my $desc;
        if (open my $F, '<', $version_file) {
          chomp($desc = <$F>);
          close $F;
        }
        $SNP->{'Source'}{'Sequence'} = $desc if defined $desc && length $desc;
      }
    }
    if (exists $CONFIG{'PDB2FASTA'}) {
      my $version_file = $CONFIG{'PDB2FASTA'} . '/VERSION';
      if (-s $version_file) {
        my $desc;
        if (open my $F, '<', $version_file) {
          chomp($desc = <$F>);
          close $F;
        }
        $SNP->{'Source'}{'Structure'} = $desc if defined $desc && length $desc;
      }
    }
    # Note: use saved CWD for output target directory path since
    # the script may be run chdir()'ed into a temporary directory
    my $xmlfile = $CONFIG{'CWD'} . '/';
    # Use the original SNP ID for XML file name since it is guaranteed
    # to be always present in a query
    $xmlfile .= "$SNP_id.xml";
    open my $F, '>', $xmlfile or die "Can't create file: $xmlfile\n";
    $SNP->snp2xml($F);
    close $F;
  }
  #-----------------------------------------------------

  #-----------------------------------------------
  # Printing tab-delimited summary table to STDOUT
  #-----------------------------------------------
  print join("\t", $pph->format_result($SNP)), "\n";
  #-----------------------------------------------

  last if $CONFIG{'QUERYSTRING'};
}

warn "TOTAL: $range_line_no queries processed\n" if $array_mode;

# Calculate [start,end] range of a stripe in dataset
# Note: all indices passed and returned are one-based
#
# Arguments:  stripe index, number of stripes, dataset size
# Returns:    data start index, data end index; both indices
#             returned are zero if no more data fall within
#             a stripe specified (e.g. number of stripes is
#             larger than the number of data items)
#
sub stripe_range {
  my ($stripe_no, $stripes, $size) = @_;

  die "ERROR: Stripe index too large: $stripe_no > $stripes\n" if $stripe_no > $stripes;

  my $p = int($size / $stripes);
  my $r = $size % $stripes;

  my $shift_start = ($stripe_no <= $r ? $stripe_no : $r + 1) - 1;
  my $shift_end   = ($stripe_no <= $r ? $stripe_no : $r);

  my $start = ($stripe_no - 1)*$p + 1 + $shift_start;
  my $end   = $stripe_no*$p + $shift_end;

  if ($start > $end) {
    return (0, 0);
  } else {
    return ($start, $end);
  }
}

END {
  # Newer rmtree() versions refuse to remove directory which is
  # still current so we need to first go back to where we belong
  chdir $CONFIG{'CWD'} if exists $CONFIG{'CWD'};
  if (!$DEBUG && exists $CONFIG{'CWD'} && -d $CONFIG{'TMP_OUT'}) {
    rmtree($CONFIG{'TMP_OUT'})
      or warn "Failed to remove temporary directory: $CONFIG{TMP_OUT}\n";
  }
}
