package PPH;
use Carp qw(cluck confess);
use warnings;
use strict;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH - Main Module of PolyPhen - former pph.pl, pph_main.pl

=head1 SYNOPSIS

 use SNP;
 use PPH;

 my $SNP = SNP->new( aa1 => 'D' , aa2 => 'N', Pos => '8', Id => 'HIS4_ECOLI');
 my $pph = PPH->new( 'my_sequences.fas' ); # Argument is optional
 $pph->analyze($SNP);
 $pph->interpret_results($SNP);
 print $pph->return_header(), "\n";
 print $pph->print_result(), "\n";


=head1 DESCRIPTION

Access to PolyPhen routines

=head1 AUTHOR

StS

=head1 SUBVERSION

 $LastChangedDate: 2009-11-28 00:15:09 -0500 (Sat, 28 Nov 2009) $
 $LastChangedRevision: 254 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

#BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 0.1;
    @ISA           = qw(Exporter);
    @EXPORT        = qw(&new);
    @EXPORT_OK     = qw();
    %EXPORT_TAGS = ();
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}

our @EXPORT_OK;

use Fcntl qw(:flock);
use File::Copy;

use PPH::Data;
use PPH::Seq;
use PPH::Profile;
use PPH::3D;
use PPH::Rules;
use PPH::BLAT;
use PPH::Codon;
use PPH::Pfam;

#BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB

# Header for tabulated output (51 columns, including optional Comments)
our @tabheader = qw(
  o_snp_id
  o_acc
  o_pos
  o_aa1
  o_aa2
  snp_id
  acc
  pos
  aa1
  aa2
  nt1
  nt2
  prediction
  based_on
  effect
  site
  region
  PHAT
  dScore
  Score1
  Score2
  Nobs
  Nstruct
  Nfilt
  PDB_id
  PDB_pos
  PDB_ch
  ident
  length
  NormAcc
  SecStr
  MapReg
  dVol
  dProp
  B-fact
  H-bonds
  AveNHet
  MinDHet
  AveNInt
  MinDInt
  AveNSit
  MinDSit
  Transv
  CodPos
  CpG
  MinDJnc
  PfamHit
  IdPmax
  IdPSNP
  IdQmin
);
if ($CONFIG{'ALNSCORES'}) { # optional alignment-based scores
  push @tabheader,
		'Len',
    'Nseqs',
    'Nsites',
    'Nsubs',
    'Nvars',
    'Ntree',
    'Sites',
    'Tdist'
  ;
}

# Tabulated output format
our @tabformat = (
  # Original query values
  '%-28s', #  1    o_snp_id
  '%16s',  #  2    o_acc
  '%6s',   #  3    o_pos
  '%5s',   #  4    o_aa1
  '%5s',   #  5    o_aa2
  # Mapped query
  '%-28s', #  6    snp_id
  '%8s',   #  7    acc
  '%6s',   #  8    pos
  '%3s',   #  9    aa1
  '%3s',   # 10    aa2
  '%3s',   # 11    nt1
  '%3s',   # 12    nt2
  # PolyPhen-1 annotations
  '%18s',  # 13    prediction
  '%20s',  # 14    based_on
  '%10s',  # 15    effect
  # Features
  '%8s',   # 16    site
  '%8s',   # 17    region
  '%8s',   # 18    PHAT
  '%6s',   # 19    dScore
  '%6s',   # 20    Score1
  '%6s',   # 21    Score2
  '%6s',   # 22    Nobs
  '%8s',   # 23    Nstruct
  '%6s',   # 24    Nfilt
  '%6s',   # 25    PDB_id
  '%7s',   # 26    PDB_pos
  '%6s',   # 27    PDB_ch
  '%6s',   # 28    ident
  '%6s',   # 29    length
  '%7s',   # 30    NormAcc
  '%6s',   # 31    SecStr
  '%6s',   # 32    MapReg
  '%6s',   # 33    dVol
  '%6s',   # 34    dProp
  '%6s',   # 35    B-fact
  '%8s',   # 36    H-bonds
  '%8s',   # 37    AveNHet
  '%8s',   # 38    MinDHet
  '%8s',   # 39    AveNInt
  '%8s',   # 40    MinDInt
  '%8s',   # 41    AveNSit
  '%8s',   # 42    MinDSit
  '%6s',   # 43    Transv
  '%6s',   # 44    CodPos
  '%3s',   # 45    CpG
  '%8s',   # 46    MinDJnc
  '%12s',  # 47    PfamHit
  '%8s',   # 48    IdPmax
  '%8s',   # 49    IdPSNP
  '%8s'    # 50    IdQmin
);
if ($CONFIG{'ALNSCORES'}) { # optional alignment-based scores
  push @tabformat,
    '%6s', # 51    Len
    '%6s', # 52    Nseqs
    '%6s', # 53    Nsites
    '%6s', # 54    Nsubs
    '%6s', # 55    Nvars
    '%6s', # 56    Ntree
    '%s',  # 57    Sites
    '%s',  # 58    Tdist
  ;
}

#CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

#----------------------------------------------------------------------

=head1 ROUTINES

=head2 new

 Usage:    $pph->new();
           $pph->new('sequences.fa')
 Function: Initialize Object PPH
 Args    : FASTA file containing the protein sequences for the SNPs
           (optional)
 Returns : PPH object

=cut

#----------------------------------------
sub new {
    my ($class, $FASTA_file) =  @_;

    ### [<now>] Starting new PPH instance...
    my $self = {};
    bless($self, $class);

    (defined($FASTA_file) and (-e $FASTA_file))
        and do {
            #----------------------------------------
            # Do a simple test if formatting of
            # uniprot.seq file is like:
            # >id
            # seq
            # >id
            # ...
            #----------------------------------------
            open my $F, '<', "$CONFIG{UNIPROT}.seq"
                or confess "Can't open file $CONFIG{UNIPROT}.seq";
            <$F>; <$F>; $_ = <$F>;
            /^>/ or confess "Wrong format of $CONFIG{UNIPROT}.seq\n".
                "Sequence should be in one line only\n$_\n";
            close $F;
            #----------------------------------------

            $self->{'FASTA_file'} = $FASTA_file;
        };
    return $self;
}
#----------------------------------------------------------------------

#CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

#======================================================================

=head2 prepareInput

 Usage   : PPH->prepareInput($SNP)
 Function: This routine checks if protein of SNP is contained in
           Uniprot - otherwise retrieve sequence from given sequence
           file
 Args    : Object of type SNP
 Returns : empty string if successful
           - the SNP object will contain the info

=cut

#========================================
# IAA: Changes to allow sequence extraction from user-supplied sequence database
#      to fail with a warning (instead of fatal error); in such case an attempt is
#      made to extract sequence with the matching protein ID from UniProt/Swiss-Prot
sub prepareInput {
  my ($self, $SNP) = @_;

  $SNP->{UniProt} = 0;              # Assume initially that the protein is not in UniProtKB
  my $orig_acc    = $SNP->{Acc};    # Original Acc may be overwritten by uniprot_id2acc()
  my $orig_snpid  = $SNP->{SNP_Id}; # Original SNP_Id may be overwritten by .snp2acc lookup
  # All calculated files are saved under SCRATCH directory by default
  $CONFIG{'SAVEPATH'} = $CONFIG{'SCRATCH'};

  # Check if we ONLY have a dbSNP rsID submitted and then try to extract
  # corresponding sequence variant data from the .snp2acc file.
  my $dbsnp =
    $SNP->{SNP_Id} =~ /^rs/ &&
    !defined $SNP->{Acc} &&
    !defined $SNP->{Pos} &&
    !defined $SNP->{aa1} &&
    !defined $SNP->{aa2};
  if ($dbsnp) {
    ($SNP->{Acc}, $SNP->{Pos}, $SNP->{aa1}, $SNP->{aa2}) = PPH::Seq::snp2acc($SNP->{SNP_Id});
    return "dbSNP $SNP->{SNP_Id} was not found in the available version of UniProtKB"
      unless length $SNP->{Acc};
  }

  # Attempt to extract sequence from the user-supplied sequence database if
  # specified with -s option (skip if this is a dbSNP SNP ID).
  if (exists $self->{FASTA_file} && !$dbsnp) {
    (undef, undef, $SNP->{Desc}, $SNP->{Seq}) =
      PPH::Seq::fetch_FASTA($SNP->{Acc}, $self->{FASTA_file});
    # Do some cleanup of the user-submitted sequence
    if (defined $SNP->{Seq} && length $SNP->{Seq}) {
      # Users often submit sequences with a translated terminator codon
      $SNP->{Seq} =~ s/\*+$//;
      # Occasionally users submit bogus sequences with multiple ORFs
      warn "WARNING: Terminator codon(s) encountered in $SNP->{Acc} query sequence for\n"
        if $SNP->{Seq} =~ tr/*/X/;
    }
  }

  # Attempt to extract sequence from uniprot_sprot_varsplic.fasta file if
  # it is present and if an isoform accession was specified (e.g. P01234-2).
  # Such sequences will not be considered "known proteins" (unless
  # successfully mapped to UniProt down below) since they may differ
  # from canonical ones substantially, causing annotated positions to
  # shift or disappear.
  unless (defined $SNP->{Seq} && length $SNP->{Seq}) {
    my $varsplic_file = $CONFIG{'UNIPROT'} . '.varsplic';
    if ($SNP->{Acc} =~ /^[A-QR-Z]\d[A-Z0-9]{3}?\d-\d+$/ && -e $varsplic_file) {
      (undef, undef, $SNP->{Desc}, $SNP->{Seq}) =
        PPH::Seq::fetch_FASTA($SNP->{Acc}, $varsplic_file);
    }
    # Save all files to the global precomputed repository if SAVEKNOWN
    # and if sequence has been successfully mapped to a known isoform
    # even though this is not a UniProtKB "known protein".
    $CONFIG{'SAVEPATH'} = $CONFIG{'PRECOMPATH'} if $CONFIG{'SAVEKNOWN'} &&
      defined $SNP->{Seq} && length $SNP->{Seq};
  }

  # Sequence successfully extracted, now attempt to match it to UniProtKB/Swiss-Prot
  if (defined $SNP->{Seq} && length $SNP->{Seq}) {
    my ($acc, $id, $desc, $seq) =
      PPH::Seq::grep_in_FASTA($SNP->{Seq}, "$CONFIG{UNIPROT}.seq");
    # Exact (sub)sequence match found in UniProt/Swiss-Prot
    if (defined $acc) {
      $SNP->{UniProt}   = 1;
      $SNP->{Orig_Acc}  = $orig_acc;
      $SNP->{Orig_Desc} = $SNP->{Desc};
      $SNP->{Orig_Pos}  = $SNP->{Pos};
      $SNP->{Id}        = $id;
      $SNP->{Acc}       = $acc;
      $SNP->{Desc}      = $desc;
      # New position
      $SNP->{Pos}       = $SNP->{Pos} + index($seq, $SNP->{Seq});
      $SNP->{Seq}       = $seq;
      # Save all files to the global precomputed repository if SAVEKNOWN,
      # since sequence has been successfully mapped to a known protein.
      $CONFIG{'SAVEPATH'} = $CONFIG{'PRECOMPATH'} if $CONFIG{'SAVEKNOWN'};
    }

  # Either user sequence file was not specified, or sequence failed to be
  # extracted from it, or this is a dbSNP SNP. Attempt to match and extract
  # sequence by its UniProtKB protein accession.
  } else {
    #warn "WARNING: FASTA entry $orig_acc was not found in $self->{FASTA_file}\n"
    #  if exists $self->{FASTA_file} && !$dbsnp;

    # Check if the protein ID submitted is from one of the supported alien databases
    # (RefSeq, GenBank) and map it to a primary UniProtKB accession if it is.
    # Note, this operation is cheap for known UniProtKB accessions since it will not
    # trigger ID hash load/search and will simply return an empty string instead.
    # Skip ID mapping for dbSNP since it's already been mapped to a primary accession.
    $SNP->{Acc} = PPH::Seq::xdb2acc($orig_acc) or
      $SNP->{Acc} = PPH::Seq::uniprot_id2acc($orig_acc) unless $dbsnp;

    return "Unable to locate protein entry $orig_acc in the available version of UniProtKB"
      unless length $SNP->{Acc};

    # Fetch sequence from the built-in UniProtKB database by its accession
    (undef, undef, $SNP->{Desc}, $SNP->{Seq}) =
      PPH::Seq::fetch_FASTA($SNP->{Acc}, "$CONFIG{UNIPROT}.seq");

    return "Failed to extract $orig_acc query sequence from the available version of UniProtKB"
      unless exists $SNP->{Seq} && length $SNP->{Seq};

    # At this point the submitted variant has been successfully mapped to a UniProtKB entry.
    # Save all files for known proteins to the global precomputed repository if SAVEKNOWN
    # option is set.
    $SNP->{UniProt}  = 1;
    $CONFIG{'SAVEPATH'} = $CONFIG{'PRECOMPATH'} if $CONFIG{'SAVEKNOWN'};
    $SNP->{Orig_Acc} = $orig_acc if defined $orig_acc && $orig_acc ne $SNP->{Acc};

  }

  #----------------------------------------
  # Check if everything is ok
  #----------------------------------------
  unless ($SNP->{'DoNotVerify'}) {
    my $seqlen = length $SNP->{Seq};
    return "Invalid variation position ($SNP->{Pos}) specified for $SNP->{Acc} query sequence, valid positions are [1..$seqlen]"
      if $SNP->{Pos} <= 0 || $SNP->{Pos} > $seqlen;
    # Query sequence residue at the position of substitution
    my $qres = substr($SNP->{Seq}, $SNP->{Pos} - 1, 1);
    return "Neither AA1 ($SNP->{aa1}) nor AA2 ($SNP->{aa2}) in input matches $SNP->{Acc} query sequence residue ($qres) at position ($SNP->{Pos})"
      if $SNP->{aa1} ne $qres && $SNP->{aa2} ne $qres;
    # Default substitution direction is:
    #        (aa1==qres)->aa2
    # but if REVERSEDIRECTION is set then we will look instead at:
    #   aa1->(aa2==qres)
    # This is sometimes useful for analysing mutations in reference
    # organism (e.g., human) lineage
    my $refkey  = $CONFIG{'REVERSEDIRECTION'} ? 'aa2' : 'aa1';
    if ($SNP->{$refkey} ne $qres) {
      if ($CONFIG{'SWAPRESIDUES'}) {
        warn "WARNING: Swapped input residues AA1 ($SNP->{aa1}) and AA2 ($SNP->{aa2}) for $SNP->{Acc} query sequence at position ($SNP->{Pos})\n";
        # Save original AA pair
        $SNP->{o_aa1} = $SNP->{aa1};
        $SNP->{o_aa2} = $SNP->{aa2};
        # Now swap them
        $SNP->{aa1} = $SNP->{o_aa2};
        $SNP->{aa2} = $SNP->{o_aa1};
      } else {
        return "Wrong amino acid residue ($qres) found instead of ($SNP->{$refkey}) in $SNP->{Acc} query sequence at position ($SNP->{Pos})";
      }
    }
  }

  # If this is a UniProt protein and no dbSNP rsID was supplied in input then
  # try to locate rsID matching this substitution in a local index file
  if ($CONFIG{'MAP2DBSNPS'} && !$dbsnp && $SNP->{UniProt}) {
    my $snp2acc = $CONFIG{'UNIPROT'} . '.snp2acc';
    if (-e $snp2acc) {
      open my $F, '-|', "egrep -s -m 1 '\t$SNP->{Acc}\t$SNP->{Pos}\t$SNP->{aa1}\t$SNP->{aa2}\$|\t$SNP->{Acc}\t$SNP->{Pos}\t$SNP->{aa2}\t$SNP->{aa1}\$' $snp2acc";
      my $match = <$F>;
      close $F;
      if (defined $match) {
        my $rsid = (split /\t/, $match)[0];
        if ($rsid) {
          $SNP->{SNP_Id}      = $rsid;
          $SNP->{Orig_SNP_Id} = $orig_snpid;
        }
      }
    }
  }

  # Replace Selenocysteine (U) by Cysteine (C) in the sequence since
  # MAFFT as well as most other tools don't understand 'U' as aa code
  $SNP->{Seq} =~ tr/Uu/Cc/;

  $SNP->{Len} = length $SNP->{Seq};

  my $seqfile = "$SNP->{Acc}.seq";
  open my $W, '>', $seqfile or confess "Can't create file: $seqfile";
  print $W ">" . $SNP->{'Acc'} . "\n" . $SNP->{'Seq'} . "\n";
  close $W;

  return '';  # Success
}
#======================================================================


#======================================================================

=head2 analyze

 Usage   : PPH->analyze($SNP)
 Function: This is the main routine of PolyPhen:

             o It validates the SNP object by:
               a) Checking if SNP is not a synonymous substitution
               b) The protein sequence is given either by Uniprot
                  (preferrably) or by the sequence file (PolyPhen
                  checks if sequence is nevertheless in Uniprot)
             o It finds Uniprot annotations for the SNP
               (e.g. transmembrane) or which are in proximity to SNP
             o It calculates the PSIC Profile score for the
               substitution based on a sequence alignment
             o It checks structural parameters and checks for possible
               interactions to Uniprot sites

  Args    : Object of type SNP
  Returns : undef if successful

=cut

#========================================
sub analyze {
    my ($self, $SNP) = @_;
    ### PPH__analyze: $SNP

    #----------------------------------------
    # Check for UniProt annotations
    #### PPH__Seq__get_uniprot_annotation: $SNP
    #### check: $CONFIG{CALCUNIPROT} and $SNP->{UniProt}
    PPH::Seq::get_uniprot_annotation($SNP) if $CONFIG{'CALCUNIPROT'} and $SNP->{'UniProt'};

    #----------------------------------------
    # Check for Pfam annotations
    #### PPH__Pfam__find_pfam: $SNP
    #### check: $CONFIG{FINDPFAM} and $SNP->{UniProt}
    PPH::Pfam::find_pfam($SNP) if $CONFIG{'FINDPFAM'} and $SNP->{'UniProt'};

    #----------------------------------------
    # Calculate PSIC
    #### PPH__Profile__set_PSIC_profile: $SNP
    if ($CONFIG{'CALCPRF'}) {
      my $lockfile = "$CONFIG{SAVEPATH}/lock/$SNP->{Acc}.lock";
      open(FLK, ">$lockfile") or die "ERROR: Can't create lockfile $lockfile: $!\n";
      flock(FLK, LOCK_EX) or die "ERROR: Failed to obtain lock for $lockfile: $!\n";
      PPH::Profile::set_PSIC_profile($SNP);
      close(FLK);
    }

    #----------------------------------------
    # Calculate 3D
    #### PPH__3D__set_3D_parameters: $SNP
    if ($CONFIG{'CALC3D'}) {
      my $lockfile = "$CONFIG{SAVEPATH}/lock/$SNP->{Acc}.lock";
      open(FLK, ">$lockfile") or die "ERROR: Can't create lockfile $lockfile: $!\n";
      flock(FLK, LOCK_EX) or die "ERROR: Failed to obtain lock for $lockfile: $!\n";
      PPH::3D::set_3D_parameter($SNP);
      close(FLK);
    }

    #---------------------------------------
    # Calculate CpG / Exon junction distance
    #### PPH__BLAT__SNP2refGene: $SNP
    PPH::BLAT::SNP2refGene($SNP) if $CONFIG{'CALCCPG'};

    #### Upon return: $SNP
    return 1;
}
#======================================================================


#======================================================================

=head2 interpret_results

 Usage   : PPH->interpret_results($SNP)
 Function: This routine evaluates all parameters given for the SNP
           object and stores them in $SNP->{Rules}
 Args    : Object of type SNP
 Returns : TRUE if successful - SNP will contain the prediction

=cut

#========================================

sub interpret_results {
    my ($self, $SNP) = @_;
    ### PPH__interpret_results: $SNP

    $SNP->{'Rules'} = PPH::Rules::test($SNP);
    #### Upon return: $SNP->{'Rules'}
    return 1;
}
#======================================================================

#======================================================================

=head2 return_header

 Usage   : PPH->return_header([-format=>1])
 Args    : Optional -format flag to return formatted column headers
 Returns : Array containing all column headers as either raw or
           formatted strings depending on the format flag

=cut

#========================================
sub return_header {
  my ($self, %p);
  if (ref($_[0])) { ($self, %p) = @_; } # object method call
  else            { %p = @_;          } # package subroutine call
  if ($p{'-formatted'} || $p{'formatted'}) {
    my @formatted = ();
    for (my $i=0; $i<@tabheader; $i++) {
      push @formatted, sprintf($tabformat[$i], $tabheader[$i]);
    }
    return @formatted;
  } else {
    return @tabheader;
  }
}
#======================================================================


#======================================================================

=head2 format_result

 Usage   : PPH->format_result($SNP)
 Function: Prints values of SNP object which were filled by PolyPhen
 Args    : Object of type SNP
 Returns : Array of values computed from PolyPhen

=cut

#========================================
sub format_result {
    my ($self, $SNP) = @_;
    ##### PPH__format_result
    my @out;
    #----------------------------------------
    # Original input values if present,
    # otherwise they are same as mapped
    push @out, (
        _value($SNP, 'Orig_SNP_Id') || _value($SNP, 'SNP_Id'),
        _value($SNP, 'Orig_Acc')    || _value($SNP, 'Acc'),
        _value($SNP, 'Orig_Pos')    || _value($SNP, 'Pos'),
        _value($SNP, 'o_aa1')       || _value($SNP, 'aa1'),
        _value($SNP, 'o_aa2')       || _value($SNP, 'aa2')
    );
    #----------------------------------------

    #----------------------------------------
    # Mapped input values
    push @out, (
        _value($SNP, 'SNP_Id'),
        _value($SNP, 'Acc'),
        _value($SNP, 'Pos'),
        _value($SNP, 'aa1'),
        _value($SNP, 'aa2'),
        _value($SNP, 'Gene', 'nt1'),
        _value($SNP, 'Gene', 'nt2')
    );
    #----------------------------------------

    #----------------------------------------
    # Predictions
    push @out, (
        _value($SNP, 'Rules', 'Prediction'),
        _value($SNP, 'Rules', 'Basis'),
        _value($SNP, 'Rules', 'Effect')
    );
    # ((_value($SNP, "Rules", "Effect") and
    # exists($PPH::Data::Effect{$SNP->{Rules}{Effect}}{Name}))
    # ?
    # $PPH::Data::Effect{$SNP->{Rules}{Effect}}{Name}
    # : ""));
    #----------------------------------------

    #----------------------------------------
    # Uniprot annotation if available
    push @out, (
        _value($SNP, 'Annot', 'Site'),
        _value($SNP, 'Annot', 'Region'),
        _value($SNP, 'Annot', 'PHAT')
    );
    #----------------------------------------

    #----------------------------------------
    # General alignment info
    push @out, (
        _value($SNP, 'PSIC', 'Delta'),
        _value($SNP, 'PSIC', 'Score1'),
        _value($SNP, 'PSIC', 'Score2'),
        _value($SNP, 'PSIC',   'Nobs'),
        _value($SNP, 'Struct', 'NumHits'),
        _value($SNP, 'Struct', 'NumHitsFiltered')
    );
    #----------------------------------------

    #----------------------------------------
    # Structure part
    # - get info for first hit only
    #----------------------------------------
    my ($map, $cont);
    $map = $SNP->{Struct}{Maps}[0] if exists $SNP->{Struct};
    unless (defined $map) {
      push @out, ('') x 18;
    } else {
      my $nhsp = $CONFIG{CONTALLHITS} ? scalar @{$SNP->{Struct}{Maps}} : 1;
      $cont = $SNP->{Struct}{BestContact} if exists $SNP->{Struct}{BestContact};
      push @out, (
        _value($map, 'Hsp', 'PDB_id'),
        _value($map, 'Hsp', 'PDB_pos'),
        _value($map, 'Hsp', 'PDB_chain'),
        sprintf('%.2f', _value($map, 'Hsp', 'Identity')),
        _value($map, 'Hsp', 'Hsp_align-len'),
        (_value($map, 'Params', 'AccNorm') ?
              sprintf('%.3f', $map->{'Params'}{'AccNorm'}) : ''),
        _value($map, 'Params', 'SecStr'),
        _value($map, 'Params', 'MapRegion'),
        _value($map, 'Params', 'dVol'),
        _value($map, 'Params', 'dProp'),
        _value($map, 'Params', 'NormedB'),
        (_value($map, 'Params', 'Hbonds') ?
              scalar @{$map->{Params}{Hbonds}} : ''),
        (_value($cont, 'Number', 'Heteroatoms') ?
              sprintf('%.2g', $cont->{Number}{Heteroatoms} / $nhsp) : ''),
        _value($cont, 'Heteroatoms', 'Dist'),
        (_value($cont, 'Number', 'InterChain') ?
              sprintf('%.2g', $cont->{Number}{InterChain} / $nhsp) : ''),
        _value($cont, 'InterChain', 'Dist'),
        (_value($cont, 'Number', 'CritSites') ?
              sprintf('%.2g', $cont->{Number}{CritSites} / $nhsp) : ''),
        _value($cont, 'CritSites', 'Dist')
      );
    }
    #----------------------------------------

    #----------------------------------------
    # Adding CpG / Exon-Intron Junction info
    #----------------------------------------
    my $transv = _value($SNP, 'Gene', 'Transversion');
    if ($transv ne '') {
        push @out, $transv;
    } elsif (_value($SNP, 'Gene', 'nt1') ne '') {
        push @out,
            PPH::Codon::is_transversion(
                _value($SNP, 'Gene', 'nt1'),
                _value($SNP, 'Gene', 'nt2')
            );
    } else {
        push @out, '';
    }
    my $cpg = '';
    if ( _value($SNP, 'Gene', 'CodonPos') ne '') {
        push @out, _value($SNP, 'Gene', 'CodonPos');
        $cpg =  0;
        $cpg =  1 if (_value($SNP, 'Gene', 'CpG1') ne '');
        $cpg += 2 if (_value($SNP, 'Gene', 'CpG2') ne '');
        push @out, $cpg;

        my $d  = _value($SNP, 'Gene', 'd3junction');
        $d     = 10000 if ($d eq ''  or $d  == -1);
        my $d2 = _value($SNP, 'Gene', 'd5junction');
        $d2    = 10000 if ($d2 eq '' or $d2 == -1);
        $d     = (($d2 < $d) ? (-1 * $d2) : $d);

        if ($d == 10000) {
            push @out, '';
        } else {
            push @out, $d;
        }
    } else {
      push @out, ('') x 3;
    }

    #----------------------------------------
    # Adding Pfam
    #----------------------------------------
    push @out, _value($SNP, 'Annot', 'Domain', 'pfam');

    #----------------------------------------
    # Adding Substitution Score
    # IdPmax
    # IdPSNP
    # IdPQmin
    #----------------------------------------
    if (_value($SNP, 'PSIC', 'Aln')) {
        my $val = _value($SNP, 'PSIC', 'Aln', 'IdentPmax');
        push @out, ($val eq '') ? '' : sprintf('%.3f', $val * 100);
        $val = _value($SNP, 'PSIC', 'Aln', 'IdentPSNP');
        push @out, ($val eq '') ? '' : sprintf('%.3f', $val * 100);
        $val = _value($SNP, 'PSIC', 'Aln', 'IdentQmin');
        push @out, ($val eq '') ? '' : sprintf('%.3f', $val * 100);
    } else {
        push @out, ('') x 3;
    }

    #----------------------------------------
    # Optional alignment-based scores
    # Len
    # Nseqs
    # Nsites
    # Nsubs
    # Nvars
    # Ntree
    # Sites
    # Tdist
    #----------------------------------------
    if ($CONFIG{'ALNSCORES'}) {
			my $val;
			$val = _value($SNP, 'Len');
			push @out, ($val eq '') ? '' : sprintf('%d', $val);
      if (_value($SNP, 'PSIC', 'Aln')) {
          $val = _value($SNP, 'PSIC', 'Aln', 'Nseqs');
          push @out, ($val eq '') ? '' : sprintf('%d', $val);
          $val = _value($SNP, 'PSIC', 'Aln', 'Nsites');
          push @out, ($val eq '') ? '' : sprintf('%d', $val);
          $val = _value($SNP, 'PSIC', 'Aln', 'Nsubs');
          push @out, ($val eq '') ? '' : sprintf('%d', $val);
          $val = _value($SNP, 'PSIC', 'Aln', 'Nvars');
          push @out, ($val eq '') ? '' : sprintf('%d', $val);
          $val = _value($SNP, 'PSIC', 'Aln', 'Ntree');
          push @out, ($val eq '') ? '' : sprintf('%.3f', $val);
          $val = _value($SNP, 'PSIC', 'Aln', 'Sites');
          push @out, ($val eq '') ? '' : $val;
          $val = _value($SNP, 'PSIC', 'Aln', 'Tdist');
          push @out, ($val eq '') ? '' : $val;
      } else {
					push @out, ('') x 7;
      }
    }

    # Let's ignore the above formatting and start from scratch ;)
    map { if (defined) { s/^\s+//; s/\s+$//; } else { $_ = ''; } } @out;

    my @formatted = ();
    for (my $i=0; $i<@out; $i++) {
      push @formatted, sprintf($tabformat[$i], $out[$i]);
    }

    # Add comments if present
    push @formatted, '# ' . $SNP->{'Comments'} if exists $SNP->{'Comments'};

    return @formatted;
}

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

#PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP

#----------------------------------------

=head3 _value

 Usage   : _value($hashref, @keys)
 Function: This small routine returns the value of a multidim hash
           - it carefully checks if each of the keys really exists
             and otherwise returns undef
 Args    : Hashref
           List of keys
 Result  : Value of innermost key

=cut

#----------------------------------------
sub _value {
    my ($hr, @keys) = @_;
    (scalar(@keys)) or return defined $hr ? $hr : '';
    my $k = shift @keys;
    exists $hr->{$k} ? _value($hr->{$k}, @keys) : '';
}
#----------------------------------------------------------------------

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

1;
