package PPH::BLAT;
use strict;
use warnings;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::BLAT - module for BLAT routines - running blat - retrieving sequences

=head1 DESCRIPTION

Using UCSC BLAT utility to find sequences very fast

=head1 SUBVERSION

 $LastChangedDate: 2009-11-30 00:39:17 -0500 (Mon, 30 Nov 2009) $
 $LastChangedRevision: 261 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Carp qw(cluck confess);

BEGIN {
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}

use PPH::Codon;

use vars qw(@ISA %STATUS);

my %STATUS = (
    'ok'       =>  1,   # analysis successful
    'no_hit'   => -1,   # No BLAT hit found
    'gap'      => -2,   # BLAT with gaps
    'overlap'  => -3,   # BLAT alignment doesnt cover SNP
    'wrong_aa' => -4,   # Codon is not coding for one of the aa
    'SNPmut'   => -5,   # The variants can't be explained by a SINGLE nt mutation
    'codwrong' => -6,   # Codon provided is wrong;
);

#------------------------------------------------------------------------------

=head2 prot_blat

 Title   : prot_blat
 Usage   : my $PSL = BLAT::prot_blat($dbase, $seq, [$minIdentity])
 Function: returns a psl hit for a protein protein BLAT
           o only takes first hit / sorted by match
           o takes first codon which has only one substitution to code
             for aa1 / aa2
 Input   : fasf        - protein sequence file (FASTA formatted)
           minIdentity - min Identity for hit [default 90%]
 Returns : PSL object - hashref

=cut

#----------------------------------------
sub prot_blat {
    my ($fasf, $minIdentity) = @_;
    #### PPH__BLAT__prot_blat
    $minIdentity = 90 unless defined($minIdentity);

    confess "Sequence file '$fasf' not found" unless (-e $fasf);
    my $exec = "$CONFIG{BLAT} $CONFIG{BLAT_HOST} $CONFIG{BLAT_PORT} $CONFIG{BLAT_ARG} -q=prot -minIdentity=$minIdentity $fasf stdout | sort -n -r ";

    ##### Executing BLAT: $exec
    open  my $R, '-|', $exec or confess "BLAT failed";
    $_ = <$R>;
    close $R;

    return {} unless(defined($_));
    chomp;
    return {} if (/^\s*$/);
    ##### PPH__BLAT___prot_blat returns: _psl2hash($_)
    return _psl2hash($_);
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head3 _psl2hash

 Title   : _psl2hash
 Usage   : my $PSL = psl2hash($psl)
 Function: Converts a psl formatted blat output to a hash structure
           ! In case of prot -> dnax it converts T positions to protein pos
 Input   : psl - one line of psl formatted blat

=cut

#----------------------------------------
sub _psl2hash {
    my ($psl) = shift;
    ##### _psl2hash: $psl
    my (@KEY) = qw(match mismatch repmatch Ns QgapCount QgapBases TgapCount TgapBases strand Qname Qsize Qstart Qend Tname Tsize Tstart Tend blockCount blockSizes qStarts tStarts);
    my %PSLH;
    my (@PSL) = split /\s+/;

    confess "Something wrong - more PSL fields than descriptors $#PSL $#KEY\n\t$psl" if ($#PSL != $#KEY);

    for (0..$#PSL) {
        $PSLH{$KEY[$_]} = $PSL[$_];
    }

    $PSLH{'blockSizes'} = [ (split /,/, $PSLH{'blockSizes'}) ];
    $PSLH{'qStarts'}    = [ (split /,/, $PSLH{'qStarts'})    ];
    $PSLH{'tStarts'}    = [ (split /,/, $PSLH{'tStarts'})    ];

    #----------------------------------------
    # PROT - DNAX mapping
    #----------------------------------------
    if ($PSLH{'strand'} eq "++") {
        $PSLH{'strand'} = substr($PSLH{strand}, 1, 1);
        $PSLH{Tsize}  /=  3;
        $PSLH{Tstart} /= 3;
        $PSLH{Tend}   /= 3;
        $PSLH{'tStarts'} = [ map { ($_ /= 3) } @{ $PSLH{'tStarts'} } ];
    }#/

    return \%PSLH;
}
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------

=head2 SNP2refGene

 Title   : SNP2refGene
 Usage   : my $status = SNP2refGene($SNP)
 Function: This procedure retrives the nucleotides for a non-synonymous SNP
           at a given position and evaluates the following info:
           o distance from intron / exon junction (+ 5' exon), (- 3' exon)
           o CpG
           This routine heavily depends on the refGene.prot.nt.fas.pl script
           which prepares a file this routine uses to find the corresponding
           sequence - it uses BLAT for this
 Input   : $SNP     - object of type SNP - fields required:
           $SNP = {
            'aa1' => 'Q',
            'aa2' => 'H',
            'Pos' => 633,
            'Seqfile' => '../..seq'
            'Gene' => {
                     'cod1' => '', 'cod2' => '' # OPTIONAL
                     }
           }

 Output  : $status  - 0 - no hit found / 1 - ok
           $SNP->{'Gene'} = {
             'refSeqID'  => 'NM..',
             'prot => prot sequence (may differ from real refGene entry),
             'orf' => orf sequence  (" ")
             'cod1' => ' ', 'cod2' => ' ',
             'CodonPos' => '0|1|2', # Pos of SNP in codon
             'triplet1' => ' ', 'triplet2' => ' ', # 3 nt's with SNP in middle
             'd5junction' => '', 'd3junction' => '', # distance to exon/intron
             'CpG1' => '1|2', 'CpG2' => '1|2', # which snp is cpg which nt is affected
             }

=cut

#----------------------------------------
# $SNP = {
#          'aa1' => 'Q',
#          'aa2' => 'H',
#          'Seq' => 'MERCSR..',
#          'Pos' => 633,
#          'Seqfile' => '../..seq'
#        }
#----------------------------------------
sub SNP2refGene {
    my $SNP = shift;

    #### PPH__BLAT__SNP2refGene
    #----------------------------------------
    # Alright first try to find a refSeq
    # sequence with good hit i.e. no gaps!
    #----------------------------------------
    my $PSL;
    if ($CONFIG{MAPREFSEQ}) {
      $PSL = prot_blat($SNP->{'Seqfile'}, '90');
    }

    #  return $STATUS{'no_hit'}  unless defined($PSL);
    # no gaps
    #  return $STATUS{'gap'} unless ($PSL->{'blockCount'} == 1);
    # SNP in aligned region?
    #  return $STATUS{'overlap'} unless ($PSL->{'Qend'} > $SNP->{'Pos'});

    if (defined($PSL)) {
        undef($PSL) unless
          (exists($PSL->{'blockCount'}) and
          $PSL->{'blockCount'} == 1 and ($PSL->{'Qend'} > $SNP->{'Pos'}));
    }

    my $status = 0;
    if (defined($PSL)) {
        $status = _retrieve_gene($SNP, $PSL);
        return $status unless (defined($status) and $status == $STATUS{'ok'});
    }
    #----------------------------------------

    #----------------------------------------
    # Get possible codons for aa and find the
    # mutated position
    # unless Codons are given
    #----------------------------------------
    $status = _assign_codons($SNP, $PSL);
    return $status unless ($status == $STATUS{'ok'});

    #----------------------------------------
    # Test for CpG
    #----------------------------------------
    # Possible scenarios:        Codpos
    #  AAC GTT TTT | AAG CTT TTT 0
    #  AAA CGT TTT | AAA GCT TTT 0 1
    #  AAA ACG TTT | AAA AGC TTT 1 2
    #  AAA AAC GTT | AAA AAG CTT   2
    #----------------------------------------

    my @genecod;                # Codons of gene (if available)
    if (exists($SNP->{'Gene'}) and exists($SNP->{'Gene'}{'orf'})) {
        (@genecod) = $SNP->{'Gene'}{'orf'} =~ m/\|?.\|?.\|?./g;
    }

    # Codon 1
    my $dna = "";
    if (@genecod) {
        $dna  = (($SNP->{'Gene'}{'aa_pos'} > 1) ? $genecod[($SNP->{'Gene'}{'aa_pos'} - 2)] : "XXX");
        $dna .= $SNP->{'Gene'}{'cod1'};
        $dna .= $genecod[($SNP->{'Gene'}{'aa_pos'})];
    } else {
        $dna = join '', "XXX", $SNP->{'Gene'}{'cod1'}, "XXX";
    }
    $dna =~ s/\|//g;
    $dna = substr($dna, (2 + $SNP->{'Gene'}{'CodonPos'}), 3); # Retrieve nt surrounding SNP
    ##### |    DNA: $dna
    $SNP->{'Gene'}{'triplet1'} = $dna;
    my $cg = (index $dna, "CG");
    $SNP->{'Gene'}{'CpG1'} = $cg + 1 if ($cg >= 0);
    ##### |     cg: $cg

    # Codon 2
    if (@genecod) {
        $dna  = (($SNP->{'Gene'}{'aa_pos'} > 1) ? $genecod[($SNP->{'Gene'}{'aa_pos'} - 2)] : "XXX");
        $dna .= $SNP->{'Gene'}{'cod2'};
        $dna .= $genecod[($SNP->{'Gene'}{'aa_pos'})];
    } else {
        $dna = join '', "XXX", $SNP->{'Gene'}{'cod2'}, "XXX";
    }
    $dna =~ s/\|//g;
    $dna = substr($dna, (2 + $SNP->{'Gene'}{'CodonPos'}), 3); # Retrieve nt surrounding SNP
    ##### |    DNA: $dna
    $SNP->{'Gene'}{'triplet2'} = $dna;
    $cg = (index $dna, "CG");
    $SNP->{'Gene'}{'CpG2'} = $cg + 1 if ($cg >= 0);
    ##### |    cg:  $cg
    #----------------------------------------

    #----------------------------------------
    # Find closest Exon/Intron boundary
    #----------------------------------------

    if (@genecod) {
        $dna = join '', @genecod[($SNP->{'Gene'}{'aa_pos'} - 1)..$#genecod];
        my $i = $SNP->{'Gene'}{'CodonPos'} + 1;
        $dna =~ s/^(\|?\w){$i}//;
        ##### |    DNA: $dna
        my $d = index $dna, "|";
        $SNP->{'Gene'}{'d3junction'} = $d unless ($d < 0);

        $dna = reverse ( split '', ( join '', @genecod[0..($SNP->{'Gene'}{'aa_pos'} - 1)]));
        $i = 3 - $SNP->{'Gene'}{'CodonPos'};
        $dna =~ s/(\|?\w){$i}//;
        ##### |    DNA: $dna
        $d = index $dna, "|";
        $SNP->{'Gene'}{'d5junction'} = $d unless ($d < 0);
    }
    #----------------------------------------

    return $STATUS{'ok'};
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 _retrieve_gene

 Title   : _retrieve_gene
 Usage   : my $status = _retrieve_gene($SNP, $PSL);
 Function: This routine greps the sequences of the refseq entry and extracts
           also the codon of the mutation
 Input   : $SNP - hashref
           $PSL - hashref
 Output  : $SNP datatype extended by 'Gene' field
           $SNP->{'Gene'}{'refSeqID'}
           $SNP->{'Gene'}{'prot'}
           $SNP->{'Gene'}{'orf'}
           $SNP->{'Gene'}{'mismatch'}
           $SNP->{'Gene'}{'genomic_aa'} # 1|2 : the variant found in the genome

=cut

#----------------------------------------
sub _retrieve_gene {
    my $SNP = shift;
    my $PSL = shift;
    ##### PPH__BLAT__retrieve_gene: $PSL

    open my $F, '-|', "grep -A 2 '^>$PSL->{Tname} ' $CONFIG{REFGENE}"
        or confess "Can't grep refGene sequence";

    my $Tid   =    (<$F>);
    $Tid =~ s/^>(\S+)/$1/;                # The RefSeq ID (mRNA)
    my $Tprot = uc (<$F>); chomp($Tprot); # The protein sequence
    my $Tdna  = uc (<$F>); chomp($Tdna ); # The genomic sequence
    close $F;

    #----------------------------------------
    # Target codons
    # incl. '|' for exon/intron junctions
    #----------------------------------------
    my (@Tcod)= $Tdna =~ m/\|?.\|?.\|?./g;

    # Correct postion if hit doesn't start at first postion
    my $Taa_pos = $PSL->{'Tstart'} - $PSL->{'Qstart'} + $SNP->{'Pos'};
    my $Tcod = $Tcod[$Taa_pos - 1]; # The codon without '|'

    return $STATUS{'codwrong'} unless defined($Tcod);

    $Tcod =~ s/\|//g;
    ##### |    Tcod: $Tcod
    #----------------------------------------

    #----------------------------------------
    # Check if aa in target is one of SNPs'aa
    #----------------------------------------
    my $Taa = substr($Tprot, ($Taa_pos - 1), 1);
    my $Taa_is_aa;         # Determine which variant we find in genome
  SWITCH: {
        ($Taa eq $SNP->{'aa1'}) and do { $Taa_is_aa = 1; last SWITCH};
        ($Taa eq $SNP->{'aa2'}) and do { $Taa_is_aa = 2; last SWITCH};

        # check: ($Taa eq $SNP->{'aa1'}) or ($Taa eq $SNP->{'aa2'})
        return $STATUS{'wrong aa'};
    }
    #----------------------------------------

    #----------------------------------------
    # Basic assignment is o.k.
    #----------------------------------------
    $SNP->{'Gene'}{'refSeqID'} = $Tid;
    $SNP->{'Gene'}{'prot'} = $Tprot;
    $SNP->{'Gene'}{'orf'}  = $Tdna;
    $SNP->{'Gene'}{'mismatch'} = $PSL->{'mismatch'};
    $SNP->{'Gene'}{'genomic_aa'} = $Taa_is_aa;
    $SNP->{'Gene'}{'aa_pos'} = $Taa_pos;
    $SNP->{'Gene'}{'genomic_cod'} = $Tcod;
    if ($Taa_is_aa == 1) {
        return $STATUS{'wrong aa'} if (exists($SNP->{'Gene'}{'cod1'}) and $SNP->{'Gene'}{'cod1'} != $Tcod);
        $SNP->{'Gene'}{'cod1'} = $Tcod
    } else {
        return $STATUS{'wrong aa'} if (exists($SNP->{'Gene'}{'cod2'}) and $SNP->{'Gene'}{'cod2'} != $Tcod);
        $SNP->{'Gene'}{'cod2'} = $Tcod
    }
    #----------------------------------------
    ##### PPH__BLAT__retrieve_gene returns: $SNP->{'Gene'}
    return $STATUS{'ok'};
}
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------

=head2 _assign_codons

 Title   : _assign_codons
 Usage   : my $status = _assign_codons($SNP, $PSL);
 Function: Retrieve the correct codon for from the refseq sequence and assign
           a valid codon for the given mutation
 Input   : $SNP - hashref
           $PSL - hashref
 Output  : $SNP datatype extends the 'Gene' field

=cut

#----------------------------------------
sub _assign_codons {
    my $SNP = shift;
    my $PSL = shift;

    my (@Var1cod) = ();		# Codons for variant1
    my (@Var2cod) = ();		# Codons for variant2
    ##### PPH__BLAT__assign_codons

    #----------------------------------------
    # Not both codons provided
    #----------------------------------------
    unless (exists($SNP->{'Gene'}{'cod1'}) and exists($SNP->{'Gene'}{'cod2'})) {
        #----------------------------------------
        # Genome mapping successful?
        # i.e. at least on codon is known
        #----------------------------------------
        if (exists($SNP->{'Gene'}{'genomic_aa'})) {
            if ($SNP->{'Gene'}{'genomic_aa'} == 1) {
                @Var1cod = $SNP->{'Gene'}{'cod1'};
                @Var2cod = PPH::Codon::revtranslate($SNP->{'aa2'});
            } else {
                @Var1cod = PPH::Codon::revtranslate($SNP->{'aa1'});
                @Var2cod = $SNP->{'Gene'}{'cod2'};
            }
        }
        #----------------------------------------
        # No Codons provided take all possible
        #----------------------------------------
        else {
            @Var1cod = PPH::Codon::revtranslate($SNP->{'aa1'});
            @Var2cod = PPH::Codon::revtranslate($SNP->{'aa2'});
        }
    }
    #----------------------------------------
    # Both codons have been provided
    #----------------------------------------
    else {
        ##### |    codons provided
        @Var1cod = $SNP->{'Gene'}{'cod1'};
        @Var2cod = $SNP->{'Gene'}{'cod2'};
    }
    #----------------------------------------

    my $CodonPos = 0;		# Codon position where mutation occurs
    my $Mcod = "";		# Mcod = mutation of Tcod to code for
				# other aa
    my $Nmut = 0;               # No of mutations per codon pair
    my ($cod1, $cod2);

  COD_FOUND:
    foreach my $c1 (@Var1cod) {
        foreach my $c2 (@Var2cod) {
            for my $p (0..2) {
                if (substr($c1, $p,1) ne substr($c2,$p,1)) {
                    $Nmut++; $CodonPos = $p;
                }
            }
            if ($Nmut == 1) {
                ($cod1, $cod2) = ($c1, $c2); last COD_FOUND;
            }
            ;
            $Nmut = 0;
        }
    }

    ##### |    aa1, cod1, aa2, cod2: $SNP->{'aa1'}, $cod1, $SNP->{'aa2'}, $cod2

    if ($Nmut != 1) {
        ##### |    Codon / AA is not correct
        return $STATUS{'SNPmut'};
    }
    #----------------------------------------

    #----------------------------------------
    # Codons found
    #----------------------------------------
    $SNP->{'Gene'}{'CodonPos'} = $CodonPos;
    $SNP->{'Gene'}{'nt1'}  = substr($cod1, $CodonPos, 1);
    $SNP->{'Gene'}{'nt2'}  = substr($cod2, $CodonPos, 1);
    $SNP->{'Gene'}{'Transversion'} = PPH::Codon::is_transversion($SNP->{'Gene'}{'nt1'}, $SNP->{'Gene'}{'nt2'});
    $SNP->{'Gene'}{'cod1'} = $cod1;
    $SNP->{'Gene'}{'cod2'} = $cod2;
    #----------------------------------------

    return $STATUS{'ok'};
}
#------------------------------------------------------------------------------



#------------------------------------------------------------------------------
1;
