package PPH::Codon;
use strict;
use warnings;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::Codon - Package to translate codons to protein

=head1 DESCRIPTION

 - adopted from BioPerl : Bio::Tools::CodonTable

=head1 SUBVERSION

 $LastChangedDate: 2009-11-28 00:15:09 -0500 (Sat, 28 Nov 2009) $
 $LastChangedRevision: 254 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Carp qw(cluck confess);

BEGIN {
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}

use vars qw(@ISA $TABLE $START $CODON $TRCOL);

#------------------------------------------------------------------------------
# Creating the Tables
#----------------------------------------
use constant CODONSIZE => 3;
use constant GAP       => '-';
use constant CODONGAP  => GAP x CODONSIZE;

#----------------------------------------
# Standard codon table
#----------------------------------------
my $TABLE = "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG";
my $START = "---M---------------M---------------M----------------------------";
#----------------------------------------

#----------------------------------------
# Code for nucleotides
#----------------------------------------
my %BASE   = (
    'Purine' => [ 'A', 'G' ],
    'Pyrimidine' => [ 'T', 'C', 'U' ]
);
my %NTCODE = (
    'A' => [ 'A' ],
    'C' => [ 'C' ],
    'G' => [ 'G' ],
    'T' => [ 'T' ],
    'M' => [ 'A', 'C' ],
    'R' => [ 'A', 'G' ],
    'W' => [ 'A', 'T' ],
    'S' => [ 'C', 'G' ],
    'Y' => [ 'C', 'T' ],
    'K' => [ 'G', 'T' ],
    'V' => [ 'A', 'C', 'G' ],
    'H' => [ 'A', 'C', 'T' ],
    'D' => [ 'A', 'G', 'T' ],
    'B' => [ 'C', 'G', 'T' ],
    'X' => [ 'A', 'C', 'G', 'T' ],
    'N' => [ 'A', 'C', 'G', 'T' ],
);

my %NTREVERSE = (
    'A' => 'T',
    'C' => 'G',
    'G' => 'C',
    'T' => 'A',
    'M' => 'K',
    'R' => 'Y',
    'W' => 'W',
    'S' => 'S',
    'Y' => 'R',
    'K' => 'M',
    'V' => 'B',
    'H' => 'D',
    'D' => 'H',
    'B' => 'V',
    'X' => 'X',
    'N' => 'N'
);

my %AACODE = (
    'F' => [ 'F' ],
    'S' => [ 'S' ],
    'T' => [ 'T' ],
    'N' => [ 'N' ],
    'K' => [ 'K' ],
    '*' => [ '*' ],
    'E' => [ 'E' ],
    'Y' => [ 'Y' ],
    'V' => [ 'V' ],
    'Z' => [ 'E', 'Q' ],
    'Q' => [ 'Q' ],
    'M' => [ 'M' ],
    'C' => [ 'C' ],
    'L' => [ 'L' ],
    'A' => [ 'A' ],
    'W' => [ 'W' ],
    'X' => [ 'X' ],
    'P' => [ 'P' ],
    'B' => [ 'D', 'N' ],
    'H' => [ 'H' ],
    'D' => [ 'D' ],
    'R' => [ 'R' ],
    'I' => [ 'I' ],
    'G' => [ 'G' ],
    'U' => [ 'U' ]
);

my $VALID_AA = (join '', keys %AACODE)."*";

my %AANAME = (
    'Leu' => 'L',
    'Lys' => 'K',
    'S' => 'Ser',
    'T' => 'Thr',
    'N' => 'Asn',
    'K' => 'Lys',
    'Gln' => 'Q',
    'E' => 'Glu',
    'Y' => 'Tyr',
    'His' => 'H',
    'Z' => 'Glx',
    'Asx' => 'B',
    'Asn' => 'N',
    'Trp' => 'W',
    'Cys' => 'C',
    'Asp' => 'D',
    'Gly' => 'G',
    'Phe' => 'F',
    'Met' => 'M',
    'W' => 'Trp',
    'Val' => 'V',
    'Thr' => 'T',
    'B' => 'Asx',
    'Ter' => '*',
    'H' => 'His',
    'Pro' => 'P',
    'D' => 'Asp',
    'I' => 'Ile',
    'G' => 'Gly',
    'U' => 'Sec',
    'F' => 'Phe',
    'Glu' => 'E',
    '*' => 'Ter',
    'Xaa' => 'X',
    'V' => 'Val',
    'Q' => 'Gln',
    'M' => 'Met',
    'Ser' => 'S',
    'C' => 'Cys',
    'Ala' => 'A',
    'Arg' => 'R',
    'L' => 'Leu',
    'A' => 'Ala',
    'Tyr' => 'Y',
    'X' => 'Xaa',
    'P' => 'Pro',
    'Glx' => 'Z',
    'R' => 'Arg',
    'Ile' => 'I',
    'Sec' => 'U'
);
#----------------------------------------

#----------------------------------------
# Creating the Codon table
#----------------------------------------
my ($CODON, $TRCOL) = ( {}, {} );
my @NUCS = qw(t c a g);
my $pos = 0;
for my $i (@NUCS) {
    for my $j (@NUCS) {
        for my $k (@NUCS) {
            my $codon = "$i$j$k";
            $CODON->{$codon} = $pos;
            $TRCOL->{$pos}   = $codon;
            $pos++;
        }
    }
}
#----------------------------------------

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 translate

  Title   : translate
  Usage   : PPH::Codon::translate($seq);
  Function: Returns a string in one letter amino acid code from nt input.
            The input can be of any length!

            Returns 'X' for unknown codons and codons that code for more than
            one aa.
  Input   : nt sequence
  Returns : corresponding aa sequence

=cut

#----------------------------------------
sub translate {
    my ($seq) = shift;
    #### PPH__Codon__translate: $seq

    #----------------------------------------
    # Seq is partial?
    #----------------------------------------
    my $partial = 0;
    $partial = 2 if length($seq) % CODONSIZE == 2;
    #----------------------------------------

    $seq = lc $seq;
    $seq =~ tr/u/t/;

    my $protein = "";

    if ($seq =~ /[^actg]/ ) {   #ambiguous chars
        for (my $i = 0; $i < (length($seq) - (CODONSIZE-1)); $i+= CODONSIZE) {
            my $triplet = substr($seq, $i, CODONSIZE);

            if ( $triplet eq CODONGAP ) {
                $protein .= GAP;
            } elsif (exists $CODON->{$triplet}) {
                $protein .= substr($TABLE, $CODON->{$triplet},1);
            } else {
                $protein .= _translate_ambiguous_codon($triplet, $partial);
            }
        }
    } else {                    # simple, strict translation
        for (my $i = 0; $i < (length($seq) - (CODONSIZE -1)); $i+=CODONSIZE) {
            my $triplet = substr($seq, $i, CODONSIZE);

            if ( $triplet eq CODONGAP ) {
                $protein .= GAP;
            }
            if (exists $CODON->{$triplet}) {
                $protein .= substr($TABLE, $CODON->{$triplet}, 1);
            } else {
                $protein .= 'X';
            }
        }
    }
    if ($partial == 2) {        # 2 overhanging nucleotides
        my $triplet = substr($seq, ($partial -4)). "n";
        if ( $triplet eq CODONGAP ) {
            $protein .= GAP;
        } elsif (exists $CODON->{$triplet}) {
            my $aa = substr($TABLE, $CODON->{$triplet},1);
            $protein .= $aa;
        } else {
            $protein .= _translate_ambiguous_codon($triplet, $partial);
        }
    }
    ##### |    PPH__Codon__translate returns: $protein
    return $protein;
}
#------------------------------------------------------------------------------


#----------------------------------------

=head2 _translate_ambiguous_codon

 Title   : _translate_ambiguous_codon
 Usage   : $aa = _translate_ambiguous_codon($triplet, $partial);
 Function:
 Returns : single amino acid if possible - otherwise place holders
 Args    : the DNA triplet, and code for partial (0 = all, 2 = partial)

=cut

#----------------------------------------
sub _translate_ambiguous_codon {
    my ($triplet, $partial) = @_;
    ##### PPH__Codon__translate_ambiguous_codon: $triplet, $partial
    $partial ||= 0;

    my $aa;
    my @codons = _unambiquous_codons($triplet); # get all possible triplets
    my %aas =();
    foreach my $codon (@codons) { # get them translated
        $aas{substr($TABLE,$CODON->{$codon}, 1)} = 1;
    }
    my $count = scalar keys %aas;
    if ( $count == 1 ) {        # all codons coding for 1 aa
        $aa = (keys %aas)[0];
    } elsif ( $count == 2 ) {	# well, any similar aa?
        if ($aas{'D'} and $aas{'N'}) {
            $aa = 'B';
        } elsif ($aas{'E'} and $aas{'Q'}) {
            $aa = 'Z';
        } else {                # Out of options
            $partial ? ($aa = '') : ($aa = 'X');
        }
    } else {
        $partial ? ($aa = '') :  ($aa = 'X');
    }
    return $aa;
    ###### |    PPH__Codon__translate_ambiguous_codon returns: $aa
}
#----------------------------------------



#----------------------------------------

=head2 _unambiquous_codons

 Title   : _unambiquous_codons
 Usage   : @codons = _unambiquous_codons('ACN')
 Function:
 Example :
 Returns : array of strings (one letter unambiguous amino acid codes)
 Args    : a codon = a three IUPAC nucleotide character string

=cut

#----------------------------------------
sub _unambiquous_codons{
    ##### PPH Codon _unambiquous_codons
    my ($value) = @_;
    my @nts = ();
    my @codons = ();
    my ($i, $j, $k);
    @nts = map { $NTCODE{uc $_} }  split(//, $value);
    for my $i (@{$nts[0]}) {
        for my $j (@{$nts[1]}) {
            for my $k (@{$nts[2]}) {
                push @codons, lc "$i$j$k";
            }
        }
    }
    ##### |    PPH Codon _unambiquous_codons returns: @codons
    return @codons;
}
#----------------------------------------

#------------------------------------------------------------------------------

=head2 revcomp

 Title   : revcomp
 Usage   : PPH::Codon::revcomp('CAT')
 Function: returns reverse complement of a sequence

 Example : PPH::Codon::revcomp('CAT')
 Returns : reverse complement
 Args    : sequence

=cut

#----------------------------------------
sub revcomp {
    my ($seq) = shift;
    #### PPH Codon revcomp: $seq

    return join '', (map { exists($NTREVERSE{$_}) ? $NTREVERSE{$_} : '?' } reverse ( split '', $seq ));
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 revtranslate

 Title   : revtranslate
 Usage   : PPH::Codon::revtranslate('F')
 Function: returns possible codons for an amino acid

           Returns an empty string for unknown amino acid
           codes. '*' is used for terminator.

           By default, the output codons are shown in DNA.  If the
           output is needed in RNA (tr/t/u/), add a second argument
           'RNA'.

 Example : PPH::Codon::revtranslate('G')
 Returns : An array of three lower case letter strings i.e. codons
 Args    : amino acid

=cut

#----------------------------------------
sub revtranslate {
    my ($aacode) = uc (shift);
    #### PPH__Codon__revtranslate: $aacode
    my (@aas,  $p);
    my (@codons) = ();

    if (length($aacode) == 3 ) {
        $aacode = $AANAME{ (ucfirst lc $aacode) };
    }

    if ( defined $aacode && $aacode =~ /[$VALID_AA]/ && length($aacode) == 1 ) {

        (@aas) = @{$AACODE{$aacode}};

        foreach my $aa (@aas) {

            $aa = '\*' if $aa eq '*';

            while ($TABLE =~ m/$aa/g) {
                $p = pos $TABLE;
                push @codons, uc $TRCOL->{($p - 1)};
            }
        }
    }
    return @codons;
    ##### |   PPH__Codon__revtranslate returns: @codons
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 is_transversion

 Title   : is_transversion
 Usage   : my $test = is_transversion($nt1, $nt2);
 Function: Return TRUE (1) if substitution is a transversion
 Input   : $nt1
           $nt2
 Output  : TRUE is transversion

=cut

#----------------------------------------

sub is_transversion {
    my ($nt1, $nt2) = @_;
    #### PPH__Codon__is_transversion: $nt1, $nt2

    (grep /$nt1/, @{$BASE{Purine}}) and (grep /$nt2/, @{$BASE{Pyrimidine}}) and return 1;
    (grep /$nt2/, @{$BASE{Purine}}) and (grep /$nt1/, @{$BASE{Pyrimidine}}) and return 1;

    return 0;
}
#------------------------------------------------------------------------------


#PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP
1;
