package PPH::Data;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::Data

=head1 DESCRIPTION

Contains biological data and other data objects for PolyPhen

=head1 AUTHOR

StS

=head1 SUBVERSION

 $LastChangedDate: 2010-11-27 20:35:38 -0500 (Sat, 27 Nov 2010) $
 $LastChangedRevision: 338 $
 $LastChangedBy: ivan $

=head1 ROUTINES / VARIABLES

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Carp qw(confess);
use strict;
use warnings;

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 0.1;
    @ISA           = qw(Exporter);
    @EXPORT        = qw();
    @EXPORT_OK     = qw(
      %Effect
      %BONDS %SITES1 %SITES2
      @PhiPsiMap %MaxAccessibility %AA_Volume
      @Non_biol_ligands @Predictions
      %Propensities @GARBAGE_FILES
      %BLOSUM62_F
    );
    %EXPORT_TAGS = ();
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}
our @EXPORT_OK;

#----------------------------------------

=head2 @Predictions

 Function: Possible predictions set by PPH::Rules

=cut

#----------------------------------------
our @Predictions = (
    'unknown',
    'benign',
    'possibly damaging',
    'probably damaging',
);
#----------------------------------------

#----------------------------------------------------------------------

=head2 %Effect

 Function: Predicted effect of a substituion set by PPH::Rules

=cut

#----------------------------------------

our %Effect = (
    '1' =>   {
        Name  => 'structural effect',
    },

    '1.1' => {
        Name  => 'buried site',
    },

    '1.1.1'  => {
        Name  => 'hydrophobicity',
        Comment  => 'Hydrophobicity change at buried site',
    },

    '1.1.2'  => {
        Name  => 'overpacking',
        Comment  => 'Overpacking at buried site',
    },

    '1.1.3'  => {
        Name  => 'cavity',
        Comment  => 'Cavity creation at buried site',
    },

    '1.2' => {
        Name  => 'bond formation',
        Comment  => 'Disruption of annotated bond formation site',
    },

    '1.2.1'  => {
        Name  => 'covalent bond',
    },

    '1.2.1.1'   => {
        Name  => 'disulphide',
    },

    '1.2.1.2'   => {
        Name  => 'thioesther',
    },

    '1.2.1.3'   => {
        Name  => 'thioether',
    },

    '1.2.2'  => {
        Name  => 'non-covalent',
    },

    '1.2.2.1'   => {
        Name  => 'hydrogen',
    },

    '1.2.2.2'   => {
        Name  => 'salt bridge (electrostatic)',
    },



    '2'   => {
        Name  => 'functional effect',
    },

    '2.1' => {
        Name  => 'indirect',
        Comment  => 'Contact with functional site',
    },

    '2.2' => {
        Name  => 'functional site',
        Comment  => 'Disruption of annotated functional site',
    },

    '2.2.1'  => {
        Name  => 'signal peptide',
    },

    '2.2.2'  => {
        Name  => 'transmembrane',
        Comment  => 'Improper substitution in the transmembrane region',
    },

    '2.2.3'  => {
        Name  => 'ligand binding',
        Comment  => 'Disruption of ligand binding site',
    },

    '2.2.4'  => {
        Name  => 'protein interaction',
    },


    '3'   => {
        Name  => 'unknown',
        Comment  => 'Derived from multiple alignment',

    },


);
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 UniProt Annotations

 Function: The given annotations extracted from UniProt will be used to
           calculate e.g. transmembrane PHAT scores or distances to
           sites such as catalytic site or ligand

=head3 BONDS

=cut

our %BONDS;
@BONDS{  qw(DISULFID THIOLEST THIOETH) } = ();


=head3 %REGIONS

=cut

our %REGIONS;
@REGIONS{qw(TRANSMEM TRANSMEM_PRED COIL_PRED SIGNAL_PRED SIGNAL PROPEP) } = ();

=head3 %SITES1

 Function: Critical sites - c.f. PPH::Seq::get_uniprot_annotation
           The sites are not in %REGIONS

=cut

our %SITES1;
@SITES1{ qw(BINDING ACT_SITE LIPID METAL) } = ();

=head3 %SITES2

=cut

our %SITES2;
@SITES2{ qw(SITE DISULFID THIOLEST THIOETH MOD_RES CARBOHYD SE_CYS) } = ();

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 Constants for PPH::3D

=head3 @PhiPsiMap

 Function: needed for mapping phipsi angles (c.f. PPH::3D)

=cut

#----------------------------------------
our @PhiPsiMap = qw(

                       bBBBBBBBBBBbbbbb.....pppppgggggbbbbb
                       bBBBBBBBBBBBbbbbb......ggggggggbbbbb
                       bBBBBBBBBBBBbbbbbbb....ggggggggbbbbb
                       bbBBBBBBBBBBBbbbbbb....ggggggggbbbbb
                       bbBBBBBBBBBBBBbbbbb....ggggggggbbbbb
                       bbBBBBBBBBBBBbbbbbb....gggggggggbbbb
                       bbbBBBBBBBBBbbbbbbb...llllllgggggbbb
                       bbbbBBBBBBBbbbbbbbblllllllllgggggbbb
                       bbbbbBbbBbbbbbbbbbblllllllllgggggbbb
                       bbbbbbbbbbbbbbbbbbblllllllllgggggbbb
                       bbbbbbbbbbbbbbbbbbblllllllllgggggbbb
                       bbbbbbbbbbbbbbb....llllllllllggggbbb
                       bbbbbbbbbbbbbbb....llllllllllggggbbb
                       aaaaaaaaaaaaaaa....llllLlllllggggbbb
                       aaaaaaaaaaaaaaa....llllLLllllggggbbb
                       aaaaaaaAaaaaaaaa...llllllllllggggbbb
                       aaaaaaAAAAaaaaaaa...lllllllllggggbbb
                       aaaaaaAAAAAaaaaaa...lllllllllggggbbb
                       aaaaaAAAAAAAaaaaaa..lllllllllggggbbb
                       aaaaaaAAAAAAAaaaaaa.lllllllllggggbbb
                       aaaaaaAAAAAAAAaaaaa.lllllllllggggbbb
                       aaaaaaaAAAAAAAaaaaaa....gggggggggbbb
                       aaaaaaaaAAAAAAAaaaaa...ggggggggggbbb
                       aaaaaaaaaAAAAAAaaaaaa..ggggggggggbbb
                       aaaaaaaaaaaAAAAaaaaaa..ggggggggggbbb
                       aaaaaaaaaaaaaaaaaaaaa..ggggggggggbbb
                       aaaaaaaaaaaaaaaaaaaaa.gggggggggggbbb
                       .aaaaaaaaaaaaaaaaaaaa.gggggggggggggg
                       gggaaaaaaaaaaaaaaaaa..gggggggggggggg
                       gggbbbbbbbbbbbbbb.....gggggggggggggg
                       bbbbbbbbbbbbbb.......pppppgggggggggg
                       bbbbbbbbbbbbbb.......pppppgggggggggg
                       bbbbbbbbbbbbbbb......ppPpppggggggggg
                       bbbbbbbbbbbbbbb......ppPpppggggggbbb
                       bbbbbbbbbbbbbbb......ppPPppggggggbbb
                       bbbbbbbbbbbbbbb......ppppppggggggbbb

               );

#----------------------------------------

=head3 %MaxAccessibility

 Function: Maximum accessibility of residues

=cut

#----------------------------------------
our %MaxAccessibility
    = (
        'A' => 103, 'C' => 105, 'D' => 164, 'E' => 187,
        'F' => 169, 'G' =>  87, 'H' => 170, 'I' => 133,
        'K' => 197, 'L' => 137, 'M' => 157, 'N' => 160,
        'P' => 133, 'Q' => 181, 'R' => 234, 'S' => 125,
        'T' => 139, 'V' => 120, 'W' => 195, 'Y' => 181
    );
#----------------------------------------

#----------------------------------------

=head3 %AA_Volume

 Function: Volumina of AA

=cut

#----------------------------------------
our %AA_Volume
    = (
        'A' => 88, 'C' => 108, 'D' => 111, 'E' => 138, 'F' => 190,
        'G' => 60, 'H' => 153, 'I' => 167, 'K' => 168, 'L' => 167,
        'M' => 163,'N' => 114, 'P' => 112, 'Q' => 144, 'R' => 173,
        'S' => 89, 'T' => 116, 'V' => 140, 'W' => 227, 'Y' => 193
    );
#----------------------------------------

#----------------------------------------

=head3 @Non_biol_ligands

 Function: Ligands which are not considered for distance measurments

=cut

#----------------------------------------
our @Non_biol_ligands = qw( PO4 PI SO4 SUL CL BR NO3 SCN NH4 K NA LI MG DOD
			    NAG MAN GOL SO4 CL CO3 FS4 );
#----------------------------------------

#----------------------------------------

=head3 %Propensities

=cut

our %Propensities
    = (
        A => [  0.59,   0.06,  -0.03,  -0.20,  -0.23,  -0.18,   0.18 ],
        R => [ -2.99,  -1.14,   0.02,   0.33,   0.47,   0.25,  -0.31 ],
        N => [ -1.48,  -0.55,  -0.27,   0.06,   0.26,   0.45,   0.39 ],
        D => [ -1.35,  -0.69,  -0.33,   0.10,   0.28,   0.42,   0.50 ],
        C => [  0.61,   0.52,   0.51,  -0.02,  -0.75,  -1.80,  -2.00 ],
        Q => [ -1.55,  -0.58,  -0.25,   0.06,   0.40,   0.44,  -0.17 ],
        E => [ -2.13,  -0.98,  -0.35,  -0.01,   0.40,   0.58,   0.28 ],
        G => [  0.31,  -0.23,  -0.07,  -0.21,  -0.12,   0.05,   0.68 ],
        H => [ -1.09,  -0.11,   0.25,   0.38,   0.10,  -0.18,  -0.54 ],
        I => [  0.87,   0.54,   0.09,  -0.04,  -0.79,  -1.29,  -2.00 ],
        L => [  0.59,   0.61,   0.28,  -0.11,  -0.76,  -1.21,  -1.42 ],
        K => [ -4.06,  -2.28,  -1.22,  -0.08,   0.60,   0.75,   0.39 ],
        M => [  0.53,   0.61,  -0.01,  -0.11,  -0.40,  -1.17,  -0.62 ],
        F => [  0.47,   0.64,   0.36,  -0.01,  -0.86,  -1.31,  -1.75 ],
        P => [ -0.98,  -0.46,  -0.28,  -0.03,   0.11,   0.41,   0.74 ],
        S => [ -0.29,  -0.21,   0.04,  -0.11,  -0.01,   0.21,   0.49 ],
        T => [ -0.49,  -0.16,  -0.11,   0.10,   0.23,   0.09,  -0.15 ],
        W => [ -0.29,   0.53,   0.63,   0.19,  -0.52,  -1.42,  -2.54 ],
        Y => [ -0.76,   0.27,   0.53,   0.43,  -0.22,  -0.95,  -1.87 ],
        V => [  0.69,   0.50,   0.19,  -0.13,  -0.46,  -1.13,  -1.76 ]
    );
#----------------------------------------

#------------------------------------------------------------------------

#----------------------------------------

=head2 blastmatrix

 Usage   : PPH::Data->blastmatrix($filename)
 Function: Simply Cache BLAST matrix
 Args    : filename of BLAST matrix
 Returns : hash of matrix

=cut

#----------------------------------------
sub blastmatrix {
    my $file = shift;
    my (@alph, $i, %matr);

    open my $MAT, '<', $file or confess "Can't read matrix from $file";

    while (<$MAT>) {
        (/^\#/) and next;

        if ( /^\s+(\w.*)$/ ) {
            @alph = (split);    ### leading space is lost only for $_
        } elsif ( /^[A-Z\*] / ) {
            my($let, @row) = (split);
            foreach $i (0..($#row)) {
                $matr{$let}{$alph[$i]} = $row[$i];
            }
        }                       # end elsif
    }
    close $MAT;
    return %matr;
}
#----------------------------------------

# IAA: Weight BLOSUM62_F matrix by column totals, i.e. convert to conditional probabilities

my @BLA = qw( A R N D C Q E G H I L K M F P S T W Y V );

# BLOSUM62 raw frequencies = fij pairs (off-diagonals = 2*fij)
# Source: http://blocks.fhcrc.org/blocks/uploads/blosum/blosum.tar.gz
#         (file: blosum62.out)
my @BL62F = qw(
 26782.81
  5848.09  22123.18
  4857.14   4930.09  17616.72
  5400.73   3953.21   9269.63  26504.98
  3962.65    980.88   1092.50    994.42  14864.71
  4794.70   6194.03   3813.41   4106.64    770.86   9131.13
  7444.98   6711.13   5505.96  12248.73    955.29   8817.09  20100.50
 14491.39   4291.00   7124.25   6284.36   1917.42   3409.41   4829.16  47098.64
  2759.96   3091.55   3563.41   2376.71    572.50   2613.56   3405.63   2387.39  11561.82
  7943.72   3098.65   2477.68   3076.61   2730.06   2220.18   3037.90   3450.79   1447.42  22975.42
 11009.91   6028.42   3411.38   3787.90   3907.74   4030.19   4990.96   5198.87   2459.23  28361.76  46272.78
  8338.88  15533.06   6080.29   6092.99   1248.97   7716.37  10296.37   6327.07   2958.56   3900.98   6138.15  20074.90
  3341.90   2001.08   1319.20   1156.88    939.80   1843.69   1692.11   1825.96    953.44   6249.52  12282.50   2264.38   5042.88
  4076.59   2321.82   1868.88   1894.29   1280.49   1351.88   2122.42   2984.24   2019.30   7589.45  13492.91   2364.01   2965.57  22771.30
  5374.35   2386.63   2143.28   3083.07    899.80   2109.68   3542.26   3398.98   1190.29   2508.69   3524.79   3930.17   1017.22   1308.84  23753.47
 15579.38   5646.43   7840.40   6985.53   2599.50   4717.08   7360.29   9553.90   2753.84   4291.98   6049.14   7728.29   2132.99   2975.00   4151.82  15680.23
  9264.29   4435.93   5571.65   4724.65   2318.32   3437.77   5106.27   5446.48   1853.12   6716.05   8281.94   5847.29   2515.49   2896.40   3366.54  11712.22  15591.62
  1003.70    662.10    402.65    404.11    360.67    566.56    660.02   1015.14    377.94    901.59   1824.09    677.71    495.17   2115.89    352.63    715.91    712.02   8060.58
  3239.21   2308.18   1745.40   1491.09    862.22   1684.00   2168.88   2079.76   3790.82   3443.82   5505.85   2489.43   1423.89  10562.84   1126.86   2566.42   2346.22   2211.24  12765.13
 12627.73   3939.43   2993.54   3278.68   3390.37   2905.57   4232.85   4539.45   1616.69  29832.35  23617.94   4824.05   5761.66   6419.39   3102.50   5877.26   9070.39    886.51   3859.61  24458.47
);

# Calculate sums of fgrequencies first
my $BL62S;      # Total sum of the BL62F table (e.g.: 829541.9)
my @BL62CS;     # Array of column sums for BL62F, i.e. frequencies of each of the 20 aa residues
my @BL62idx;    # Array of indices into plain BL62F array which trace it by aa column
foreach my $i (0 .. $#BLA) {
  my $sum;
  $sum += $_ for (0 .. $i);
  push @BL62idx, $sum;
}
for my $i (0 .. $#BLA) {
  for my $j ($i .. $#BLA) {
    my $k = $BL62idx[$j] + $i;
    my $fij = $BL62F[$k];
    $fij /= 2 if $i != $j;
    $BL62CS[$i] += $fij;
    $BL62S += $fij;
  }
}

# Calculate normalized matrix
our %BLOSUM62_F; # Full 20X20 matrix of conditional probabilities
for my $i (0 .. $#BLA) {
  for my $j ($i .. $#BLA) {
    my $k = $BL62idx[$j] + $i;
    my $fij = $BL62F[$k];
    $fij /= 2 if $i != $j;
    my $pij = $fij / $BL62CS[$i];
    $BLOSUM62_F{$BLA[$i]}{$BLA[$j]} = $pij;
    $BLOSUM62_F{$BLA[$j]}{$BLA[$i]} = $pij;
  }
}

1;
