package PPH::Profile;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::Profile

=head1 DESCRIPTION

Calculate a PSIC matrix from a sequence alignment and also provide
routines to return PSIC scores for a given substitution

(former pph_psic)

=head1 SUBVERSION

 $LastChangedDate: 2011-03-29 17:25:56 -0400 (Tue, 29 Mar 2011) $
 $LastChangedRevision: 350 $
 $LastChangedBy: ivan $


=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD
use Carp qw(cluck confess);
use strict;
use warnings;

use File::Basename;
use File::Copy;

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 0.1;
    @ISA           = qw(Exporter);
    @EXPORT        = qw();
    @EXPORT_OK     = qw();
    %EXPORT_TAGS = ();
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}
our @EXPORT_OK;

use PPH::BLAST;
use PPH::Misc;
use PPH::Align;

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

=head1 ROUTINES

#----------------------------------------------------------------------

=head2 set_PSIC_profile

 Usage   : PPH::Profile->set_PSIC_profile($SNP)
 Function: Calculate PSIC score for the given SNP
           It starts a calculation of the PSIC matrix if no profile was found
 Args    : Object of type SNP
 Returns : PSIC score stored in SNP->{PSIC}

=cut

#----------------------------------------------------------------------
sub set_PSIC_profile {
    my ($SNP, $SkipPSIC) = @_;
    #### PPH__Profile__set_PSIC_profile: $SNP, $SkipPSIC

    # Create PSIC section in the SNP object to store filenames
    my %PSIC;
    $SNP->{'PSIC'}     = \%PSIC;
    $PSIC{'Blastfile'} = "$SNP->{Acc}.blast";
    $PSIC{'Psicfile'}  = "$SNP->{Acc}.prf";
    $PSIC{'Alnfile'}   = "$SNP->{Acc}.aln";

    $PSIC{'Score1'} = $PSIC{'Score2'} = $PSIC{'Delta'} = undef;
    $PSIC{'Nobs'}   = 0;
    $PSIC{'Result'} = 'undefined';

    # Attempt to calculate MULTIZ scores if MULTIZPATH option was specified
    my ($mz_found, $mz_aln, $mz_prf, $mz_score1, $mz_score2, $mz_nobs);
    if ($CONFIG{'MULTIZPATH'}) {
      $mz_aln = $PSIC{'Alnfile'};
      $mz_prf = $PSIC{'Psicfile'};
      if (PPH::Misc::locate_file($mz_aln, "$CONFIG{MULTIZPATH}/alignments") &&
          PPH::Misc::locate_file($mz_prf, "$CONFIG{MULTIZPATH}/profiles")) {
        $mz_found = 1;
        ($mz_score1, $mz_score2, $mz_nobs) =
          PPH::Profile::get_aln_scores_for_SNP($mz_prf, $SNP->{'Pos'}, $SNP->{'aa1'}, $SNP->{'aa2'});
      }
    }

    # Skip MSA pipeline completely if and only if both .aln and .prf files
    # are already present and accounted for
    my $aln_found = PPH::Misc::locate_file($PSIC{'Alnfile'},  "$CONFIG{PRECOMPATH}/alignments", "$CONFIG{SCRATCH}/alignments");
    my $prf_found = PPH::Misc::locate_file($PSIC{'Psicfile'}, "$CONFIG{PRECOMPATH}/profiles", "$CONFIG{SCRATCH}/profiles");
    unless ($aln_found && $prf_found) {
        # Profile calculation is not allowed
        if ($CONFIG{'PRFPRECOMPONLY'}) {
          if ($CONFIG{'MULTIZPATH'} && $mz_found) {
            $PSIC{'Alnfile'}  = $mz_aln;
            $PSIC{'Psicfile'} = $mz_prf;
            $PSIC{'Score1'}   = $mz_score1;
            $PSIC{'Score2'}   = $mz_score2;
            $PSIC{'Nobs'}     = $mz_nobs;
            $PSIC{'Delta'}    = $mz_nobs ? sprintf('%+.3f', $PSIC{'Score1'} - $PSIC{'Score2'}) : undef;
            $PSIC{'Result'}   = 'multiz';
            $PSIC{'Precomputed'} = [ 'alignment', 'profile' ];
            PPH::Align::ident_fij_score($SNP) if $CONFIG{'SUBSTSCORES'};
            return 1;
          } else {
            $PSIC{'Result'} = 'no precomputed';
            return 0;
          }
        }
        # Do the whole shebang (BLAST / PSIC)
        ##### |    No PSIC profile found: $PSIC{Psicfile}
        $PSIC{'Result'} = calculate_profile($SNP, $SkipPSIC);
        $PSIC{'Psicfile'} = $PSIC{'Alnfile'} = undef unless $PSIC{'Result'} eq 'calculated';
    } else {
      if      ($aln_found == 1) {
        push @{ $PSIC{'Precomputed'} }, 'alignment';
      } elsif ($aln_found == 2) {
        push @{ $PSIC{'Cached'} }, 'alignment';
      }
      if      ($prf_found == 1) {
        push @{ $PSIC{'Precomputed'} }, 'profile';
        $PSIC{'Result'} = 'precomputed';
      } elsif ($prf_found == 2) {
        push @{ $PSIC{'Cached'} }, 'profile';
        $PSIC{'Result'} = 'cached';
      }
      delete $PSIC{'Blastfile'};
    }
    #----------------------------------------

    return 1 if $SkipPSIC;

    #----------------------------------------
    # Now calculate MSA PSIC scores
    #----------------------------------------
    if (defined $PSIC{'Psicfile'} && -e $PSIC{'Psicfile'}) {
      ($PSIC{'Score1'}, $PSIC{'Score2'}, $PSIC{'Nobs'}) =
        PPH::Profile::get_aln_scores_for_SNP($PSIC{'Psicfile'}, $SNP->{'Pos'}, $SNP->{'aa1'}, $SNP->{'aa2'});
    }

    # Switch to MULTIZ alignment if MULTIZPATH option was specified and MULTIZ alignment
    # coverage at this position equals or exceeds that of the regular MSA alignment
    if ($CONFIG{'MULTIZPATH'} && $mz_found && $mz_nobs && $mz_nobs >= $PSIC{'Nobs'}) {
      # DEBUG
      my $nobs = defined $PSIC{'Nobs'} ? $PSIC{'Nobs'} : 'NA';
      #warn "($SNP->{'Acc'}, $SNP->{'Pos'}, $SNP->{'aa1'}, $SNP->{'aa2'}) MZ: NobsMZ=$mz_nobs, Nobs=$nobs\n";
      $PSIC{'Alnfile'}  = $mz_aln;
      $PSIC{'Psicfile'} = $mz_prf;
      $PSIC{'Score1'}   = $mz_score1;
      $PSIC{'Score2'}   = $mz_score2;
      $PSIC{'Nobs'}     = $mz_nobs;
      $PSIC{'Result'}   = 'multiz';
      $PSIC{'Precomputed'} = [ 'alignment', 'profile' ];
    }

    #----------------------------------------
    # IAA: Changed Delta to signed value of scores difference instead of abs()
    #      Note: PPH::Rules::test() method still uses abs(Delta) value for compatibility with PPH v1
    #----------------------------------------
    $PSIC{'Delta'} = $PSIC{'Nobs'} ? sprintf('%+.3f', $PSIC{'Score1'} - $PSIC{'Score2'}) : undef;

    #----------------------------------------
    # Calculate substitution scores
    #----------------------------------------
    PPH::Align::ident_fij_score($SNP) if $CONFIG{'SUBSTSCORES'} && defined $PSIC{'Psicfile'} && -e $PSIC{'Psicfile'};

    return 1;
}
#----------------------------------------------------------------------


#------------------------------------------------------------------------------

=head2 calculate_profile

 Usage   : $PPH::Profile::calculate_profile($SNP, $SkipPSIC)
 Function: calculates the PSIC matrix
 Args    : Object of type SNP
           BOOLEAN - do not calculate PSIC profile matrix
 Returns : 'calculated' if successful

=cut

#----------------------------------------
sub calculate_profile {
    my ($SNP, $SkipPSIC) = @_;
    #### PPH__Profile__calculate_profile: $SkipPSIC

    # Need this for locate_file() searches to be able
    # to store located full pathnames persistently.
    my $PSIC = $SNP->{'PSIC'};
    my $seq_file    = "$SNP->{Acc}.seq";
    my $blast_file  = $PSIC->{'Blastfile'};
    my $aln_file    = $PSIC->{'Alnfile'};

    #-----------------------------------------
    # BLAST
    # Skip if either one of the BLAST or
    # alignment files is already present.
    #-----------------------------------------
    my ($aln_found, $blast_found);
    if      ($aln_found = PPH::Misc::locate_file($PSIC->{'Alnfile'}, "$CONFIG{PRECOMPATH}/alignments", "$CONFIG{SCRATCH}/alignments")) {
      if    ($aln_found == 1) { push @{ $PSIC->{'Precomputed'} }, 'alignment'; }
      elsif ($aln_found == 2) { push @{ $PSIC->{'Cached'} }, 'alignment'; }
      $aln_file = $PSIC->{'Alnfile'};
      delete $PSIC->{'Blastfile'};
    } elsif ($blast_found = PPH::Misc::locate_file($PSIC->{'Blastfile'}, "$CONFIG{PRECOMPATH}/blastfiles", "$CONFIG{SCRATCH}/blastfiles")) {
      if    ($blast_found == 1) { push @{ $PSIC->{'Precomputed'} }, 'blast'; }
      elsif ($blast_found == 2) { push @{ $PSIC->{'Cached'} }, 'blast'; }
      $blast_file = $PSIC->{'Blastfile'};
    } else {
        PPH::BLAST::run_BLAST($seq_file, $blast_file, $CONFIG{'BLAST_ALN'})
          or return 'BLAST error';
        copy($blast_file, "$CONFIG{SAVEPATH}/blastfiles/")
          or confess "Failed to save file: $blast_file";
        $PSIC->{'Blastfile'} = "$CONFIG{SAVEPATH}/blastfiles/$blast_file";
        push @{ $PSIC->{'Calculated'} }, 'blast';
    }

    #----------------------------------------
    # MSA
    # Fall back to BLAST pairwise pseudo-MSA
    # if MSA_LEON_CLUSPACK pipeline fails.
    #----------------------------------------
  SWITCH:
    {
      # Skip MSA creation if alignment file is already present
      last if $aln_found;

      # Test whether the BLAST search results are too large to handle
      return 'Huge BLAST' if -s $blast_file > $CONFIG{'BIGBLAST'};

      my $status = '';

      # Do not even attempt PPH-2 MSA pipleine when USEBLASTMSA config option is set
      unless ($CONFIG{'USEBLASTMSA'}) {
        # IAA: Use different cutoffs for BLAST_MSA_LEON_CLUSPACK v2 pipeline
        #      vs. old PPH-1 BLAST version
        $CONFIG{'MINPSICHITIDE'} = $CONFIG{'MSA_MINPSICHITIDE'};
        $CONFIG{'MAXPSICHITIDE'} = $CONFIG{'MSA_MAXPSICHITIDE'};
        $PSIC->{'MSAVersion'}    = 2; # PPH2 MSA pipeline version used

        $status =
          PPH::Align::Create_aln_BLAST_MSA_LEON_CLUSPACK(
            $SNP, $blast_file, $aln_file
          );

        # Break out here if PPH-2 MSA pipeline was successful...
        last SWITCH if $status eq 'OK';

        chomp $status;
        # ...or warn and proceed with PPH-1 BLAST pipeline if it failed
        warn "WARNING: MSA_LEON_CLUSPACK failed ($status), switching to BLAST alignment\n";
      }

      # IAA: Use different cutoffs for PPH-1 BLAST pipeline
      $CONFIG{'MINPSICHITIDE'} = $CONFIG{'BLAST_MINPSICHITIDE'};
      $CONFIG{'MAXPSICHITIDE'} = $CONFIG{'BLAST_MAXPSICHITIDE'};
      $CONFIG{'BLAST_EVALUE'}  = '1e-4';
      $PSIC->{'MSAVersion'}    = 1; # PPH1 MSA pipeline version used

      $status = PPH::Align::Create_aln_from_BLAST_m7(
        $SNP, $blast_file, $aln_file
      );
      last SWITCH if $status eq 'OK';

      chomp $status;
      warn "ERROR: BLAST alignment failed ($status), giving up\n";
      return $status;
    }
    #----------------------------------------

    #----------------------------------------
    # PSIC
    #----------------------------------------
    unless ($SkipPSIC) {
      my $psic_file = "$SNP->{Acc}.prf";  # forget about any old prf file hanging around
      my $exec = "$CONFIG{PSIC} $CONFIG{PSIC_ARG} $aln_file $CONFIG{MATRIX}/Blosum62.txt $psic_file 2>&1";
      ##### |    Execute: $exec
      my $ret = `$exec`;
      if (length $ret) {
        chomp $ret;
        my $msg = "PSIC failed: $ret";
        warn "$msg\n";
        return $msg;
      }
      copy($psic_file, "$CONFIG{SAVEPATH}/profiles/")
        or confess "ERROR: Failed to save PSIC profile file: $psic_file";
      $PSIC->{'Psicfile'} = "$CONFIG{SAVEPATH}/profiles/$psic_file";
      push @{ $PSIC->{'Calculated'} }, 'profile';
    }
    #----------------------------------------

    return 'calculated';
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 get_aln_scores_for_SNP

 Usage   : $PPH::Profile::get_aln_scores_for_SNP($psicfile, $pos, $var1, $var2)
 Function: Retrieve the PSIC score for a given SNP
 Args    : PSIC filename
           Amino acid position
           Amino acid variation 1 & 2

=cut

#----------------------------------------
sub get_aln_scores_for_SNP {
    my ($psic_file, $pos, $var1, $var2) = @_;
    #### PPH Profile get_aln_scores_for_SNP: $psic_file, $pos, $var1, $var2
    open my $F, '<', $psic_file or confess "Can't read file: $psic_file";
    my @lines = <$F>;
    close $F;

    ($#lines + 1 < $pos)
        and warn "Insufficient No. of lines in profile $psic_file for position: $pos"
            and return(undef, undef, 0);

    my @alphabet = (split /\s+/, $lines[1])[1..20];

    ((grep $var1, @alphabet) and (grep $var2, @alphabet))
        or do {
            warn "Unable to locate variant aa1 ($var1) or aa2 ($var2) in profile: $psic_file";
            return(undef, undef, 0);
        };

    my ($num, @scores) = split /\s+/, $lines[$pos + 1];
    my $Nobs           = pop @scores; # Last column in matrix is Nobs

    ($num != $pos)
        and warn "Can't locate corresponding line in profile $psic_file for pos: $pos"
            and return(undef, undef, 0);

    ($#scores != 19)
        and warn "Invalid line in profile $psic_file for pos: $pos"
            and return(undef, undef, 0);

    my %aaidx;
    @aaidx{@alphabet}  = (0..$#alphabet);

    ##### |    get_aln_scores_for_SNP returns: $scores[$aaidx{$var1}], $scores[$aaidx{$var2}], $Nobs

    return($scores[$aaidx{$var1}], $scores[$aaidx{$var2}], $Nobs);
}
#------------------------------------------------------------------------------

#SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

1;
