package PPH::Rules;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::Rules

=head1 DESCRIPTION

Return a prediction for the effect of a non-synonymous substitution -
rule based

=head1 AUTHOR

StS

=head1 SUBVERSION

 $LastChangedDate: 2009-11-28 00:15:09 -0500 (Sat, 28 Nov 2009) $
 $LastChangedRevision: 254 $
 $LastChangedBy: ivan $



=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD


use Carp qw(cluck);
use Data::Dumper;
use strict;
use warnings;

BEGIN {
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR


=head1 ROUTINES

=head2 test

 Usage   : $PPH::Rules->test(SNP)
 Function: Rule based prediction of the effect of a non-synonymous
           substitution
 Args    : Object of type SNP
 Returns : Hashref of prediction
           {
             Prediction
             Avail
             Basis
             Effect
             Data
           }

=cut

#----------------------------------------------------------------------
sub test {
    my $SNP = shift;

    ### PPH__Rules__test

    my $features  = exists($SNP->{Annot}) && defined($SNP->{Annot}) ?
                          $SNP->{Annot} : undef;
    my $delta     = exists($SNP->{PSIC}) && exists($SNP->{PSIC}{Delta}) &&
                      defined($SNP->{PSIC}{Delta}) ?
                          abs($SNP->{PSIC}{Delta}) : undef;
    my $structure = exists($SNP->{Struct}) && exists($SNP->{Struct}{NumHitsFiltered}) &&
                      defined($SNP->{Struct}{NumHitsFiltered}) &&
                      $SNP->{Struct}{NumHitsFiltered} > 0 ?
                          $SNP->{Struct}{NumHitsFiltered} : undef;

    my $struct_params  = defined($structure) ?
                            $SNP->{Struct}{Maps}[0]{Params} : undef;
    my $struct_contact = defined($structure) && exists($SNP->{Struct}{BestContact}) ?
                            $SNP->{Struct}{BestContact} : undef;

    my @avail;
    defined($features)  and push @avail, 'FT';
    defined($delta)     and push @avail, 'alignment';
    defined($structure) and push @avail, 'structure';

    #--------------------------------------------------------------------
    # Worst case : no prediction is possible ('unknown')
    #--------------------------------------------------------------------
    defined($features) or defined($delta) or defined($structure)
        or  return
            {
                Prediction => $PPH::Data::Predictions[0],
                Avail      => undef,
                Basis      => undef,
                Effect     => undef,
                Data       => undef
            };
    #---------------------------------------------------------------------

    #---------------------------------------------------------------------
    # We have a UniProt Site features
    #---------------------------------------------------------------------
    if (defined($features)) {
        #### features defined
        #----------------------------------------
        # Site features
        #----------------------------------------
        exists($features->{Site})
            and ($features->{Site} ne 'CARBOHYD')
                and return
                    {
                        Prediction => $PPH::Data::Predictions[3],
                        Avail      => [@avail],
                        Basis      => 'sequence annotation',
                        Effect     => (grep {/$features->{Site}/}
                                           @PPH::Data::BONDS) ? '1.2' : '2.2',
                        Data       => "Site type: ".$features->{Site}
                    };
        #----------------------------------------

        #----------------------------------------
        # Region features
        #----------------------------------------
        exists($features->{Region})
            and ($features->{Region} eq 'TRANSMEM')
                and return
                    {
                        Prediction => ($features->{PHAT} < 0) ?
                            $PPH::Data::Predictions[2] :
                                $PPH::Data::Predictions[1],
                        Avail      => [@avail],
                        Basis      => ($features->{RegionStatus} eq
                                           'predicted') ?
                                               'sequence prediction' :
                                                   'sequence annotation',
                        Effect     => ($features->{PHAT} < 0) ? '2.2.2' : undef,
                        Data       => "PHAT matrix element difference: ".
                            $features->{PHAT}
                    };
        #----------------------------------------
    }
    #---------------------------------------------------------------------

    #---------------------------------------------------------------------
    # if nothing else...
    defined($delta)
        or return
            {
                Prediction => $PPH::Data::Predictions[0],
                Avail      => [@avail],
                Basis      => undef,
                Effect     => undef,
                Data       => undef
            };
    #---------------------------------------------------------------------

    #---------------------------------------------------------------------
    # Structure - contacts to important residues or substrates / cofactors
    #---------------------------------------------------------------------
    if ($delta > 1 and defined($structure) and defined($struct_contact)) {
        #### Delta lt 1: $delta, $structure, $struct_contact
        #----------------------------------------
        # A) Heteroatom (e.g.substrate / cofactor)
        #----------------------------------------
        (exists($struct_contact->{Heteroatoms}) and
             exists($struct_contact->{Heteroatoms}{Dist})
                 and ($struct_contact->{Heteroatoms}{Dist} <= 3))
            and return
                {
                    Prediction => $PPH::Data::Predictions[3],
                    Avail      => [@avail],
                    Basis      => 'structure',
                    Effect     => '2.2.3',
                    Data       => "PSIC score difference: $SNP->{PSIC}{Delta}. ligand name: $SNP->{Struct}{BestContact}{Heteroatoms}{Name}, distance: $SNP->{Struct}{BestContact}{Heteroatoms}{Dist} A"
                };
        #----------------------------------------

        #----------------------------------------
        # B) Critical site (e.g. cat site)
        #----------------------------------------
        (exists($struct_contact->{Critical})
             and exists($struct_contact->{CritSites}{Dist})
                 and ($struct_contact->{CritSites}{Dist} <= 3))
            and return
                {
                    Prediction => $PPH::Data::Predictions[3],
                    Avail      => [@avail],
                    Basis      => 'structure',
                    Effect     => '2.1',
                    Data       => "PSIC score difference: $SNP->{PSIC}{Delta}. functional site: $SNP->{Struct}{BestContact}{CritSites}{Name} $SNP->{Struct}{BestContact}{CritSites}{Num}, distance: $SNP->{Struct}{BestContact}{CritSites}{Dist} A"
                };
        #----------------------------------------
    }
    #---------------------------------------------------------------------

    #---------------------------------------------------------------------
    # PSIC - sequence conservation plus structure parameters if available
    #---------------------------------------------------------------------
    ($delta <= 0.5)
        and return
            {
                Prediction => $PPH::Data::Predictions[1],
                Avail      => [@avail],
                Basis      => 'alignment',
                Effect     => undef,
                Data       => "PSIC score difference: $SNP->{PSIC}{Delta}"
            };
    #----------------------------------------

    #----------------------------------------
    ($delta > 0.5 and $delta <= 1.5)
        and do {
            #### $delta, $struct_params->{AccNorm}, $struct_params->{dProp}, $struct_params->{dVol}
            #----------------------------------------
            # Is SNP on surface?
            #----------------------------------------
            if (defined($structure) and
                    exists($struct_params->{AccNorm}) and
                        ($struct_params->{AccNorm} <= 0.05)) {
                #----------------------------------------
                # Propensity changes of aa variants
                #----------------------------------------
                ($struct_params->{dProp} >= 1)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[3],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.1',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, hydrophobicity change: $SNP->{Struct}{Maps}[0]{Params}{dProp}"
                        };
                #----------------------------------------

                #----------------------------------------
                # Vol changes of aa variants
                #----------------------------------------
                ($struct_params->{dVol} >= 80)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[3],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.2',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };

                ($struct_params->{dVol} <= -80)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[3],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.3',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };
                #----------------------------------------
            }
            #----------------------------------------

            #----------------------------------------
            # Lower threshold for SNP accesibility
            #----------------------------------------
            if (defined($structure) and
                    exists($struct_params->{AccNorm}) and
                        ($struct_params->{AccNorm} > 0.05) and
                            ($struct_params->{AccNorm} <= 0.15)) {
                #----------------------------------------
                # Propensity changes
                #----------------------------------------
                ($struct_params->{dProp} >= 0.75)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[2],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.1',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, hydrophobicity change: $SNP->{Struct}{Maps}[0]{Params}{dProp}"
                        };
                #----------------------------------------

                #----------------------------------------
                # Vol changes
                #----------------------------------------
                ($struct_params->{dVol} >= 60)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[2],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.2',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };

                ($struct_params->{dVol} <= -60)
                    and return
                        {
                            Prediction => $PPH::Data::Predictions[2],
                            Avail      => [@avail],
                            Basis      => 'structure',
                            Effect     => '1.1.3',
                            Data       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };
            }
            #----------------------------------------

            return {
                Prediction => $PPH::Data::Predictions[1],
                Avail      => [@avail],
                Basis      => 'alignment',
                Effect     => undef,
                Data       => "PSIC score difference: $SNP->{PSIC}{Delta}"
            };
        };
    # end ($delta > 0.5 and $delta <= 1.5)
    #----------------------------------------

    #----------------------------------------
    ($delta > 1.5 and $delta <= 2.0)
        and do {
            #### $delta, $struct_params->{AccNorm}, $struct_params->{dProp}, $struct_params->{dVol}

            #----------------------------------------
            # Accessbility
            #----------------------------------------
            if (defined($structure) and
                    exists($struct_params->{AccNorm}) and
                        $struct_params->{AccNorm} <= 0.05) {

                #----------------------------------------
                # Propensity
                #----------------------------------------
                ($struct_params->{dProp} >= 1)
                    and return
                        {
                            Prediction  => $PPH::Data::Predictions[3],
                            Avail       => [@avail],
                            Basis       => 'structure',
                            Effect      => '1.1.1',
                            Data        => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, hydrophobicity change: $SNP->{Struct}{Maps}[0]{Params}{dProp}"
                        };
                #----------------------------------------
                # Vol change
                #----------------------------------------
                ($struct_params->{dVol} >= 80)
                    and return
                        {
                            Prediction  => $PPH::Data::Predictions[3],
                            Avail       => [@avail],
                            Basis       => 'structure',
                            Effect      => '1.1.2',
                            Data        => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };

                ($struct_params->{dVol} <= -80)
                    and return
                        {
                            Prediction  => $PPH::Data::Predictions[3],
                            Avail       => [@avail],
                            Basis       => 'structure',
                            Effect      => '1.1.3',
                            Data	       => "PSIC score difference: $SNP->{PSIC}{Delta}, normed accessibility: $SNP->{Struct}{Maps}[0]{Params}{AccNorm}, volume change: $SNP->{Struct}{Maps}[0]{Params}{dVol}"
                        };
            }
            #----------------------------------------

            return
                {
                    Prediction  => $PPH::Data::Predictions[2],
                    Avail       => [@avail],
                    Basis       => 'alignment',
                    Effect      => undef,
                    Data        => "PSIC score difference: $SNP->{PSIC}{Delta}"
                };
        };
    # end ($delta > 1.5 and $delta <= 2.0)
    #------------------------------

    #------------------------------
    ($delta > 2.0)
        and return
            {
                Prediction  => $PPH::Data::Predictions[3],
                Avail       => [@avail],
                Basis       => 'alignment',
                Effect      => undef,
                Data        => "PSIC score difference: $SNP->{PSIC}{Delta}"
            };
    #------------------------------




    #------------------------------
    #
    warn "This shouldn't happen - all rules failed!";
    return
        {
            Prediction => $PPH::Data::Predictions[0],
            Avail      => [@avail],
            Basis      => undef,
            Effect     => undef,
            Data       => undef
        };
    #------------------------------
}
#----------------------------------------------------------------------

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
1;
