package PPH::Seq;
use strict;
use warnings;
use Carp qw(cluck confess);

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::Seq

=head1 DESCRIPTION

Routines to process / convert sequence / annotation database files
(former pph_seq)

=head1 SUBVERSION

 $LastChangedDate: 2010-05-24 19:43:27 -0400 (Mon, 24 May 2010) $
 $LastChangedRevision: 305 $
 $LastChangedBy: ivan $



=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 0.1;
    @ISA           = qw(Exporter);
    @EXPORT        = qw();
    @EXPORT_OK     = qw();
    %EXPORT_TAGS = ();
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}
our @EXPORT_OK;

use File::Temp;
use Storable;

#IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII


#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

=head1 ROUTINES

#------------------------------------------------------------------------------

=head2 uniprot_id2acc

 Usage   : PPH::Seq->uniprot_id2acc($id)
 Function: Converts a UniProt identifier into a accession number
 Args    : sequence identifier
 Returns : accession number

=cut

#----------------------------------------
sub uniprot_id2acc {
  my $search_id = shift;

  #### PPH__Seq__uniprot_id2acc: $search_id
  my $file = "$CONFIG{UNIPROT}.id2acc";

  ##### |    Open: $file
  open my $ID2ACC, '<', $file
      or confess "can't open id->acc mapping file $file";

  while (<$ID2ACC>) {
    my($id, $acc) = (split);
    ($search_id eq $id) and do { close $ID2ACC; chomp($acc); return uc($acc); };
  }
  close $ID2ACC;

  return '';
}
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------

=head2 xdb2acc

 Usage   : PPH::Seq->xdb2acc($id)
 Function: maps a number of different protein identifiers from
           various databases to a unique UniProtKB accession number
 Args    : protein (!) sequence identifier, currently supported
           identifiers are RefSeq and GenBank gi
 Returns : UniProtKB accession number, empty string if not matched

=cut

#------------------------------------------------------------------------------
{
my $db2acc;

sub xdb2acc {
  my $dbid = shift;
  die "Sequence ID missing\n" unless defined $dbid;
  $dbid =~ s/^\s+//; $dbid =~ s/\s+$//;
  die "Empty sequence ID\n" unless length $dbid;
  my $dbtag;
  # RefSeq
  if      ($dbid =~ /^[ANXYZ]P_\d+(\.\d+)?$/) {
    $dbtag = 'refseq';
  # GI: prefix is case-insensitive
  } elsif ($dbid =~ /^gi:\d+$/i) {
    $dbid =~ s/^gi://i;
    $dbtag = 'gi';
  } else {
    return '';
  }
  unless (defined $db2acc) {
    $db2acc = retrieve($CONFIG{'UNIPROT'}.'.xdb2acc');
    die "ERROR: Failed to load idmapping file: $!\n" unless defined $db2acc;
  }
  if (exists $db2acc->{$dbtag}{$dbid}) {
    return $db2acc->{$dbtag}{$dbid};
  } else {
    return '';
  }
}
}

#------------------------------------------------------------------------------

=head2 snp2acc

 Usage   : PPH::Seq->snp2acc($rsid)
 Function: maps a dbSNP rsID a UniProtKB variant
 Args    : rsID SNP identifier
 Returns : UniProtKB accession number, variant position and two aa residues

=cut

#------------------------------------------------------------------------------
sub snp2acc {
  my $rsid = shift;
  die "ERROR: snp2acc(): empty/missing rsID\n" unless defined $rsid && length $rsid;

  #### PPH__Seq__snp2acc: $rsid
  my $file = "$CONFIG{UNIPROT}.snp2acc";

  ##### |    Open: $file
  open my $SNP2ACC, '<', $file
      or confess "can't open snp->acc mapping file $file";

  while (<$SNP2ACC>) {
    my ($id, @variant) = split;
    ($rsid eq $id) and do { close $SNP2ACC; chomp($variant[-1]); return @variant; };
  }
  close $SNP2ACC;

  return '';
}

#------------------------------------------------------------------------------

=head2 fastacmd

  Usage   : PPH::Seq->fastacmd($id, $db)
  Function: Fetch a sequence from a database prepared with formatdb (NCBI)
  Args    : Identifier
            Database
  Returns : Unprocessed header
            Accession no
            Identifier
            Description
            The sequence (without white spaces)

=cut

#----------------------------------------
sub fastacmd {
    my ($id, $db) = @_;
    #### PPH__Seq__fastacmd: @_
    ##### |     Open: $db, $id
    my ($head, $seq);
    eval {
        my $pid;
        my $tmpf = File::Temp->new();
#         system("$CONFIG{FASTACMD}", '-d', "$db", '-p', 'T', '-s', "$id", '-o', "$tmpf") == 0 or
#           die "failed to execute fastacmd";
        # IAA: switched to blastdbcmd (NCBI BLAST+) because of fastacmd random crashes
        system( "$CONFIG{FASTACMD}",
          '-db', $db, '-dbtype', 'prot', '-entry', $id, '-out', $tmpf, '-outfmt', '%f' ) == 0 or
          die 'ERROR: Failed to execute blastdbcmd';

        open my $F, '<', $tmpf or die "Can't open file: $tmpf";
        $head = <$F>;
        $seq  = join '', <$F>;
        close $F;

        defined($head) or die 'ERROR: fastacmd failed';
        chomp($head);
        $seq =~ s/\s+//g;

    };
    if ($@) {
        warn "Warning: $@";
        return ();
    }

    #----------------------------------------
    # Take care of PDB ids
    # (e.g. 2FWR_D should be pdb|2FWR|D)
    #----------------------------------------
    if ($head !~ /^>/ and $id =~ /\_/) {
        my ($pdbid, $chain) = split /_/, $id;
        my $cmd = qq($CONFIG{FASTACMD} -db $db -dbtype prot -entry 'pdb|$pdbid|$chain' -outfmt '%f' 2>&1);
        open my $C, '-|', $cmd
            or confess "Can't execute $CONFIG{FASTACMD} $db pdb|$pdbid|$chain";
        $head = <$C>; chomp($head);
        close $C;
    }
    #----------------------------------------

    ($head =~ /^>/) or return ();

    my ($tmp, $desc) = split /\s+/, $head, 2;
    my (@fields)     = split /\|/,  $tmp;

    my ($acc);
    if ($#fields == 4) {
        if ($fields[2] eq 'pdb') {
            $acc = $fields[1], $id = $fields[3];
        } else {
            $acc = $fields[3], $id = $fields[4];
        }
    } elsif ($#fields == 3) {
        $acc = $fields[1], $id = $fields[3];
    } elsif ($#fields == 2) {
        $acc = $fields[1], $id = $fields[2];
    } else {
        $acc = $id = $tmp;
    }

    ##### |     return(head, $acc, $id, $desc, $seq)
    return ($head, $acc, $id, $desc, $seq);
}
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------

=head2 fetch_FASTA

 Usage   : PPH::Seq->fetch_FASTA($id, $db)
 Function: Fetch a sequence from a database
 Args    : Sequence identifier
           FASTA database file name
 Returns : Accession no
           Identifier
           Description
           The sequence (without white spaces)

=cut

#----------------------------------------
sub fetch_FASTA {
    my ($id, $db) = @_;
    #### PPH__Seq__fetch_FASTA: $id, $db

    ##### |    $CONFIG{FETCHSEQ}: $CONFIG{FETCHSEQ}
    open my $F, '-|', "$CONFIG{FETCHSEQ} $db $id"
      or confess "Can't execute fetchseq $db $id\n";

    my $output = do { local $/; <$F> }; close $F;

    return undef unless defined $output && length $output;

    my ($head, $sequence) = split /\n/m, $output, 2;

    $head =~ s/^>\s*//;
    my $tmp; my $description;
    ($tmp, $description) = split /\s+/, $head, 2;

    defined($tmp)         or return ();
    defined($description) or $description = '';

    my $acc;
    if ($tmp =~ /^\w+\|\w+\|\w+\|\w+$/) {
        (undef, $acc, undef, $id) = split /\|/, $tmp;
    } elsif ($tmp=~ /^\w+\|\w+\|\w+$/) {
        (undef, $acc, $id) = split /\|/, $tmp;
    } else {
        $acc = $id = $tmp;
    }

    $sequence =~ s/\s|\n//mg;
    $sequence =~ s/\*$//;

    ##### |    return: $acc, $id, $description, $sequence
    return ($acc, $id, $description, $sequence);
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 slurp_dbase - OUTDATED

 Usage   : my $DBref = PPH::Seq::slurp_dbase($dbfile)
           print $DBref->[0]{Id}
 Function: Reads in a sequence database ???Do we need this???
 Args    : FASTA database file name
 Returns : Pointer to Array of Hashes:
           [{ Acc, Id, Des, Seq }]

=cut

#----------------------------------------
sub slurp_dbase {
    confess "OUTDATED";
    my $file = shift;

    open my $FASTA, '<', $file or confess "can't open $file";

    my ($seq, $id, $acc, $desc, @dbase);

    my $out;
    $out = do { local $/; <$FASTA>};

    foreach my $fas (split /^>/m, $out) {
        ($fas =~ /^\s*$/) and next;
        my ($header, $seq) = split /\n/, $fas, 2;

        ($id, $desc)        = split /\s+/, $header, 2;
        (undef, $acc, $id) = split /\|/,  $id;

        $seq =~ s/\n|\s//mg;

        push @dbase, {Seq => $seq, Acc => $acc, Id => $id, Des => $desc};
    }

    return \@dbase;
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 get_uniprot_annotation

 Usage   : my $status = $PPH::Seq::get_uniprot_annotation($SNP)
 Function: Find UniProt annotations to a given SNP
 Args    : Object of type SNP
 Returns : $SNP->{Annot} =
           { Region, RegionStatus, PHAT, Site, CritSites }
           1 if ok

=cut

#----------------------------------------
sub get_uniprot_annotation {
    my ($SNP) = shift;

    ### PPH Seq get_uniprot_annotation: $SNP

    my %Annot;
    my $Featr = _grep_features($SNP->{Acc});
    defined($Featr) or return 0;


    foreach my $feature (keys %$Featr) {

        #----------------------------------------
        # Is the Feature at the SNP Position?
        # !!! Allows only one SITE and REGION !!!

        if (grep { $_ eq $SNP->{Pos} } @{$Featr->{$feature}}) {

            #----------------------------------------
            # Is the Feature a region?
            #----------------------------------------
            if (exists($PPH::Data::REGIONS{$feature})) {
                (exists($Annot{Region}) or # Allow Only one region feature
                     $feature =~ /_PRED/) and next; # Ignore PRED if FT

                $Annot{Region} = $feature;

                #----------------------------------------
                # In case we have a transmembrane region
                #----------------------------------------

                if ($Annot{Region} =~ /TRANSMEM/) {
                    my %PHAT = PPH::Data::blastmatrix("$CONFIG{MATRIX}/PHAT_T75_B73.bla");
                    $Annot{PHAT} = $PHAT{$SNP->{aa1}}{$SNP->{aa2}};
                }
            }
            #----------------------------------------
            # ...or a site?
            elsif (exists($PPH::Data::SITES1{$feature}) or
                       exists($PPH::Data::SITES2{$feature})) {
                $Annot{Site} = $feature;
            }
        }
        # Feature at SNP position -------------------

        exists($PPH::Data::SITES1{$feature})
            and push @{$Annot{CritSites}}, @{$Featr->{$feature}};
    }
    # Foreach feature -----------------------

    #----------------------------------------
    # Nothing found
    (%Annot) or return 0;
    (exists($Annot{Region})     or
         exists($Annot{Site})       or
             scalar(@{$Annot{CritSites}})) or return 0;
    #----------------------------------------

    #----------------------------------------
    if (exists($Annot{Region})) {
        $Annot{RegionStatus} = ($Annot{Region} =~ /_PRED/) ? 'predicted' : 'proven'
    }
    #----------------------------------------

    ##### |    get_uniprot_annotation return: %Annot

    $SNP->{Annot} = \%Annot;
    return 1;
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head3 _grep_features

 Usage   : my $Featr = _grep_features($SNP->{Acc});
 Function: Find features from a already prepared feature file
 Args    : Accession
 Returns : HashRef to features
           { [ description ] }

=cut

#----------------------------------------
sub _grep_features {
    my ($acc) = shift;
    ##### PPH__Seq__grep_features: $acc
    my %Features;

    open my $F, '<', "$CONFIG{UNIPROT}.ft"
        or confess "Features file $CONFIG{UNIPROT} not found!";

    while (<$F>) {
        /^$acc/ or next;

        my (undef, $source, $desc, $from, $to) = split;

        #----------------------------------------
        # Take only features with defined pos
        #----------------------------------------
        ($from =~ /^\d+$/ and $to =~ /^\d+$/) or next;

        # ?? Not found in .ft file
        $desc =~ s/(RPT)_\d+/$1/;

        # ?? Also not found
        ($source eq 'PRED') and $desc .= '_PRED';

        #----------------------------------------
        # If site   - return only position
        # If region - return array from..to
        #----------------------------------------
        push @{$Features{$desc}},
            (exists($PPH::Data::BONDS{$desc})) ? ($from, $to) : ($from..$to);
    }
    close $F;

    ###### |    PPH__Seq__grep_features returns: %Features

    return scalar(keys %Features) ? \%Features : undef;
}
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

=head2 grep_in_FASTA

 Usage   : my ($acc, $id, $desc, $seq) = grep_in_FASTA($query_seq, $dbfile)
 Function: Find a (sub) sequence in a FASTA formatted file
           File has to keep each sequence on ONE LINE!!!
 Args    : query sequence
           FASTA filename
 Returns : Returns FIRST hit which contains the match to query sequence
           Accession
           Identifier
           Description
           Sequence

=cut

#----------------------------------------
sub grep_in_FASTA {
  my ($query_seq, $db) = @_;

  #### PPH Seq grep_in_FASTA($query_seq, $db)
  open my $F, '-|', "fgrep -s -m 1 -B 1 '$query_seq' $db";
  my $head = <$F>;
  close($F), return () unless defined $head;
  my $seq = <$F>;
  close $F;

  chomp $head;
  chomp $seq;

  ($head =~ s/^>\s*//) or confess "ERROR: Incorrect format of FASTA defline: $head\n";

  my ($tmp, $desc) = split /\s+/, $head, 2;
  defined($tmp) or return ();

  my ($acc, $id);
  if ($tmp =~ m/^\S+?\|\S+?\|\S+/) {
    (undef, $acc, $id) = split /\|/, $tmp;
  } else {
    $acc = $id = $tmp = undef;
  }

  ##### |     return($acc, $id, $desc, $seq)
  return ($acc, $id, $desc, $seq);
}


#MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM

1;
