#!/usr/bin/env perl

use strict;
use warnings;

BEGIN {
  use FindBin qw($RealBin);
  BEGIN {
    unless ($ENV{'PPH'}) {
      my $binpath = $RealBin;
      $binpath =~ s|/[^/]+$||;
      $ENV{'PPH'} = $binpath;
    }
  }
  use lib "$ENV{PPH}/perl";
}

use PPH::Codon;
use Pod::Usage;
use PPH::Config;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

refGene.prot.nt.fas.pl - convert UCSC refGene track to protein / DNA sequences

=head1 SYNOPSIS

cat refGene.hg18.txt | refGene.prot.nt.fas.pl >refGene.prot.nt.fas

=head1 DESCRIPTION

This script simply relies on the mappings of refGene:
 - be careful since many of these entries will contain stop codons and
   other mistakes
 - duplicated entries will have a "_#N" added
 - format:
   >id
   prot seq
   nt   seq


=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use constant TMPF => "$CONFIG{TMP_OUT}/$$.2bit";
my %NAME;

<>; # skip header with the column names
while (<>) {

  next if /^#/;

  /^\d/ or pod2usage();

  $CONFIG{DEBUG} and print;
  my (@PSL) = split /\s+/;

  my $name       = $PSL[1];
  my $chr        = $PSL[2];
  my $reverse    = ($PSL[3] eq "-") ? 1 : 0;
  my $cdsStart   = $PSL[6];
  my $cdsEnd     = $PSL[7];

  #----------------------------------------
  # non protein coding gene
  #----------------------------------------
  ($cdsEnd == $cdsStart) and next;
  #----------------------------------------

  my (@ExS)      = split /,/, $PSL[9]; # Start
  my (@ExE)      = split /,/, $PSL[10]; # End
  my (@ExF)      = split /,/, $PSL[15];  # Frame

  #----------------------------------------
  # Removing non-coding exons
  #----------------------------------------
  while ($#ExE) {
    $CONFIG{DEBUG} and print "shift $ExE[0] > $cdsStart\n";
    last if $ExE[0] > $cdsStart;
    shift(@ExS), shift(@ExE), shift(@ExF);
  }
  while ($#ExS) {
    last if $ExS[$#ExS] < $cdsEnd;
    pop(@ExS), pop(@ExE), pop(@ExF);
  }
  #----------------------------------------

  $ExS[0]        = $cdsStart;
  $ExE[$#ExE]    = $cdsEnd;

  if ($reverse) {
    (@ExS) = reverse @ExS;
    (@ExE) = reverse @ExE;
    (@ExF) = reverse @ExF;
  }

  my @EXONS;
  for (0..$#ExS) { push @EXONS, "$chr:$ExS[$_]-$ExE[$_]" }

  $CONFIG{DEBUG} and print join "\n", @EXONS, "\n";

  open W, "|"."$CONFIG{TWOBIT} $CONFIG{CHROM2BIT} -seqList=stdin ".TMPF or die;
  print W join "\n", @EXONS;
  close W;

  open F, TMPF or die;
  my (@DNA) =
    map { $reverse ? PPH::Codon::revcomp($_) : $_ }
      map { uc((s/\s+//g, $_)[1]) }
      (split />.*/, (join '', (<F>)));
  close F;
  unlink TMPF;

  shift(@DNA);

  $CONFIG{DEBUG} and do {
    for (0..$#DNA) {
      print "\n$DNA[$_]\n";
      print $ExF[$_], "\n";
      my ($shift) = ($ExF[$_] == 0) ? 0 : (3 - $ExF[$_]);
      print substr($DNA[$_], $shift), "\n";
      print join "  ", PPH::Codon::translate(substr($DNA[$_], $shift)), "\n";
    }
  };

  my $prot = join '', (split '', PPH::Codon::translate(join '', @DNA));
  # ($prot =~ /\*.+/ or $prot !~ /\*$/ or $prot !~ /^M/) and next;
  # ($prot !~ /\*.+/) and next;
  print ">$name".( (exists $NAME{$name}) ? "_".$NAME{$name} : "" );
  print " $chr $PSL[3] $cdsStart $cdsEnd \n";
  print "$prot\n";
  print join "|", @DNA; print "\n";
  $NAME{$name}++;

}

#------------------------------------------------------------------------------
DESTROY {
  unlink TMPF;
}
