#!/usr/bin/env perl
use warnings;
use strict;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

unipfam.pl - a simple script to extract Pfam annotations for UniProt proteins

=head1 SYNOPSIS

unipfam.pl [-n <org_name>] [<seq_file>]

=head1 DESCRIPTION

This script uses UniProt sequence database file (default: 'human.seq')
prepared by 'uniprot.pl' script to get a list of protein accessions
and then downloads and parses swisspfam.gz to extract Pfam annotations
for all proteins.

=head1 SUBVERSION

 $LastChangedDate: 2011-05-13 19:15:44 -0400 (Fri, 13 May 2011) $
 $LastChangedRevision: 363 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use LWP::Simple;
use Getopt::Std;

my %opts;
getopts('n:', \%opts);

# Organism common name: human (default), mouse, rat, dog, zebrafish
my $CNAME = 'human';

# SPECIES hash indexed by CNAME
my %SPECIES = (
  # Human
  'human'    => {
    'os'     => 'Homo sapiens',
    'spname' => 'HUMAN',
    'file'   => 'human',
  },
  # Orangutan
  'orangutan'    => {
    'os'     => 'Pongo pygmaeus|Pongo abelii',
    'spname' => 'PONPY|PONAB',
    'file'   => 'mammals',
  },
  # Mouse
  'mouse'    => {
    'os'     => 'Mus musculus',
    'spname' => 'MOUSE',
    'file'   => 'rodents',
  },
  # Rat
  'rat'      => {
    'os'     => 'Rattus norvegicus|Rattus rattus',
    'spname' => 'RAT|RATRT',
    'file'   => 'rodents',
  },
  # Dog
  'dog'      => {
    'os'     => 'Canis familiaris',
    'spname' => 'CANFA',
    'file'   => 'mammals',
  },
  # Zebrafish
  'zebrafish' => {
    'os'      => 'Danio rerio',
    'spname'  => 'DANRE',
    'file'    => 'vertebrates',
  },
);

if (my $cname = $opts{'n'}) {
  die "Illegal or unsupported organism name: $cname\n"
    unless exists $SPECIES{$cname};
  $CNAME = $cname;
}

my $uniseq_name = "$CNAME.seq";
$uniseq_name    = shift if @ARGV;
my $outfile     = "$CNAME.pfam";

open(F, $uniseq_name) or die "Can't open input file: $uniseq_name\n";

my %HsAcc;
warn "Reading protein accessions...\n";
while (<F>) {
    /^>/ or next;
    my (undef, $acc) = split /\|/;
    $HsAcc{$acc} = 1;
}
close F;
warn "Done.\n";

my $fetched = 0;
# Reuse existing swisspfam.gz file
if (-e 'swisspfam.gz') {
  warn "Found swisspfam.gz in the current directory, reusing.\n";
# Download recent swisspfam.gz file
} else {
  my $PFAM_FTP = 'ftp://ftp.sanger.ac.uk/pub/databases/Pfam/current_release/swisspfam.gz';
  warn "Fetching swisspfam.gz...\n";
  die "Can't fetch file: swisspfam.gz\n" unless mirror($PFAM_FTP, 'swisspfam.gz') == 200;
  warn "Done.\n";
  $fetched = 1;
}

warn "Extracting Pfam annotations...\n";
open(F, 'gunzip -c swisspfam.gz |') or die "Can't open input file: swisspfam.gz\n";
open(FOUT, ">$outfile") or die "Can't create output file: $outfile\n";
while (<F>) {
    /^>/ or next;
    my (undef, undef, $acc) = split;
    $acc =~ s/\.\d+$//;
    exists $HsAcc{$acc} or next;
    while (<F>) {
        /^\s*$/ and last;
        /^Pfam-B/ and next;
        chomp;
        my ($id, $desc, @f) = split;
        my @loc;
        while ($f[$#f] =~ /^\d+-\d+$/) {
            push @loc, pop @f;
        }
        while ($f[$#f] !~ /^PF/) {
            $desc = (pop @f)." $desc";
        }
        print FOUT join("\t", $acc, $f[$#f], $id, $desc, @loc), "\n";
    }
}
close FOUT;
close F;
unlink 'swisspfam.gz' if -e 'swisspfam.gz' && $fetched;
warn "Done.\n";
