package PPH::3D;
use strict;
use warnings;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

PPH::3D - routines for analyzing structures - former pph_3d.pl

=head1 DESCRIPTION

Depends on the PDB database...

=head1 AUTHOR

StS

=head1 SUBVERSION

 $LastChangedDate: 2011-05-28 00:55:28 -0400 (Sat, 28 May 2011) $
 $LastChangedRevision: 365 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Carp qw(cluck confess);
use File::Copy;
use Storable;
use PPH::Data;
use PPH::Misc;
use File::Basename;

BEGIN {
    #----------------------------------------
    # DEBUGGING
    #----------------------------------------
    use PPH::Config; # imports: $DEBUG @DEBUGGING_MODE %CONFIG
    if ($DEBUG) {
      # Test if we have Smart::Comments installed
      eval { require Smart::Comments; Smart::Comments->import(-ENV) };
      die "Please install Smart::Comments module or switch off verbose/DEBUG mode\n" if $@;
    }
    #----------------------------------------
}

#GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG

my @SUSPICIOUS_CONTACTS;
my %FRAGM;

#GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

=head1 ROUTINES

=head2 set_3D_parameter

 Usage   : PPH::3D::set_3D_parameter($SNP)
 Function: Find homologous protein structures for the given SNP and retrieve
           structural informations
 Args    : Object of type SNP
 Returns : TRUE if successful
           $SNP->{Struct} will contain all information

=cut

#----------------------------------------
sub set_3D_parameter {
    my ($SNP) = shift;

    #### PPH__3D__set_3D_parameter

    #----------------------------------------
    # Locate the BLAST file or otherwise
    # run BLAST
    #----------------------------------------
    my $blast_3d = "$SNP->{Acc}.blast_pdb";
    my $seqfile  = "$SNP->{Acc}.seq";

    #----------------------------------------

    #----------------------------------------
    # BLAST if this file doesn't exist
    #----------------------------------------
    unless (PPH::Misc::locate_file($blast_3d,
      "$CONFIG{PRECOMPATH}/structures", "$CONFIG{SCRATCH}/structures")) {
        #----------------------------------------
        # Well, don't
        #----------------------------------------
        if ($CONFIG{'PRECOMPONLY3D'}) {
          $blast_3d = undef;
          return 0;
        }
        #----------------------------------------
        PPH::BLAST::run_BLAST($seqfile, $blast_3d, $CONFIG{'BLAST_3D'})
          or do {
              cluck "BLAST warning! Multiple accession numbers";
              return 0;
          };
        copy($blast_3d, "$CONFIG{SAVEPATH}/structures")
          or warn "Failed to save file: $blast_3d\n";
    }
    #----------------------------------------

    #----------------------------------------
    # Parse the BLAST XML file and return
    # HSPs (the keys correspond to XML plus
    # Hit-id (Id) and Hit-desc (Desc)
    #----------------------------------------
    my @HSPs;
    PPH::BLAST::parse_XML_BLAST($blast_3d, \@HSPs);
    $SNP->{Struct}{NumHits} = scalar(@HSPs);
    #----------------------------------------

    #----------------------------------------
    # Filter the HSPs according to rules
    # given in %CONFIG
    #----------------------------------------
    @HSPs = grep PPH::Align::filter_3D_HSP($SNP->{Pos}), @HSPs;
    #----------------------------------------

    #----------------------------------------
    # No hits found
    #----------------------------------------
    ((scalar(@HSPs) > 0) and $SNP->{Struct}{NumHitsFiltered} = scalar(@HSPs))
        or return 0;
    #----------------------------------------

    #----------------------------------------
    # Sort HSPs by identity...
    #----------------------------------------
    ($CONFIG{'SORTBYIDE'})
        and @HSPs = sort {$b->{Identity} <=> $a->{Identity}} @HSPs;
    #----------------------------------------

    #----------------------------------------
    # Analyze each hsp
    #----------------------------------------
    my $hsp_no = 1;

    my %best_contact;		# Best contact in all HSPs
    my $dbname = $CONFIG{'PDB_ID_SUFFIX'};
    foreach my $hsp (@HSPs) {
        my %Map;
        $Map{Hsp} = $hsp;

        # IAA: Remove GI prepended by some versions of BLAST
        $hsp->{Hit_id} =~ s/^gi\|[^|]+\|//;

        #----------------------------------------
        # Check if id is really pdb|id|chain
        # and modify Id and Desc if necessary
        #----------------------------------------
        unless ($hsp->{Hit_id} =~ /^$dbname\|\S+\|\S+/) {
          if ($hsp->{Hit_def} =~ /^$dbname\|\S+\|\S+\s+/) {
            ($hsp->{Hit_id}, $hsp->{Hit_def}) = split /\s+/, $hsp->{Hit_def}, 2;
          } else {
            warn "Problems identifying database $CONFIG{PDB_ID_SUFFIX} in '$hsp->{Hit_id}' or '$hsp->{Hit_def}'";
            next;
          }
        }
        #----------------------------------------

        #----------------------------------------
        # Set DB, PDB_id, Chain
        #----------------------------------------
        ($hsp->{DB}, $hsp->{PDB_id}, $hsp->{PDB_chain}) = split /\|/, $hsp->{Hit_id};
        #----------------------------------------

        #----------------------------------------
        # Find in HSP corresponding PDB position
        #----------------------------------------
        $hsp->{PDB_pos} = PPH::3D::pos2pdb($hsp->{PDB_id},
                                           $hsp->{PDB_chain},
                                           $hsp->{Hit_pos},
                                           $hsp->{DB});
        ##### |    set_3D_parameter returns: $hsp->{PDB_pos}
        #----------------------------------------

        if ($hsp_no == 1 || $CONFIG{STRUCTALLHITS}) {
            #----------------------------------------
            # DSSP info
            #----------------------------------------
            my $DSSP = PPH::3D::get_DSSP($SNP->{aa1}, $SNP->{aa2}, $hsp);
            ##### |    set_3D_parameter returns: $DSSP
            (keys(%$DSSP)) and $Map{Params} = $DSSP;
            #----------------------------------------

            #----------------------------------------
            # Hbonds
            #----------------------------------------
            ##### |     Hit_pos or Hit_aa?
            my @HBonds = $CONFIG{CALCHBONDS} ? PPH::3D::get_HBonds($hsp) : ();
            (@HBonds) and $Map{Params}{Hbonds} = [ @HBonds ];
            #----------------------------------------

            #----------------------------------------
            # B Factor
            #----------------------------------------
            $Map{Params}{NormedB} = PPH::3D::get_BFactor($hsp);
            #----------------------------------------
        }
        #end of StructAllHits ------------------

        #----------------------------------------
        # Check now for possible SNP contacts
        # to UniProt annotated residues
        #----------------------------------------
        if ($hsp_no == 1 || $CONFIG{CONTALLHITS}) {

            ##### |    This needs testing
            my %ValidContacts;
            #----------------------------------------
            # Map Annot of critical sites to struct
            #----------------------------------------
            if (exists($SNP->{Annot}) and exists($SNP->{Annot}{CritSites})) {
                foreach my $site (@{$SNP->{Annot}{CritSites}}) {
                    # 1st map query site pos to hit pos
                    my ($pos, undef, undef) = PPH::Align::pos2hit($hsp, $site);
                    defined($pos) or next;
                    # 2nd map hit pos to pos in pdb
                    $pos = PPH::3D::pos2pdb($hsp->{PDB_id},
                                            $hsp->{PDB_chain},
                                            $pos,
                                            $hsp->{DB});
                    defined($pos) or next;
                    # concatenate pos & chain
                    push @{$Map{AsMapped}}, ($pos.$hsp->{PDB_chain});
                }
            }
            #----------------------------------------

            #----------------------------------------
            # Get contacts of SNP in structure
            # (if there are sites also look for
            #  intrachain contacts - last argument)
            #----------------------------------------
            my $FoundContacts = PPH::3D::get_contacts($hsp,
                                                      exists($Map{AsMapped}));
            #----------------------------------------

            #----------------------------------------
            # Check for critical Sites
            #----------------------------------------

            if (keys(%$FoundContacts) and exists($Map{AsMapped})) {
                foreach my $c (@{$FoundContacts->{INT}}) {

                    (grep { $c->{Pos}.$c->{Chain} eq $_ } @{$Map{AsMapped}}) or next;

                    push  @{$ValidContacts{CritSites}}, $c;

                    (!defined($best_contact{CritSites}{Dist})
                         or ($c->{Dist} < $best_contact{CritSites}{Dist}))
                        and $best_contact{CritSites} = Storable::dclone($c);

                    $best_contact{Number}{CritSites}++;
                }
            }
            #----------------------------------------

            #----------------------------------------
            # HeteroAtoms
            #----------------------------------------
            foreach my $c (@{$FoundContacts->{HET}}) {

                #----------------------------------------
                # Ignore contacts to non biol ligands
                #----------------------------------------
                next if (grep { $_ eq $c->{Name} } @PPH::Data::Non_biol_ligands);
                #----------------------------------------

                push @{$ValidContacts{Heteroatoms}}, $c;
                $best_contact{Number}{Heteroatoms}++;

                #----------------------------------------
                # Need to clone it here to get a clean
                # data structure without refs
                #----------------------------------------
                (!exists($best_contact{Heteroatoms}) or
                     ($c->{Dist} < $best_contact{Heteroatoms}{Dist}))
                    and $best_contact{Heteroatoms} = Storable::dclone($c);
            }
            #----------------------------------------

            #----------------------------------------
            # InterChain
            #----------------------------------------
            if (exists $FoundContacts->{EXT} and @{$FoundContacts->{EXT}}) {

                #----------------------------------------
                # Grep the contacts
                #----------------------------------------
                @{$ValidContacts{InterChain}} = @{$FoundContacts->{EXT}};
                #----------------------------------------

                #----------------------------------------
                # Check for the closest contact
                #----------------------------------------
                foreach my $c (@{$ValidContacts{InterChain}}) {
                    (!exists($best_contact{InterChain})
                         or ($c->{Dist} < $best_contact{InterChain}{Dist}))
                        and $best_contact{InterChain} = Storable::dclone($c);
                    $best_contact{Number}{InterChain}++;
                }
                #----------------------------------------
            }
            # end interchain
            #----------------------------------------

            scalar(keys(%ValidContacts)) and $Map{Contacts} = \%ValidContacts;
        }
        #end of ContAllHits ---------------------

        # Note, that all successfully filtered 3D HSP maps are saved if either one
        # or both of STRUCTALLHITS and/or CONTALLHITS options are set; otherwise
        # only map for a single top hit is saved (this is the default behavior).
        if ($hsp_no == 1 || $CONFIG{STRUCTALLHITS} || $CONFIG{CONTALLHITS}) {
          push @{$SNP->{Struct}{Maps}}, \%Map;
        }

        $hsp_no++;
    }
    #end foreach @HSPs ---------------------

    $SNP->{Struct}{BestContact} = \%best_contact;
    return 1;
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 get_DSSP

  Usage   : PPH::3D->get_DSSP($aa1, $aa2, $hsp)
  Function: Retrieve secondary structure information for a given substitution
  Args    : Amino acid variation 1 & 2
            Object of type HSP (from BLAST)
  Returns : Hashref containing DSSP information

=cut

#----------------------------------------
sub get_DSSP {
    my ($aa1, $aa2, $hsp) = @_;
    #### PPH__3D__get_DSSP: $aa1, $aa2, $hsp

    my ($pdbid, $pdbpos, $chain) =
        ($hsp->{PDB_id}, $hsp->{PDB_pos}, $hsp->{PDB_chain});

    my $pdbid_short   =  lc $pdbid;
    my $dssp_file     = $pdbid_short . '.dssp';
    my $dssp_file_zip = $dssp_file . '.Z';

    #----------------------------------------
    # Already computed (uncompressed)
    #----------------------------------------
    if (PPH::Misc::locate_file($dssp_file,
                               $CONFIG{'DSSP'}, "$CONFIG{PRECOMPATH}/structures", "$CONFIG{SCRATCH}/structures")) {
        # nothing to do $dssp_file will be set
    }
    #----------------------------------------
    # Already computed (compessed)
    #----------------------------------------
    elsif (PPH::Misc::locate_file($dssp_file_zip,
                                  $CONFIG{'DSSP'}, "$CONFIG{PRECOMPATH}/structures", "$CONFIG{SCRATCH}/structures")) {
        warn("WARNING: Gunzip failed for DSSP file: $dssp_file_zip\n"), return
          if system "gunzip -c $dssp_file_zip >$dssp_file 2>/dev/null";
    }
    #----------------------------------------
    # Needs to be computed
    #----------------------------------------
    else {
        my $pdbfile  = _get_pdb_file($pdbid);
        warn("DSSP file cannot be found nor calculated for: $pdbid\n"), return
          unless defined $pdbfile;

        warn("WARNING: $CONFIG{DSSPCMBI} failed for: $pdbfile, $dssp_file\n"), return
          if system "$CONFIG{DSSPCMBI} $pdbfile $dssp_file >/dev/null 2>&1";

        copy($dssp_file, "$CONFIG{SAVEPATH}/structures/")
          or cluck "ERROR: Can't save DSSP file: $dssp_file";
    }

    ##### |      open: $dssp_file
    open my $F, '<', $dssp_file
        or do { warn "WARNING: Can't open DSSP file: $dssp_file\n"; return {} };

    #----------------------------------------
    # Ignore Header
    #----------------------------------------
    while (<$F>) { /RESIDUE\s+AA\s+STRUCTURE\s+BP1\s+BP2\s+ACC/ and last }

    #----------------------------------------
    # Now parse DSSP - use substr bc. not
    # all lines contain all elements!
    #----------------------------------------
    #01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789
    #     |     | |  |                  |                                                                   |     |
    #     RESIDUE AA STRUCTURE BP1 BP2  ACC     N-H-->O    O-->H-N    N-H-->O    O-->H-N    TCO  KAPPA ALPHA  PHI   PSI    X-CA   Y-CA   Z-CA
    #11721   51 Z D  S    S+     0   0  119     -2,-0.5     2,-0.3     1,-0.2    -1,-0.1   0.723  73.9  10.2 -83.1 -22.3   27.4   -4.2    4.8
    #11722   52 Z S        -     0   0    6    -15,-0.2     2,-0.3    13,-0.0   -15,-0.2  -0.986  56.2-152.0-158.0 158.4   25.5   -6.6    2.7
    #11723   53 Z Y  E     -P 11706   0l  19    -17,-2.5   -17,-2.7    -2,-0.3     2,-0.3  -0.982  14.4-164.1-138.5 148.2   24.1  -10.1    2.5
    #11724   54 Z V  E >   -P 11705   0l  14     -2,-0.3     3,-1.0   -19,-0.2     4,-0.2  -0.988  48.8 -17.4-139.8 144.9   23.3  -12.2   -0.7
    #----------------------------------------
    while (<$F>) {
        #----------------------------------------
        my $ntmp   = substr($_,6,5);   $ntmp =~ s/ //g; # Residue no
        my $ctmp   = substr($_,11,1);                   # Chain no
        ($ctmp eq ' ')      and $ctmp =  $CONFIG{ONECHAIN};
        (($ntmp eq $pdbpos) and ($ctmp eq $chain)) or next;
        #----------------------------------------

        #----------------------------------------
        my $aa     = substr($_,13,1);   $aa  =~tr/a-z/C/; # Cystins to C (?StS?)
        my $secstr = substr($_,16,1);                     # Secondary Structure
        ($secstr eq ' ')         and $secstr =  $CONFIG{NOSECSTR};
        #----------------------------------------

        #----------------------------------------
        my $acc    = substr($_,35,3);   $acc =~ s/ //g; # Accessibility
        my $phi    = substr($_,103,6);  $phi =~ s/ //g; # PHI
        my $psi    = substr($_,109,6);  $psi =~ s/ //g; # PSI
        #----------------------------------------

        #----------------------------------------
        my $noracc = ($acc / $PPH::Data::MaxAccessibility{$aa1});
        my $propi  = _Propensity_interval($noracc);
        my $dprop  = abs($PPH::Data::Propensities{$aa1}->[$propi] -
                             $PPH::Data::Propensities{$aa2}->[$propi]);
        my $dvol   = int($PPH::Data::AA_Volume{$aa2} -
                             $PPH::Data::AA_Volume{$aa1});
        #----------------------------------------

        close($F);
        my $mapreg = _map_PhiPsi($phi, $psi);
        ##### |    DSSP returns: $aa, $secstr, $acc, $noracc, $dprop, $dvol, $phi, $psi, $mapreg
        return {
          DSSPaa     => $aa,
          SecStr     => $secstr,
          Acc        => $acc,
          AccNorm    => $noracc,
          dProp      => $dprop,
          dVol       => $dvol,
          Phi        => $phi,
          Psi        => $psi,
          MapRegion  => $mapreg
        };
    }
    close($F);
    ##### |    DSSP returns undef for: $pdbid, $pdbpos, $chain
    return {};
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head3 _map_PhiPsi

 Usage   : _map_PhiPsi($phi, $psi)
 Function: Map the given phi psi angles to areas
 Args    : Phi & Psi angles
 Returns : Mapped region from PPH::Data::PhiPsiMap

=cut

#----------------------------------------
sub _map_PhiPsi {
    my ($phi, $psi) = @_;

    (($phi == 360) or ($psi == 360)) and return '?';

  SWITCH: {
        ($phi >=  179.9) and do { $phi =  179; last SWITCH};
        ($psi >=  179.9) and do { $psi =  179; last SWITCH};
        ($phi <= -179.9) and do { $phi = -179; last SWITCH};
        ($psi <= -179.9) and do { $psi = -179; last SWITCH};
    }
    my $CELL = 10;		# Grads in one Map Cell ?StS?
    my $i    = int((180 + $phi)/$CELL);
    my $j    = int((180 - $psi)/$CELL);

    return substr($PPH::Data::PhiPsiMap[$j], $i, 1);


}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head3 _Propensity_interval

 Usage   : _Propensity_interval
 Function: Return propensisty interval
 Args    : Accessibility
 Returns : Interval 0-6

=cut

#----------------------------------------
sub _Propensity_interval {
    my $Accessibility = @_;

    ##### |    PPH 3D _Propensity_interval

    ($Accessibility == 0 ) and return 0;
    ($Accessibility  < 5 ) and return 1;
    ($Accessibility  < 15) and return 2;
    ($Accessibility  < 30) and return 3;
    ($Accessibility  < 50) and return 4;
    ($Accessibility  < 75) and return 5;
    return 6;
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 get_HBonds

 Usage   : PPH::3D->get_HBonds($hsp)
 Function: Check if SNP is involved in an hbond
 Args    : Object of type HSP (XML BLAST)
 Returns : Array containing bond information

=cut

#----------------------------------------
sub get_HBonds {
    #### PPH__3D__get_HBonds
    my ($hsp) = shift;
    my ($PDB_id, $Qpos, $Qchain) =
        ($hsp->{PDB_id}, $hsp->{PDB_pos}, $hsp->{PDB_chain});

    my $hb2file = lc "$PDB_id.hb2";
    #----------------------------------------
    # Creating unless the hb2 file exists
    #----------------------------------------
    unless (PPH::Misc::locate_file($hb2file,
      "$CONFIG{PRECOMPATH}/structures", "$CONFIG{SCRATCH}/structures")) {

        #----------------------------------------
        # Find pdb file
        #----------------------------------------
        my $PDBid_short = $PDB_id;
        my $file        = _get_pdb_file($PDB_id);
        (defined($file) and -e $file)
            or do {
                warn "Can't find PDB file for: $PDB_id\n";
                return ()
            };
        #----------------------------------------

        #----------------------------------------
        # Execute hbplus
        #----------------------------------------
        ##### hbplus: $file
        open my $EX, '-|', "$CONFIG{HBPLUS} $file 2>&1"
            or do {
                unlink($file.'.hb2');
                confess "ERROR: $CONFIG{HBPLUS} failed for: $file";
            };

        unless (grep /hydrogen bonds found/, do { local $/; <$EX> }) {
            warn "WARNING: $CONFIG{HBPLUS} failed to locate any hydrogen bonds for: $PDB_id\n";
            return ();
        }
        close $EX;

        # Now save created hb2 file
        copy($hb2file, "$CONFIG{SAVEPATH}/structures/")
          or cluck "WARNING: Can't save HB2 file: $hb2file";
    }
    #----------------------------------------

    #----------------------------------------
    # Now parse
    # Example:
    # <---DONOR---> <-ACCEPTOR-->    atom                        ^
    # c    i                          cat <-CA-CA->   ^        H-A-AA   ^      H-
    # h    n   atom  resd res      DA  || num        DHA   H-A  angle D-A-AA Bond
    # n    s   type  num  typ     dist DA aas  dist angle  dist       angle   num
    #
    # D0001-MET N    -0035-HOH O   2.62 MH  -2 -1.00 135.3  1.80  -1.0  -1.0     1
    #----------------------------------------

    ##### |      open:  $hb2file
    open my $HB, '<', $hb2file
        or do {
            warn "WARNING: Can't open hb2 file: $hb2file\n";
            return ();
        };

    # Ignore header
    while (<$HB>) { /angle\s+num\s*$/ and last;  }

    ##### |    Bonds within AA allowed? Then no switch!

    #----------------------------------------
    # If the Qchain is only one chain DSSP
    # contains as sign '_' which has to be
    # converted to '-'
    #----------------------------------------
    ($Qchain eq $CONFIG{'ONECHAIN'}) and $Qchain = '-';
    #----------------------------------------

    #----------------------------------------
    # Now extract the relevant hbonds
    #----------------------------------------
    my @Bonds;
    while (my $l = <$HB>) {

        #----------------------------------------
        # Check if hbplus is ok
        #----------------------------------------
        (length($l) != 76) and do {
            warn "WARNING: hb2 file '$hb2file' seems to be corrupted at line\n'$l'\n";
            return ()
        };
        #----------------------------------------

        #----------------------------------------
        # Extract the relevant info - split won't
        #                             work here
        #----------------------------------------
        my ($chain1,$pos1,$aa1,$atom1) = (substr($l, 0,1), substr($l, 1,4),
                                          substr($l, 6,3), substr($l,10,3));
        my ($chain2,$pos2,$aa2,$atom2) = (substr($l,14,1), substr($l,15,4),
                                          substr($l,20,3), substr($l,24,3));
        my ($Distance, $Type)          = (substr($l,28,4), substr($l,33,2));
        #----------------------------------------

      SWITCH: {
            ($Qchain eq $chain1 and $Qpos eq $pos1)
                and ($Type eq 'SM' or $Type eq 'SS')
                    and do {
                        push @Bonds, "$atom1 -> $chain2 $pos2 $aa2 $atom2 ".
                            "$Distance $Type";
                        last SWITCH;
                    };
            ($Qchain eq $chain2 and $Qpos eq $pos2)
                and ($Type eq 'MS' or $Type eq 'SS')
                    and do {
                        push @Bonds, "$atom2 -> $chain1 $pos1 $aa1 $atom1 ".
                            "$Distance $Type";
                        last SWITCH;
                    };
        }
    }
    close $HB;
    ##### |    get_HBonds returns (n hbonds): scalar(@Bonds)

    return @Bonds;
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 get_BFactor

 Usage   : PPH::3D->get_BFactor($hsp)
 Function: Calculate the B-factor for a certain position
 Args    : Object of type HSP (XML BLAST)
 Returns : Returns the output of the C-Program bfact

=cut

#----------------------------------------
sub get_BFactor {
    my ($hsp) = @_;
    my ($PDB_id, $AApos, $chain) =
        ($hsp->{PDB_id}, $hsp->{PDB_pos}, $hsp->{PDB_chain});

    #### PPH 3D get_BFactor

    my $PDBid_short = $hsp->{PDB_id};
    my $file        = _get_pdb_file($hsp->{PDB_id});
    defined($file) or do { warn "Can't find pdb file: $PDB_id\n"; return ''; };

    my $ret = `$CONFIG{BFACT} $file $AApos $chain 2>&1`;

    ( $ret =~ /NormedB:\s+(\S+)\b/ && $ret !~ /Error/ )
        and return $1;

    chomp($ret); $ret =~ s/\n/ /; $ret =~ s/\s+/ /;
    warn "WARNING: Failed bfact($file, $AApos, $chain) returns: $ret\n";

    return '';
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 pos2pdb

 Usage   : PPH::3D->pos2pdb($id, $chain, $pos, $dbase)
 Function: Map a sequence position to the structure
           - this is necessary since the structure
             often lacks N-terminus and other parts of the sequence
 Args    : Sequence identifier
           PDB chain
           Sequence position
           PDB database name
 Returns : Postion if successful

=cut

#----------------------------------------
sub pos2pdb {
    my ($id, $chain, $pos, $dbase) = @_;

    #### PPH 3D pos2pdb

    $id = lc $id;

    foreach my $fragm ( _PDB_get_fragm($id, $chain) ) {
        ##### |    values of fragm: values(%$fragm)
        (        $fragm->{Beg1} <  $fragm->{End1})
            and (  $fragm->{Beg1} <= $pos)
                and ($pos           <= $fragm->{End1})
                    and return ($pos - $fragm->{Beg1} + $fragm->{Beg2});

        (      $fragm->{Beg1} == $fragm->{End1})
            and ($fragm->{Beg1} == $pos)
                and return $fragm->{Beg2};
    }
    confess "ERROR: Can't find residue $pos in structure '$id' chain '$chain'";
}
#----------------------------------------------------------------------


#----------------------------------------------------------------------

=head2 _PDB_get_fragm

 Usage   : _PDB_get_fragm($pdbid, $chain)
 Function: Retrieve from PDB fragment file the entries corresponding
           to the pdb identifier and chain

           (fragm1 corresponds to positions in fasta format,
           fragm2 to 'real' numbers given in ATOM records)
           id   chain fragm1_beg fragm1_end  fragm2_beg fragm2_end
           102m   _      1          154          0         153
           103l   _      1           34          1          34
           103l   _     35           35        40A         40A
           103l   _     36           36        40B         40B
           103l   _     37           37        40C         40C
           103l   _     38          159         41         162

 Args    : PDB identifier
           CHAIN of interes
 Returns : @( { Beg1, Beg2, End1 } )

=cut

#----------------------------------------

sub _PDB_get_fragm {
    #### PPH 3D PDB_get_fragm
    my ($pdbid, $chain) = @_;
    my $file =
        "$CONFIG{PDB2FASTA}/$CONFIG{PDB_ID_SUFFIX}.fragm";

    open my $F, '-|', "grep -i '^$pdbid' $file"
        or confess "ERROR: PPH::3D::PDB_get_fragm failed to read file $file";

    my @OUT;

    while (<$F>) {
        /^$pdbid\s+$chain/i or next;
        chomp;
        my ($id, $chain, $beg1, $end1, $beg2) = split;
        push @OUT, {Beg1 => $beg1, Beg2 => $beg2, End1 => $end1};
    }
    close $F;
    return @OUT;
}
#----------------------------------------------------------------------
sub PDB_read_fragm {
    confess 'ERROR: Obsolete';
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 get_contacts

 Usage   : PPH::3D->get_contacts($hsp, $intra_chain_flag)
 Function: Find Contacts in structure for given position
 Args    : Object of type HSP (XML BLAST)
           Argument for C-Program dist wether to check for intra chain contacts
 Returns : Hashref with different contact types
           { INT | EXT => [ { Pos, Name, Dist} ] }

=cut

#----------------------------------------
sub get_contacts {
    #### PPH 3D get_contatcs
    my ($hsp, $intra_chain_flag) = @_;
    my ($PDB_id, $pos, $chain) =
        ($hsp->{PDB_id}, $hsp->{PDB_pos}, $hsp->{PDB_chain});

    my $file = _get_pdb_file($PDB_id);
    defined($file)
        or do {
            warn "WARNING: PDB file '$PDB_id' can't be found"; return {}
        };

    my $PDBid_short = $PDB_id;

    #----------------------------------------

    #----------------------------------------
    # Most time chain is in uppercase
    # - there are structure such as 1d3i which
    #   contain chains with lower case:
    # - so first try chain and then lc of it
    #----------------------------------------

    my $flag = $intra_chain_flag ? '-i' : '';

    my $exec = $CONFIG{DIST}." $flag -r $pos ".uc($chain).' '.$CONFIG{CONTTHRESH}." $file 2>&1";
    ##### |    exec: $exec
    open my $DIST, '-|', $exec or do {
        ### Error executing dist: $exec
        return {}
    };

    my $ret = do { local $/; <$DIST> };
    if (grep /Error|Usage/, $ret) {
        close $DIST;
        $exec = $CONFIG{DIST}." $flag -r $pos ".lc($chain).' '.$CONFIG{CONTTHRESH}." $file 2>&1";
        ##### |    retry exec: $exec
        open $DIST, '-|', $exec or do {
            ### Error executing dist: $exec
            return {}
        };
        $ret = do { local $/; <$DIST> };
        if (grep /Error|Usage/, $ret) {
            close $DIST;
            ### Error running dist: $exec
            return {}
        };
    }

    close $DIST;

    # 1thf   INT     1D MET  O  4     m >...< m     3D ALA  N  17       3.538
    my %ret;
    foreach (split /^/, $ret) {
        my ($type, $atom1, $pos2, $aa2, $atom2, $dist) =
            (split /\s+/, $_)[1,4,9,10,11,13];
        my $chain2;
        ($pos2, $chain2) = $pos2 =~ /(^-?\d+)(\w)/;
        push @{$ret{$type}}, {
            Pos => $pos2,
            Chain => $chain2,
            Atom1 => $atom1,
            Atom2 => $atom2,
            Name => $aa2,
            Dist => $dist
        };
    }
    ##### |    contacts returns: %ret
    return \%ret;
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head3 _get_pdb_file

 Usage   : _get_pdb_file($pdbid)
 Function: returns PDB filename
           - it unzips PDB if necessary and marks this file as GARBAGE
 Args    : PDB identifier
 Returns : filename

=cut

#----------------------------------------
sub _get_pdb_file {
    # pdbid used in PDB filenames is always lowercase
    my $pdbid   = lc shift;
    ##### PPH__3D__get_pdb_file: $pdbid

    my $tmpfile = "$pdbid.pdb";
    my $file = "$CONFIG{PDB}/$CONFIG{PDB_PREFIX}$pdbid.$CONFIG{PDB_SUFFIX}";

    # Return if we have uncompressed temporary file in the current working dir
    return $tmpfile if -e $tmpfile;

    # Return if we have an uncompressed file available in the local PDB mirror
    return $file if -e $file;

    # Look up compressed file in the local PDB mirror
    my ($zipped, $zfound);
    foreach my $ext (qw(.gz .Z)) { $zipped = $file.$ext; $zfound = 1, last if -e $zipped; }
    warn("WARNING: PDB file not found for entry: $pdbid\n"), return unless $zfound;

    # Create temporary uncompressed copy of the PDB file in the current dir
    my $status = system("gunzip -c $zipped >$tmpfile 2>/dev/null");
    warn("WARNING: Gunzip failed for PDB file: $zipped\n"), return unless $status == 0;

    return $tmpfile;
}

#----------------------------------------------------------------------

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
1;
