package SNP;
use warnings;
use strict;
#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

SNP

=head1 DESCRIPTION

Object containing all information about a SNP

=head1 AUTHOR

StS

=head1 SUBVERSION

 $LastChangedDate: 2009-11-28 00:15:09 -0500 (Sat, 28 Nov 2009) $
 $LastChangedRevision: 254 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use Carp qw(cluck confess);
use XML::Simple;
use Data::Dumper;
use Storable;

#BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
# Begin
BEGIN {
use Exporter ();
  our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
  $VERSION = 0.1;
  @ISA           = qw(Exporter);
  @EXPORT        = qw();
  @EXPORT_OK     = qw();
  %EXPORT_TAGS = ();
}
our @EXPORT_OK;
#BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB

#CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
# Constructor
sub new {
  my ($class, $self) = @_;
  bless($self, $class);
  my $msg = $self->_test_init();
  die "$msg\n" if $msg;
  return $self;
}

sub new_empty {
  my ($class, $self) = @_;
  $self = {};
  bless($self, $class);
  return $self;
}
#CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
# Routines

#----------------------------------------------------------------------

=head1 ROUTINES

=head3 _test_init

 Usage   : _test_init()
 Function: Simply test if initial SNP is ok
 Args    : None (Object of type SNP)
 Returns : TRUE if successful

=cut
#----------------------------------------
sub _test_init {
  my ($self) = shift;

  # If there is ONLY a dbSNP rsID present then there is nothing to test
  return '' if $self->{SNP_Id} =~ /^rs/ &&
    !defined $self->{Acc} &&
    !defined $self->{Pos} &&
    !defined $self->{aa1} &&
    !defined $self->{aa2};

  # Check for mandatory parameters
  defined($self->{Acc})                        or return 'Protein accession missing';
  defined($self->{Pos})                        or return 'Substitution position missing';
  defined($self->{aa1})                        or return 'Reference residue (AA1) missing';
  defined($self->{aa2})                        or return 'Variant residue (AA2) missing';

  # Remove spurious leading / trailing blanks
  foreach my $key (keys %{ $self }) {
    $self->{$key} =~ s/^\s+//;
  }

  ($self->{Pos}  =~ /^\d+$/)                   or return "Illegal position specified: $self->{Pos}";
  ($self->{aa1} =~ /^[GPAVLIMCFYWHKRQNEDST]$/) or return "Illegal AA1 residue code: $self->{aa1}";
  ($self->{aa2} =~ /^[GPAVLIMCFYWHKRQNEDST]$/) or return "Illegal AA2 residue code: $self->{aa2}";

  ($self->{aa1} ne $self->{aa2})               or return "Synonymous substitution: $self->{aa1} -> $self->{aa2}";

  return '';
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 snp2dump

 Usage   : SNP->print_tree(\*F)
           SNP::snp2dump($SNP, \*STDOUT)
 Function: Print SNP structure
 Args    : Optional FILE pointer, otherwise output to STDOUT
 Return  : TRUE if successful

=cut
#----------------------------------------
sub snp2dump {
  my $self = shift;
  my $OUT = shift;
  print $OUT Dumper($self);
  return 1;
}
#-----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 snp2xml

 Usage   : snp2xml($$);
           $SNP->snp2xml(\*STDOUT);
           SNP::snp2xml($SNP, \*F);
 Function: Returns XML output of data structure SNP
           - This routine is necessary since XML::Simple doesn't support
             dumps of data structures like [ { a => {}, b => {} }, {} ];
           - The XML output should be safe to be parsed by XML::Simple,
             However, in case of a structure [ {} ] an extra field is
             added [ {}, {} ] in order to maintain the data structure
 Args    : $SNP - data structure
           FilePointer;
 Returns : TRUE if successful;
 IAA     : Changed to use XMLout()

=cut

#----------------------------------------
my @LEVEL;      # Levels of XML Tree & Token
my $TAB = "\t";
my $O;
sub snp2xml {
  my $d = shift;
  $O   = shift;
  @LEVEL = ();

  confess "Data Structure $d is not of type SNP" unless ((ref $d) eq 'SNP');

  XMLout($d, OutputFile=>$O, RootName=>'SNP', AttrIndent=>1, KeyAttr=>[], SuppressEmpty=>1);
  return 1;

  print $O "<SNP>\n";
  eval_hash($d);
  print $O "</SNP>\n";
  return 1;

  #----------------------------------------
  # parse hash
  #----------------------------------------
  sub eval_hash {
    my $d = shift;
    while (my ($k, $v) = each(%$d)) {
      push @LEVEL, $k;
      eval_ref($k, $v);
      pop @LEVEL;
    }
  }
  #----------------------------------------

  #----------------------------------------
  # parse array
  #----------------------------------------
  sub eval_array {
    my $d = shift;
    foreach my $v (@$d) {
      my $k = $LEVEL[$#LEVEL];
      eval_ref($k, $v);
    }
  }
  #----------------------------------------

  #----------------------------------------
  # evaluate reference v
  #----------------------------------------
  sub eval_ref {
    my ($k, $v) = @_;
    my $r = ref $v; ($r eq "SNP" and $r = "HASH");
    my $l = $TAB x @LEVEL;

    #----------------------------------------
    # v is simply a value
    # -> no further recursion needed
    #----------------------------------------
    if ($r eq "") {
      # Ensure valid XML - replace '&' by '&amp;'
      $v = "" unless(defined($v));
      $v =~ s/&([^amp;|quot;|lt;|gt;|apos;])/&amp;$1/g;
      $v =~ s/\|_$/\|A/; print $O "$l<$k>$v</$k>\n"
    }
    #----------------------------------------
    # v is an ARRAY
    #----------------------------------------
    elsif ($r eq "ARRAY") {
      eval_array($v);
      #----------------------------------------
      # This is annoying but we need to add an
      # undefined array to maintain data structure
      # if we parse XML with XML::Simple
      #----------------------------------------
      if (@$v == 1) { print $O "$l<$k></$k>\n"}
    }
    #----------------------------------------
    # v is a HASH
    #----------------------------------------
    elsif ($r eq "HASH")  {
      print $O "$l<$k>\n"; eval_hash($v); print $O "$l</$k>\n"
    }
    else {
      confess "something wrong '$k' '$v' '$r'\n";
    }
  }
  #----------------------------------------
}
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 xml2snp

 Usage   : xml2snp($);
           $SNP->xml2snp(\*STDOUT);
           SNP::xml2snp($SNP, \*F);
 Function: Loads the SNP-XML file in an SNP object
 Args    : $SNP - data structure
           FilePointer;
 Returns : TRUE if successful;

=cut

#----------------------------------------
sub xml2snp {
  my $SNP     = shift;
  my $xmlfile = shift;

  %$SNP = %{ XMLin($xmlfile, KeyAttr=>[], ForceArray=>[ 'Avail', 'WEKA' ]) };

  return 1;
}



#RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR

1;
