#!/usr/bin/env perl
use strict;

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

=head1 NAME

uniprot.pl

=head1 SYNOPSIS

uniprot.pl [options]

Where options are:

 -n <organism>    specify common organism name instead of the default human;
                  supported are: mouse, rat, dog, and zebrafish
 -s <dat>[,<idm>] specify local source file(s): <dat> with the protein entries
                  in UniProtKB format and optional <idm> with idmapping data;
                  default is to fetch all source files from ftp.uniprot.org
 -k               keep source files; default is to remove source files after
                  they are successfully processed
 -h               print this help

=head1 DESCRIPTION

Downloads Swiss-Prot, TrEMBL, ID mapping, and VARSPLIC data from ftp.uniprot.org
and prepares protein sequence database and other auxillary database files
for use with the PolyPhen-2

=head1 AUTHOR

 Vasily Ramensky
 Steffen Schmidt
 Ivan Adzhubey

=head1 SUBVERSION

 $LastChangedDate: 2010-03-29 01:33:25 -0400 (Mon, 29 Mar 2010) $
 $LastChangedRevision: 288 $
 $LastChangedBy: ivan $

=cut

#DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD

use LWP::Simple;
use Getopt::Std;
use Pod::Usage;
use Storable qw(nstore);

my @fttokens = qw(
  DISULFID THIOLEST THIOETH MOD_RES LIPID CARBOHYD METAL BINDING ACT_SITE SITE
  SIGNAL TRANSIT PROPEP TRANSMEM INIT_MET
  ## PEPTIDE CA_BIND DNA_BIND NP_BIND ZN_FING
);

my $UNIPROT_FTP  = 'ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase';
#-----------------------------------------------------------------------------

my %opts;
getopts('hkn:s:', \%opts);

exists $opts{'h'} and pod2usage({ -verbose=>99 });

# Organism common name: human (default), mouse, rat, dog, zebrafish
my $CNAME = 'human';

# SPECIES hash indexed by CNAME
my %SPECIES = (
  # Human
  'human'    => {
    'os'     => 'Homo sapiens',
    'spname' => 'HUMAN',
    'file'   => 'human',
  },
  # Orangutan
  'orangutan'    => {
    'os'     => 'Pongo pygmaeus|Pongo abelii',
    'spname' => 'PONPY|PONAB',
    'file'   => 'mammals',
  },
  # Mouse
  'mouse'    => {
    'os'     => 'Mus musculus',
    'spname' => 'MOUSE',
    'file'   => 'rodents',
  },
  # Rat
  'rat'      => {
    'os'     => 'Rattus norvegicus|Rattus rattus',
    'spname' => 'RAT|RATRT',
    'file'   => 'rodents',
  },
  # Dog
  'dog'      => {
    'os'     => 'Canis familiaris',
    'spname' => 'CANFA',
    'file'   => 'mammals',
  },
  # Zebrafish
  'zebrafish' => {
    'os'      => 'Danio rerio',
    'spname'  => 'DANRE',
    'file'    => 'vertebrates',
  },
);

if (my $cname = $opts{'n'}) {
  die "Illegal or unsupported organism name: $cname\n"
    unless exists $SPECIES{$cname};
  $CNAME = $cname;
}

my $seqfile     = "$CNAME.seq";
my $ftfile      = "$CNAME.ft";
my $swfile      = "$CNAME.dat";
my $id2accfile  = "$CNAME.id2acc";
my $snp2accfile = "$CNAME.snp2acc";
my $xdb2accfile = "$CNAME.xdb2acc";

#-----------------------------------------------------------------------------

my @names = (
  'uniprot_sprot_'  . $SPECIES{$CNAME}{'file'} . '.dat',
  'uniprot_trembl_' . $SPECIES{$CNAME}{'file'} . '.dat',
  'idmapping_selected.tab'
);

if (defined $opts{'s'}) {

  my @files = split /,/, $opts{'s'};
  foreach my $file (@files) {
    foreach my $ext (qw{ .seq .ft .dat .id2acc .snp2acc .xdb2acc}) {
      die "Can't simultaneously read from and write to $file, please rename input file\n"
        if $file eq "$CNAME$ext";
    }
  }
  @names = @files;

} else {

  my $suf = '.gz';

  foreach my $name (@names) {
    my $infile;
    if ($name =~ /^idmapping/) {
      $infile = "$UNIPROT_FTP/idmapping/${name}${suf}";
    } else {
      $infile = "$UNIPROT_FTP/taxonomic_divisions/${name}${suf}";
    }
    my $outfile = "${name}${suf}";
    unlink $outfile if -e $outfile; # otherwise mirror() will fail
    warn "Fetching $infile ...\n";
    my $rc = mirror($infile, $outfile);
    die "Error fetching URL $infile: ".status_message($rc)."\n" unless is_success($rc);
    warn "Decompressing $outfile ...\n";
    `gzip -d $outfile`;
    warn "Done.\n";
  }

}

open(SEQ,     ">$seqfile") || die "Can't open $seqfile";
open(FT,      ">$ftfile" ) || die "Can't open $ftfile";
open(SWALL,   ">$swfile" ) || die "Can't open $swfile";
open(ID2ACC,  ">$id2accfile" ) || die "Can't open $id2accfile";
open(SNP2ACC, ">$snp2accfile" ) || die "Can't open $snp2accfile";

my $dbase_len = 0;

foreach my $file (@names) {

 next if $file =~ /idmapping/;

 warn "Processing $file ...\n";

 my $num_filtered = my $copy = 0;
 my (@entrylines, @snps, @ftlines, @seqlines, @accs, @emblrefs);
 my ($id, $seq, $de, $len, $varpos);

 open(TMP, $file) || die "Can't open $file\n";

 LINE: while (<TMP>) {

  if(/^ID.*\b(\d+) AA/) {
   @seqlines = @snps = @ftlines = @entrylines = @accs = @emblrefs = ();
   $copy = 0;
   $id = (split)[1];
   $len = $1;
   undef $de;
   undef $varpos;
  } # end if
  elsif( /^OS/ ) {
   $copy++ if m/$SPECIES{$CNAME}{'os'}/o; ## there exist multiple OS-lines
  } # end if
  elsif( /^AC/ ) {
   s/;//g;
   my @accs_tmp = (split(/\s+/, $_));
   shift @accs_tmp; # remove ^AC
   push @accs, @accs_tmp;
  } # end elsif
  elsif( /^FT   (\w+)\s/ ) { ## important to put exact spaces in regexp
    my $feature = $1;
    if ($feature eq 'VARIANT') {
      # Read and merge all continuation lines for FT VARIANT
      # since the 'dbSNP:' tag(s) can appear on any of them.
      chomp(my $varlines = $_);
      my $inline;
      while ($inline = <TMP>) {
        last unless substr($inline, 2, 32) eq ' ' x 32;
        chomp($varlines .= ' ' . substr($inline, 34));
      }
      my $from = substr($varlines, 14, 6); $from =~ s/\s+//g;
      my $to   = substr($varlines, 21, 6); $to   =~ s/\s+//g;
      # Validate that the line annotates a single residue substitution
      if ($from =~ /^\d+$/ && $to =~ /^\d+$/ && $to - $from == 0) {
        my $description = substr($varlines, 34);
        # There can be multiple SNPs annotated at the same position
        while ($description =~ /\b([A-Z]) -> ([A-Z])\s+\(in (.+?)\)/g) {
          my ($aa1, $aa2, $pointer) = ($1, $2, $3);
          # We have a dbSNP annotation
          if ($pointer =~ /dbSNP:(rs\d+)/) {
            push @snps, [ $1, $from, $aa1, $aa2 ];
          }
        }
      }
      # since we already have next input line in the $inline buffer
      $_ = $inline, redo LINE;
    } else {
      push @ftlines, $_ if grep { $1 eq $_ } @fttokens;
    }
  } # end elsif
  elsif( /^DE\s+(\w.*)\n$/ ) {
   $de .= " $1";
  } # end elsif
  elsif( /^ / ) {
   $seq = $_;
   $seq =~ s/ //g;
   push @seqlines, $seq;
   $dbase_len += length($seq);
  } # end elsif
  elsif( /^\/\// ) {
   if ($copy && $id !~ /ALU\d_HUMAN/) {
    print SWALL @entrylines;
    print SWALL "//\n";

    my $primacc = $accs[0];

    chomp(@seqlines);
    print SEQ ">sp|$primacc|$id$de LENGTH: $len AA\n";
    print SEQ join '', @seqlines, "\n";
    print FT map { "$primacc $_" } @ftlines;
    print ID2ACC map { "$_ $primacc\n" } @accs;
    print ID2ACC "$id $primacc\n";
    print SNP2ACC map { "$_->[0]\t$primacc\t" . join("\t",@$_[1..3]) . "\n" } @snps if @snps;

    $num_filtered++;

   } # end if
  } # end elsif

  push @entrylines, $_;

 } # end while

 close TMP;

 warn "Found $num_filtered $CNAME proteins.\n";

 unlink $file unless defined $opts{'k'};

} # end foreach

warn "Total length of protein sequences in '$CNAME' database: $dbase_len\n";

close SNP2ACC; close ID2ACC; close SWALL; close FT; close SEQ;

warn "Gzipping $CNAME.dat ...\n";
unlink "$CNAME.dat.gz" if -e "$CNAME.dat.gz";
`gzip $CNAME.dat`;
warn "Done.\n";

if (@names > 1) {
  my $file = $names[-1];
  warn "Processing $file ...\n";
  my @dbtags    = qw( acc pid geneid refseq gi );
  my %db2acc;
  my $orgtag = '_' . $SPECIES{$CNAME}{'spname'};
  open DBFIN, '<', $file or die "Can't open $file: $!\n";
  while (<DBFIN>) {
    # We only process first 5 columns: acc, pid, geneid, refseq, gi
    # Note: acc and pid are both unique (acc<->pid) and only primary
    # accessions are included. All other IDs can be lists of multiple
    # emtries separated by '; ' characters.
    my @a = split /\t/;
    next unless $a[1] =~ /^\w+$orgtag$/o;  # skip all but CNAME proteins
    my $acc = $a[0];
    for (my $i=3; $i<=4; $i++) {
      my $dbids = $a[$i];
      next unless length $dbids;
      foreach my $dbid (split /;\s*/, $dbids) {
        $dbid =~ s/^\s+//; $dbid =~ s/\s+$//;
        next unless length $dbid;
        # Skip all except the first xdb ID encountered in case of ambiguous mappings
        $db2acc{$dbtags[$i]}{$dbid} = $acc unless exists $db2acc{$dbtags[$i]}{$dbid};
        # Some DB IDs in idmapping file have version numbers appended to them
        # (e.g. RefSeqs) but users do not include versions in the input more than
        # often. As a workaround, use both versioned and unversioned IDs as keys
        # pointing to the same protein accession.
        if ($dbid =~ s/\.\d+$//) {
          $db2acc{$dbtags[$i]}{$dbid} = $acc unless exists $db2acc{$dbtags[$i]}{$dbid};
        }
      }
    }
  }
  close(DBFIN);
  nstore \%db2acc, $xdb2accfile or die "Can't store $xdb2accfile: $!\n";
  warn "File saved: $xdb2accfile\n";
  warn "Warning: file has zero size: $xdb2accfile\n" unless -s $xdb2accfile;
  if ($opts{'k'}) {
  warn "Compressing $file ...\n";
    `gzip -f $file`;
  } else {
    unlink $file or die "Can't delete file: $file\n";
  }
}

# Download VARSPLIC file with the isoform sequences and convert it
# to a greppable form (i.e., with the sequences on single line)
my $gzfile  = 'uniprot_sprot_varsplic.fasta.gz';
my $fafile  = $gzfile; $fafile =~ s/\.gz$//;
my $outfile = "$CNAME.varsplic";
my $url  = "$UNIPROT_FTP/complete/$gzfile";
unless ($opts{'s'}) {
  warn "Fetching $url ...\n";
  unlink $gzfile if -e $gzfile; # otherwise mirror() will fail
  my $rc = mirror($url, $gzfile);
  die "Error fetching URL $url: ".status_message($rc)."\n" unless is_success($rc);
  warn "Decompressing $gzfile ...\n";
  `gzip -df $gzfile`;
}
# Skip VARSPLIC file processing if -s option was specified
# but uniprot_sprot_varsplic.fasta file is not present
unless ($opts{'s'} && ! -e $fafile) {
  warn "Processing $fafile ...\n";
  my $orgtag = '_' . $SPECIES{$CNAME}{'spname'};
  open(FIN, $fafile) or die "Can't open file: $fafile\n";
  open(FOUT, ">$outfile") or die "Can't create temporary file: $outfile\n";
  my $seqline = '';
  my $defline = '';
  while (<FIN>) {
    chomp;
    if (/^>/) {
      if (length $defline) {
        my ($def, $desc) = split /\s+/, $defline, 2;
        if ($def =~ /\w+$orgtag$/o) {
          print FOUT $defline, "\n";
          print FOUT $seqline, "\n" if length $seqline;
        }
      }
      $defline = $_; $seqline = '';
      next;
    }
    $seqline .= $_;
  }
  if (length $defline) {
    my ($def, $desc) = split /\s+/, $defline, 2;
    if ($def =~ /\w+$orgtag$/o) {
      print FOUT $defline, "\n";
      print FOUT $seqline, "\n" if length $seqline;
    }
  }
  close(FOUT);
  close(FIN);
  if ($opts{'k'}) {
  warn "Compressing $fafile ...\n";
    `gzip -f $fafile`;
  } else {
    unlink $fafile or die "Can't delete file: $fafile\n";
  }
}

warn "All done.\n";

#-----------------------------------------------------------------------------
